/*
 * Copyright (C) Microsoft Corporation, 1982
 */

#include "y.tab.h"
#include "a86.h"
#include <a.out.h>
#include <relsym.h>

#define loc (*locptr)
#define Putc(val,file) if(loctype != A_BSS) putc(val,file)
#define Putw(val,file) if(loctype != A_BSS) putw(val,file)

fix_up(name)
char *name;
{
	register int typ;
	register int val,rel;
	int mval,mrel;
	struct aexec head;
	struct Sym *sym;
	struct Expr *expr;
	long skip;
	FILE *out_text,*out_data,*out_trel,*out_drel;
	FILE *output,*outrel;
	unsigned *locptr = &Text;
	int loctype = A_TEXT;
	unsigned old_loc = 0;

	Text = (Text + 1) & -2;
	Data = (Data + 1) & -2;
	Bss = (Bss + 1) & -2;

	output = out_text = fopen(name,"w");
	out_data = fopen(name,"a");
	outrel = out_trel = fopen(name,"a");
	out_drel = fopen(name,"a");
	if ((output == NULL) || (out_data == NULL)) {
		fprintf(stderr,"cannot open output file %s\n",name);
		exit(4);
	}
	skip = sizeof(head) + (long)Text;
	fseek(out_data,skip,0);
	skip = sizeof(head) + (long)Text + (long)Data;
	fseek(out_trel,skip,0);
	skip = sizeof(head) + 2*(long)Text + (long)Data;
	fseek(out_drel,skip,0);

	/* build and output header */
	head.xa_magic = A_MAGIC1;
	head.xa_text = Text;
	head.xa_data = Data;
	head.xa_bss = Bss;
	head.xa_syms = Globals * sizeof(struct nlist);
	head.xa_entry = Entry;
	head.xa_unused = 0;
	head.xa_flag = 0;
	fwrite((char *)&head,sizeof(head),1,output);

	rewind(object);
	Linenum = 1;
	Text = Data = Bss = 0;

	while ( (typ = getc(object)) != R_END && typ != EOF ) {
#if DEBUG
		dbprt("fix","loc cnt value %x,old value %x\n",loc,old_loc) ;
#endif
		if ( typ >= R_LINE ) {
			if ( typ == R_LINE ) {
				Linenum++;
			} else if ( typ == R_TEXT && loctype != A_TEXT ) {
				/* switch to text location counter */
				locptr = &Text;
				loctype = A_TEXT;
				old_loc = loc & -2;
				output = out_text;
				outrel = out_trel;
#if DEBUG
				dbprt("fix","switching to text obj and rel\n") ;
#endif
			} else if ( typ == R_DATA && loctype != A_DATA ) {
				/* switch to data location counter */
				locptr = &Data;
				loctype = A_DATA;
				old_loc = loc & -2;
				output = out_data;
				outrel = out_drel;
#if DEBUG
				dbprt("fix","switching to text obj and rel\n") ;
#endif
			} else if ( typ == R_BSS && loctype != A_BSS ) {
				locptr = &Bss;
				loctype = A_BSS;
				old_loc = loc & -2;
#if DEBUG
				dbprt("fix","switching to bss loc counter\n") ;
#endif
			}
		} else { 
			switch ( typ & 017 ) {
				case A_ABS:
					if ( typ & R_BYT )
						val = getc(object);
					else
						val = getw(object);
					rel = 0;
					break;
				case A_TEXT:
				case A_DATA:
				case A_BSS:
					val = getw(object);
					rel = (typ & 017);
					break;
				case A_UND:
					sym = (struct Sym *)getw(object);
					evalsym(sym);
					if ( sym->sy_type & A_UEXT )
						val = 0;
					else
						val = sym->sy_value;
					rel = sym->sy_type;
					break;
				case A_EXPR:
					expr = (struct Expr *)getw(object);
					evalexp(expr,&mval,&mrel);
					val = mval;
					rel = mrel;
					break;
			}
			if ( typ & R_PC_REL ) {
				if ( rel != loctype && (rel & A_UEXT) == 0 ) {
					error("Illegal jump out of segment","");
					if ( typ & R_BYT ) {
						Putc(0,output);
#if DEBUG
						dbprt("fix",
						"output byte value 0\n") ;
#endif
						loc++;
					} else {
						Putw(0,output);
#if DEBUG
						dbprt("fix",
						"output word value 0\n") ;
#endif
						loc += 2;
					}
					rel = 0;
				} else if ( typ & R_BYT ) {
					loc++;
					val -= loc;
					if ( rel & A_UEXT )
						error("Short jump to external not allowed","");
					else if ( val < -128 || val > 127 )
						error("Short jump out of range","");
					rel = 0;
					Putc(val,output);
#if DEBUG
						dbprt("fix",
						"output byte value %x\n",val) ;
#endif
				} else {
					loc += 2;
					val -= loc;
					if ( rel & A_UEXT )
						rel |= 1; /* mark as pc relative */
					else
						rel = 0;
					Putw(val,output);
#if DEBUG
					dbprt("fix",
					"output word value %x\n",val) ;
#endif
				}
			} else {
				if ( typ & R_BYT ) {
					loc++;
					Putc(val,output);
#if DEBUG
					dbprt("fix",
					"output byte value %x\n",val) ;
#endif
					if ( rel != 0 ) {
						error("Relocatable values not allowed for bytes","");
						rel = 0;
					}
				} else {
					loc += 2;
					Putw(val,output);
#if DEBUG
					dbprt("fix",
					"output word value %x\n",val) ;
#endif
				}
			}
			if ( loc-old_loc > 1 ) {
				if ( loc & 1 && rel != 0 )
					rel |= 0x8000;
				Putw(rel,outrel);
#if DEBUG
				dbprt("fix","output rel value %x\n",rel) ;
#endif
				old_loc = loc & -2; /* remember last reloc word output */
			}
		}
	}
	if ( Text & 1 ) {
		Putc(0,out_text);
		Putw(0,out_trel);
	}
	if ( Data & 1 ) {
		Putc(0,out_data);
		Putw(0,out_drel);
	}
	fclose(out_text);
	fclose(out_data);
	fclose(out_trel);
	dumpsym(out_drel);

	/* if more globals added on fixup pass then rewrite header */
	if ( head.xa_syms != Globals*sizeof(struct nlist) ) {
		head.xa_syms = Globals * sizeof(struct nlist);
		rewind(out_drel);
		fwrite((char *)&head,sizeof(head),1,out_drel);
	}
	fclose(out_drel);
}

#if DEBUG
obj_list()
	{
	register int val,rel,typ ;
	int mval,mrel;
	struct Sym *sym ;
	struct Expr *expr ;
	rewind(object);
	Linenum = 1;
	Text = Data = Bss = 0;

	while ( (typ = getc(object)) != R_END && typ != EOF ) {
		dbprt("obj","\nrecord typ - %x, ",typ) ;
		if ( typ >= R_LINE ) {
			if ( typ == R_LINE ) {
				Linenum++;
				dbprt("obj","\tLinenum - %d",Linenum) ;
			} else if ( typ == R_TEXT ){
				dbprt("obj","\tR_TEXT") ;
			} else if ( typ == R_DATA ) {
				dbprt("obj","\tR_DATA") ;
			} else if ( typ == R_BSS ) {
				dbprt("obj","\tR_BSS") ;
			}
		} else { 
			switch ( typ & 017 ) {
				case A_ABS:
					if ( typ & R_BYT )
						{
						val = getc(object);
						dbprt("obj",
						"\tA_ABS:byte value %x"
						,val) ;
						}
					else
						{
						val = getw(object);
						dbprt("obj",
						"\tA_ABS:word value %x"
						,val) ;
						}
					break;
				case A_TEXT:
					dbprt("obj","\tA_TEXT ") ;
					goto value ;
				case A_DATA:
					dbprt("obj","\tA_DATA ") ;
					goto value ;
				case A_BSS:
					dbprt("obj","\tA_BSS ") ;
					goto value ;
value:
					val = getw(object);
					dbprt("obj","%x,",val) ;
					break;
				case A_UND:
					dbprt("obj","\tA_UND ") ;
					sym = (struct Sym *)getw(object);
					dbprt("obj",",symptr - %x",sym) ;
					evalsym(sym);
					if ( sym->sy_type & A_UEXT )
						{
						val = 0;
						dbprt("obj",",undef or ext,") ;
						}
					else
						{
						val = sym->sy_value;
						dbprt("obj",",newval - %x,",
						val) ;
						}
					break;
				case A_EXPR:
					dbprt("obj","\tA_EXPR,"); 
					expr = (struct Expr *)getw(object);
					dbprt("obj","expptr - %x,",expr) ;
					evalexp(expr,&mval,&mrel);
					val = mval;
					dbprt("obj","expval - %x,",val) ;
					break;
			}
			if ( typ & R_PC_REL ) {
				dbprt("obj",",R_PC_REL") ;
			} else {
				if ( typ & R_BYT ) {
					dbprt("obj",",R_BYT") ;
				}
			}
		}
	}
}
#endif

evalsym(sym)
register struct Sym *sym;
{
	register int tmp;
	int val,type;

	if ( sym->sy_type == A_EXPR ) {
					/* symbol value is in expression form */
		if ( sym->sy_flags & F_BUSY ) {
			error("Reference loop in definition of symbol ",
			       sym->sy_name);
			return (-1);
		} else {		/* not self referencing */
			sym->sy_flags |= F_BUSY;
			tmp = evalexp((struct Expr *)sym->sy_value,&val,&type);
					/* evaluate symbol expression */
			sym->sy_flags &= -F_BUSY - 1;
			if ( tmp == -1 ) 
				{	/* error in evaluation */
				sym->sy_value = 0;
				sym->sy_type = A_ABS;
				return (-1);
			} else {	/* save value and type */
				sym->sy_value = val;
				sym->sy_type = type;
			}
		}
	} else if ( sym->sy_type == A_UND ) {
		if ( (sym->sy_flags & F_LOCAL) == 0 ) {	
					/* not local */
			if( Und_glb ) {	/* treat undefines as gloabl */
				if ( (sym->sy_flags & F_EXTRN) == 0 )	
					/* not external - define as global */
					do_glb(sym);
				sym->sy_type = A_UEXT | (sym->sy_snum << 4);
			} else
				error("Undefined symbol ",sym->sy_name);
		} else {				/* undefined local */
			fprintf(list,"Undefined local symbol #%d, line %d!\n",
			sym->sy_nam.sy_int,Linenum);
			++Errors;
		}
	}
	return (0);
}

evalexp(expr,val,type)
register struct Expr *expr; register int *val,*type;
{
	if ( getval(&(expr->E_a)) == -1 || getval(&(expr->E_b)) == -1 ) {
		if ( Pass == 1 )		/* possible forward reference */
			return (1);		/* ok for first pass */
		else {
			*val = 0;		/* not evaluatable  use zero */
			*type = A_ABS;
			return (-1);		/* return error flag */
		}
	}
	switch ( expr->E_opr ) {
		case '+':
			*val = expr->E_a.e_v.e_int + expr->E_b.e_v.e_int;
			if ( expr->E_a.e_type == A_ABS )
				*type = expr->E_b.e_type;
			else
				if ( expr->E_b.e_type == A_ABS )
					*type = expr->E_a.e_type;
				else {
					error("Illegal relocatable expression","");
					return (-1);
				}
			break;
		case '-':
			*val = expr->E_a.e_v.e_int - expr->E_b.e_v.e_int;
			if ( expr->E_a.e_type == A_ABS || (expr->E_b.e_type & A_UEXT) )
				if ( expr->E_b.e_type == A_ABS )
					*type = expr->E_a.e_type;
				else {
					error("Illegal relocatable expression","");
					return (-1);
				}
			else
				if ( expr->E_b.e_type == A_ABS )
					*type = expr->E_a.e_type;
				else if ( expr->E_b.e_type == expr->E_a.e_type )
					*type = A_ABS;
				else {
					error("Illegal relocatable expression","");
					return (-1);
				}
			break;
		case '*':
			*val = expr->E_a.e_v.e_int * expr->E_b.e_v.e_int;
			*type = A_ABS;
			if ( expr->E_a.e_type != A_ABS || expr->E_b.e_type != A_ABS) {
				error("Illegal relocatable expression","");
				return (-1);
			}
			break;
		case '/':
			*val = expr->E_a.e_v.e_int / expr->E_b.e_v.e_int;
			*type = A_ABS;
			if ( expr->E_a.e_type != A_ABS || expr->E_b.e_type != A_ABS) {
				error("Illegal relocatable expression","");
				return (-1);
			}
			break;
		case '&':
			*val = expr->E_a.e_v.e_int & expr->E_b.e_v.e_int;
			*type = A_ABS;
			if ( expr->E_a.e_type != A_ABS || expr->E_b.e_type != A_ABS) {
				error("Illegal relocatable expression","");
				return (-1);
			}
			break;
		case '^':
			*val = expr->E_a.e_v.e_int % expr->E_b.e_v.e_int;
			*type = A_ABS;
			if ( expr->E_a.e_type != A_ABS || expr->E_b.e_type != A_ABS) {
				error("Illegal relocatable expression","");
				return (-1);
			}
			break;
		case '!':
			*val = expr->E_a.e_v.e_int | expr->E_b.e_v.e_int;
			*type = A_ABS;
			if ( expr->E_a.e_type != A_ABS || expr->E_b.e_type != A_ABS) {
				error("Illegal relocatable expression","");
				return (-1);
			}
			break;
		case RSH:
			*val = expr->E_a.e_v.e_int >> expr->E_b.e_v.e_int;
			*type = A_ABS;
			if ( expr->E_a.e_type != A_ABS || expr->E_b.e_type != A_ABS) {
				error("Illegal relocatable expression","");
				return (-1);
			}
			break;
		case LSH:
			*val = expr->E_a.e_v.e_int << expr->E_b.e_v.e_int;
			*type = A_ABS;
			if ( expr->E_a.e_type != A_ABS || expr->E_b.e_type != A_ABS) {
				error("Illegal relocatable expression","");
				return (-1);
			}
			break;
		case '~':
			*val = ~ expr->E_a.e_v.e_int;
			*type = A_ABS;
			if ( expr->E_a.e_type != A_ABS ) {
				error("Illegal relocatable expression","");
				return (-1);
			}
			break;
	}
	return (0);
}

getval(estk)
register struct estak *estk;
{
	register struct Sym *sym;
	int val,type;

	if ( Pass == 1 ) {
		if ( estk->e_type == A_UND || estk->e_type == A_EXPR )
			return (-1);	/* can't evaluate yet */
	} else if ( estk->e_type == A_UND ) {
		sym = estk->e_v.e_sym;	/* still undefined ? */
		if(evalsym(sym) == -1 )	/* try evaluating symbol again */
			return (-1);	/* not possible to evaluate */
		else {
			estk->e_type = sym->sy_type;
			if ( sym->sy_type & A_UEXT )
					/* if external use zero */
				estk->e_v.e_int = 0;
			else
					/* otherwise use symbol value */
				estk->e_v.e_int = sym->sy_value;
		}
	} else if ( estk->e_type == A_EXPR ) {
		if ( evalexp(estk->e_v.e_expr,&val,&type) == -1 )
			return (-1);	/* un evaluatable expression */
		else {
			estk->e_type = type;
			estk->e_v.e_int = val;
		}
	}
	return (0);
}

dumpsym(out)
FILE *out;
{
	struct nlist entry;
	register struct Sym *sym;
	register int type;

	for ( sym = gl_head ; sym ; sym = sym->sy_glink ) {
#if DEBUG
		dbprt("sym","sym -> ouput global name %s",sym->sy_name) ;
#endif
		strncpy(entry.n_name,sym->sy_name,8);
		evalsym(sym);
		entry.n_value = sym->sy_value;
#if DEBUG
		dbprt("sym",",symbol value %x,type %x\n",
			sym->sy_value,sym->sy_type) ;
#endif
		switch ( sym->sy_type ) {
			case A_TEXT:
				type = N_TEXT | ( sym->sy_flags & F_EXTRN ? N_EXT : 0);
				break;
			case A_DATA:
				type = N_DATA | ( sym->sy_flags & F_EXTRN ? N_EXT : 0);
				break;
			case A_BSS:
				type = N_BSS | ( sym->sy_flags & F_EXTRN ? N_EXT : 0);
				break;
			case A_ABS:
				type = N_ABS | ( sym->sy_flags & F_EXTRN ? N_EXT : 0);
				break;		/* undefines end up here */
			default:
				type = N_UNDF | ( sym->sy_flags & F_EXTRN ? N_EXT : 0);
				break;
		}
		entry.n_type = type;
		fwrite((char *)&entry,sizeof(entry),1,out);
	}
}
