/*
 * Copyright (C) Microsoft Corporation, 1982
 */

/*      Modification History
 *
 *      05/17/82        JGL     M001
 *              - modify getx() into in-line GETX(), ala putc in stdio
 *                This produces a significant performance gain.
 */


#include "y.tab.h"
#include "a86.h"
#include <ctype.h>

char Symbol[MAXSYM];
static char *psops[] = {
	".text",	".data",	".bss",		".byte",	".word",
	".insrt",	".globl",	".comm",	".blkb",	".blkw",
	".even",	".list",	".nlist",	".end",		
	".ascii",	".asciz",	".float",	".double", 	
	".dlptr",	".targref",	".textref",	".dataref",	
	".segrel",	".selfrel",	0 };
static int pstok[] = {
	TEXT,	DATA,	BSS,	DBYTE,	DWORD,
	INSRT,	GLOBL,	COMM,	BLKB,	BLKW,	
	EVEN,	LIST,	NLIST,	END,	ASCII,
	ASCIZ,	FLOAT,	DOUBLE,	DLPTR,	TARGREF, 
	TEXTREF,DATAREF,SEGREL, SELFREL, 0};

#define ALPHA 0
#define DIGIT 1
#define OTHER 2
#define OCT_HEX 3
#define SHIFTER 4
#define CHARCON 5

char chrcls[128] = {
	OTHER,	OTHER,	OTHER,	OTHER,	/* nul	soh	stx	etx	*/
	OTHER,	OTHER,	OTHER,	OTHER,	/* eot	enq	ack	bel	*/
	OTHER,	OTHER,	OTHER,	OTHER,	/* bs	ht	nl	vt	*/
	OTHER,	OTHER,	OTHER,	OTHER,	/* np	cr	so	si	*/
	OTHER,	OTHER,	OTHER,	OTHER,	/* dle	dc1	dc2	dc3	*/
	OTHER,	OTHER,	OTHER,	OTHER,	/* dc4	nak	syn	etb	*/
	OTHER,	OTHER,	OTHER,	OTHER,	/* can	em	sub	esc	*/
	OTHER,	OTHER,	OTHER,	OTHER,	/* fs	gs	rs	us	*/
	OTHER,	OTHER,	OTHER,	OTHER,	/* sp	!	"	#	*/
	OTHER,	OTHER,	OTHER,	CHARCON,/* $	%	&	'	*/
	OTHER,	OTHER,	OTHER,	OTHER,	/* (	)	*	+	*/
	OTHER,	OTHER,	ALPHA,	OCT_HEX,/* ,	-	.	/	*/
	DIGIT,	DIGIT,	DIGIT,	DIGIT,	/* 0	1	2	3	*/
	DIGIT,	DIGIT,	DIGIT,	DIGIT,	/* 4	5	6	7	*/
	DIGIT,	DIGIT,	OTHER,	OTHER,	/* 8	9	:	;	*/
	SHIFTER,OTHER,	SHIFTER,OTHER,	/* <	=	>	?	*/
	OTHER,	ALPHA,	ALPHA,	ALPHA,	/* @	A	B	C	*/
	ALPHA,	ALPHA,	ALPHA,	ALPHA,	/* D	E	F	G	*/
	ALPHA,	ALPHA,	ALPHA,	ALPHA,	/* H	I	J	K	*/
	ALPHA,	ALPHA,	ALPHA,	ALPHA,	/* L	M	N	O	*/
	ALPHA,	ALPHA,	ALPHA,	ALPHA,	/* P	Q	R	S	*/
	ALPHA,	ALPHA,	ALPHA,	ALPHA,	/* T	U	V	W	*/
	ALPHA,	ALPHA,	ALPHA,	OTHER,	/* X	Y	Z	[	*/
	OTHER,	OTHER,	OCT_HEX,ALPHA,	/* \	]	^	_	*/
	OTHER,	ALPHA,	ALPHA,	ALPHA,	/* `	a	b	c	*/
	ALPHA,	ALPHA,	ALPHA,	ALPHA,	/* d	e	f	g	*/
	ALPHA,	ALPHA,	ALPHA,	ALPHA,	/* h	i	j	k	*/
	ALPHA,	ALPHA,	ALPHA,	ALPHA,	/* l	m	n	o	*/
	ALPHA,	ALPHA,	ALPHA,	ALPHA,	/* p	q	r	s	*/
	ALPHA,	ALPHA,	ALPHA,	ALPHA,	/* t	u	v	w	*/
	ALPHA,	ALPHA,	ALPHA,	OTHER,	/* x	y	z	{	*/
	OTHER,	OTHER,	OTHER,	OTHER	/* |	}	~	del	*/
	} ;
#define class(x) chrcls[(x)&127]

yylex()
{
	register int tok,c;
	register struct optab *op;
	struct optab *oplook();
	struct Sym *lookup(),*loc_lk();
	char val[2],i;

	if ( strmod ) {
		if ( (c = GETX()) == delimit )
			tok = DELIMIT;
		else if ( c == '\n' || c == -1 ) {
			tok = yylval.t_int = c;
			readflg = opcflg = 1;
		} else {
			yylval.t_int = c;
			tok = CHARAC;
		}
	} else switch ( class(c = getnxt()) ) {
		case ALPHA:
			if( getsym(c) != ':' && opcflg) {
					/* not a label and need operation */
				if ( (tok = pslook()) != -1 ) {
					/* psuedo operation */
#if DEBUG
					dbprt("tok",
					"tok - pseudo op %s,t=%x,d=%x,b=%x\n",
					Symbol,Text,Data,Bss) ;
#endif
					yylval.t_opcd = 0;
					operflg = 1; 
					/* looking for operand now */
				} else if ( op = oplook(Symbol) ) {
#if DEBUG
					dbprt("tok",
					"tok - opcode  %s,t=%x,d=%x,b=%x\n",
					op->op_text,Text,Data,Bss);
#endif
					tok = optoken(op);
					yylval.t_opcd = op;
					operflg = 1;
					/* looking for operand now */
				} else {
					tok = SYMBOL;
					yylval.t_sym = lookup(Symbol);
#if DEBUG
					dbprt("tok",
					"tok - symbol  %s,t=%x,d=%x,b=%x\n",
					yylval.t_sym->sy_name,Text,Data,Bss) ;
#endif
					operflg = 0;
					/* looking for operator now */
				}
				opcflg = 0;
			} else
				if ( (tok = reglook()) == -1 ) {
					tok = SYMBOL;
					yylval.t_sym = lookup(Symbol);
#if DEBUG
					dbprt("tok",
					"tok - symbol  %s,t=%x,d=%x,b=%x\n",
					yylval.t_sym->sy_name,Text,Data,Bss) ;
#endif
					operflg = 0; 
					/* looking for operator now */
				} else {/* was a register */
					yylval.t_int = tok;
#if DEBUG
					dbprt("tok",
					"tok - register  %s,t=%x,d=%x,b=%x\n",
					reg_nam[tok],Text,Data,Bss) ;
#endif
					tok = reg_tok[tok];
				}
			break;
		case DIGIT:
			yylval.t_int = getnum(c);
			if ( (c = GETX()) == '$' ) {
				yylval.t_sym = loc_lk(yylval);
					/* found a local symbol */
				tok = LOCSYM;
			} else {
				tok = NUMBER;
				if ( c != '.' )
					/* optional decimal point */
					putbak();
			}
			operflg = 0; /* looking for operator now */
			break;
		case OCT_HEX:		/* /hexnumber or ^octalnumber */
			if ( operflg ) {
				yylval.t_int = (c == '/') ? gethex() : getoct();
				tok = NUMBER;
			} else
				yylval.t_int = tok = c;
			operflg ^= 1;	/* toggle operator/operand flag */
			break;
		case SHIFTER:		/* << or >> */
			if ( GETX() == c )
				tok = yylval.t_int = (c == '<') ? LSH : RSH;
			else {
				tok = yylval.t_int = c;
				putbak();
			}
			operflg = 1; /* looking for operand now */
			break;
		case CHARCON:
#if DEBUG
			dbprt("tok","tok - character constant") ;
#endif
			for(i=0;((c=GETX()) != '\'') && i<2;i++)
			    {
			    switch(c)
				{
				case -1:
					yyerror("unexpected end of file") ;
					tok = yylval.t_int = c ;
					return;
				case '\n':
				    yyerror("unexpected newline") ;
				    goto newline;
				case '\\':
				    switch(c=GETX())
					{
					case '\n':
						yyerror("unexpected newline") ;
						goto newline;
					case 'n':
						val[i] = '\n';
						break;
					case 'r':
						val[i] = '\r';
						break;
					case 'b':
						val[i] = '\b';
						break;
					case 't':
						val[i] = '\t';
						break;
					case 'f':
						val[i] = '\f';
						break;
					default:
						val[i] = c;
						break;
					}
				    break;
				default:
				    val[i] = c ;
				    break;
				}
			    }
			if(c != '\'')
			    yyerror("illegal character constant") ;
			if(i==1)
			    yylval.t_int = val[0];
			else
			    yylval.t_int = (val[0] << 8) | val[1];
			tok = NUMBER;
			break;
		case OTHER:
		newline:
			if ( c == '\n' || c == '|' ) {
					/* treat comments like newlines */
				c = '\n';
				readflg = opcflg = 1;
			}
			tok = yylval.t_int = c;
			operflg = 1; /* looking for operand now */
			break;
	}
	return (tok);
}

getnum(c)
register int c;
{
	register int val;

	for ( val = c-'0' ; isdigit(c = GETX()) ; val = val*10 + c-'0')
		;			/* compute till non decimal digit */
	putbak();			/* put back non decimal digit */
	return (val);			/* return computed value */
}

gethex()
{
	register int val,c;

	for ( val = 0 ; ; val = (val<<4) + c )
		if ( isdigit(c = GETX()) )
			c -= '0';
		else if ( c >= 'a' && c <= 'f' )
			c = c - 'a' + 10;
		else if ( c >= 'A' && c <= 'F' )
			c = c - 'A' + 10;
		else 
			break;		/* compute till non hex digit */
	putbak();			/* put back non hex digit */
	return (val);			/* return computed value */
}

getoct()
{
	register int val,c;

	for ( val = 0 ; (c = GETX())>='0' &&  c<='7' ; val = (val<<3) + c-'0' )
		;			/* compute till non octal digit */
	putbak();			/* put back not octal digit */
	return (val);			/* return computed value */
}

getsym(c)
register int c;
{
	register int ty;
	register char *i;

	i = Symbol;			/* start of empty character buffer */
	do {				/* scan till illegal symbol character */
		*(i++) = c;		/* save last scan character */
	} while ( (ty = class(c = GETX())) == ALPHA || ty == DIGIT );
	putbak();			/* put back non symbol character */
	c = getnxt();  			/* find 1st non-blank after symbol */
	putbak();			/* put character back for rescanning */
	*i = '\0';			/* terminate symbol in buffer */
	return(c);			/* return first non-blank after symbol*/
}

pslook()
{
	register char **op;

	if ( *Symbol == '.' )		/* psuedo operation start with . */
		for ( op = psops ; *op ; ++op )
			if ( strcmp(*op,Symbol) == 0 )
				return (pstok[op-psops]);
					/* found a psuedo operation */
	return(-1);			/* not a psuedo operation */
}

reglook()
{
	register char **reg;

	if ( strlen(Symbol) == 2 )	/* all register are two characters */
		for ( reg = reg_nam ; *reg ; reg++ )
			if ( strcmp(*reg,Symbol) == 0 )
				return ( reg-reg_nam );
					/* return index into register descrip */
	return (-1);			/* not a register */
}

getnxt()
{
	register int c;

	while ( (c = GETX()) == ' ' || c == '\t' )
		;			/* scan till not white space character*/
	return (c);			/* return found character */
}
getx()
{
	if ( readflg ) {		/* need new line image */
		if ( endflag ) 
			return (-1);
		else {
			while ( fgets(linebuf,MAXLINE,input) == NULL )
				if ( istak ) {
					fclose (input);
					input = inpstk[--istak];
				} else {
					no_eof = 0;
					strcpy(linebuf,".end\n");
					break;
				}
			readflg = 0;
			lineptr = linebuf;
		}
	}
	return ( *(lineptr++) );
}
putbak()
{
	if ( --lineptr < linebuf ) 
		lineptr = linebuf;
}

yyerror(s) char *s;
{
	fprintf(list,"***ERROR*** %s, line %d\n",s,Linenum);
	if (list != stderr)
		fprintf(stderr,"***ERROR*** %s, line %d\n",s,Linenum);
	++Errors;
}
error(s,x) char *s,*x;
{
	fprintf(list,"***ERROR*** %s%s, line %d\n",s,x,Linenum);
	if (list != stderr)
		fprintf(stderr,"***ERROR*** %s%s, line %d\n",s,x,Linenum);
	++Errors;
}
warn(s,x) char *s,*x;
{
	fprintf(list,"***ERROR*** %s%s, line %d\n",s,x,Linenum);
	if (list != stderr)
		fprintf(stderr,"***ERROR*** %s%s, line %d\n",s,x,Linenum);
	++Errors;
}
boom(s) char *s;
{
	fprintf(stderr,"Fatal assembler error: %s, line %d, after .globl %s\n",
		s, Linenum, last_glb);
	fclose(object);
	unlink(objtemp);
	exit(5);
}

struct Sym *
lookup(nam)
char *nam;
{
	register struct Sym *sym;
	register int hash,i;

	for ( i = 0, hash = 0 ; nam[i] ; i++ )
		 hash += nam[i] * (i+1);
	hash %= HASHVAL;		/* compute hash index */

	for ( sym = hashtab[hash] ; sym ; sym = sym->sy_link )
		if ( strcmp(nam,sym->sy_name) == 0 )
			{
#if DEBUG
			dbprt("sym","old symbol %s,type %d,value %x,flags %d\n",
			sym->sy_name,sym->sy_type,sym->sy_value,sym->sy_flags);
#endif
			return(sym);	/* found old entry */
			}

	if ( (sym = (struct Sym *)malloc(i-1+sizeof(struct Sym))) == NULL )
		boom("No symbol table space left!!!");
#if DEBUG
			dbprt("sym","new symbol %x,t=%x,D=%x,b=%x\n",
			sym,Text,Data,Bss) ;
#endif
					/* allocate new entry */
	sym->sy_link = hashtab[hash];	/* link into symbol table */
	hashtab[hash] = sym;
	sym->sy_type = A_UND;		/* undefined for now */
	sym->sy_flags = 0;		/* no location info */
	sym->sy_value = 0;		/* no value */
	strcpy(sym->sy_name,nam);	/* save copy of name */
	if ( Ext_only ||  		/* using only named externals or */
	    (Frm_cmp && (*nam == 'L')))	/* compiler generated name */
		sym->sy_snum = 0;	/* don't output in rel symbol tab */
	else
		{				/* assume global */
		if ( (sym->sy_snum = Globals++) == 0 )
			gl_head = sym;	/* assign global number */
		else 
			gl_tail->sy_glink = sym;
		gl_tail = sym;		/* link into global string */
		sym->sy_glink = 0;
		}
	return(sym);			/* return pointer to symbol entry */
}

struct Sym *
loc_lk(nam)
register int nam;
{
	register struct Sym *sym;
					/* search local symbol table */
	for ( sym = lcl_tab ; sym ; sym = sym->sy_link )
		if ( nam == sym->sy_nam.sy_int )
			return(sym);	/* matching number */

	if ( (sym = (struct Sym *)malloc(sizeof(struct Sym))) == NULL)
		boom("No symbol table space left!!!");
	sym->sy_link = lcl_tab;		/* add new entry */
	lcl_tab = sym;
	sym->sy_type = A_UND;		/* undefined for now */
	sym->sy_flags = F_LOCAL;		/* always local */
	sym->sy_value = 0;		/* no value for now */
	sym->sy_nam.sy_int = nam;		/* save number not name */
	return(sym);			/* return pointer to entry */
}
