/*
 * Copyright (C) Microsoft Corporation, 1982
 */

/*      Modification History
 *
 *      05/17/82        JGL     M001
 *              - modify getx() into in-line GETX(), ala putc in stdio
 *                This produces a significant performance gain.
 *      12/09/82        JGL     M002
 *              - fixed bugs in converting double to single floating
 */

#include "y.tab.h"
#include "a86.h"
#include <signal.h>

#ifdef PDP11
#define SUFFIX 'b'
#else
#define SUFFIX 'o'
#endif

/*
*
initsig is used to catch interrrupts and try to clean up a bit.
*
*/

initsig()
	{
	fclose(object) ;
	unlink(objtemp) ;

	exit(4) ;
	}
/*
*
fpsig is used to catch floating point over/undef flow.
* 
*/

fpsig()
	{
	error("floating point over/under flow- assembly aborted\n") ;
	exit(4) ;
	}


main(argc,argv)
int argc; char **argv;
{
	char *Out_name,*Inp_name,*Lst_name ;
	char *cp ;
	int nc;
#if DEBUG
	extern obj_list() ;
#endif

	hashtab[ '.' % HASHVAL ] = &Dot;
	list = stdout;
	Out_name = 0 ;			/* no output name */
	Inp_name = NULL;
	Listing = -1;
	while ( --argc > 0 )
		if ( *(cp = *(++argv)) == '-' ) {
			while ( *(++cp) ) switch ( *cp ) {
				case 'l':
#if DEBUG
					dbprt("cmd","cmd - listing option\n");
#endif
					Listing = 1; break;
				case 'e':
#if DEBUG
					dbprt("cmd","cmd - externals option\n");
#endif
					Ext_only = 1; break;
#ifdef RELFMTI
				case 'E':
#if DEBUG
					dbprt("cmd","cmd - end option\n");
#endif
					end_out = 1; break;
#endif
				case 'g':
#if DEBUG
					dbprt("cmd","cmd - globals option\n");
#endif
					Und_glb = 1; break;
				case 'o':
					Out_name = *(++argv);
#if DEBUG
					dbprt("cmd","cmd - output option %s\n",
					Out_name);
#endif
					--argc;
					break;
				case 'C':		/* from compiler */
#if DEBUG
					dbprt("cmd","cmd - from comp option\n");
#endif
					Frm_cmp = 1; break;

				case 'M':		/* medium model */
#if DEBUG
					dbprt("cmd","cmd - large med option\n");
#endif
					Med_opt = 1; break;

#ifdef DEBUG
				case 'D':
					dbinit(*(++argv)) ;
						/* turn on debugging flag */
					dbprt("cmd","cmd - debug option %s\n",
					*argv) ;
					--argc;
					break;
#endif
				case 'n':
#if DEBUG
					dbprt("cmd","cmd - not option\n") ;
#endif
					switch ( *(++cp) ) {
						case 'l':
#if DEBUG
							dbprt("cmd",
							"\tlisting\n") ;
#endif
							Listing = -1; break;
						case 'e':
#if DEBUG
							dbprt("cmd",
							"\texternal\n") ;
#endif
							Ext_only = 0; break;
						case 'g':
#if DEBUG
							dbprt("cmd",
							"\tglobal\n") ;
#endif
							Und_glb = 0; break;
						default:
							fprintf(stderr,"Unknown option %c ignored.\n",*cp);
							break;
					}
					break;
				default:
					fprintf(stderr,"Unknown option %c ignored.\n",*cp);
					break;
			}
		} else {
			if ( Inp_name == NULL )
				{
				Inp_name = *argv;
#if DEBUG
				dbprt("cmd","cmd - input file %s\n",*argv) ;
#endif
				}
			else
				fprintf(stderr,"More than one input, %s used.\n",Inp_name);
		}

	if ( Inp_name == NULL ) 
		{
#if DEBUG
		dbprt("cmd","cmd - standard input\n") ;
#endif
		input = stdin;
		}
	else
		if ( (input = fopen(Inp_name,"r")) == NULL ) {
			fprintf(stderr,
			"Cannot open input %s\n",Inp_name);
			exit(4);
		}
	if ( !Frm_cmp && ((*(Inp_name+(nc=strlen(Inp_name))-1) != 's') ||
	     (*(Inp_name+nc-2) != '.')))
		fprintf(stderr,"Warning input file %s does not end in .s\n",
		Inp_name) ;
	if (!Out_name)
		{
		register char *q ;
		Out_name = malloc(strlen(Inp_name)+1) ;
		cp = Inp_name ;
		q = Out_name ;
		while(*cp)
			{		/* strip off directories */
			if (*cp=='/') {cp++; q = Out_name;}
			*q++ = *cp++;
			}
		*q = 0 ;
		*(Out_name + strlen(Out_name) -1) = SUFFIX ;
		}
#if DEBUG
	dbprt("cmd","cmd - output file name %s\n",Out_name) ;
#endif

	if ( Listing > 0 ) {
		Lst_name = malloc(strlen(Out_name)+1) ;
		strcpy(Lst_name,Out_name);
		*(Lst_name + strlen(Lst_name) -1) = 'L';
#if DEBUG
		dbprt("cmd","cmd - listing file name %s\n",Lst_name) ;
#endif
		if ( (list = fopen(Lst_name,"w")) == NULL ) {
			fprintf(stderr,
			"Cannot create listing file %s\n",Lst_name);
			exit(4);
		}
	}

#if RELFMTI
	Rel_name = malloc(strlen(Out_name)+1) ;
	strcpy(Rel_name,Out_name) ;
	cp = Rel_name + strlen(Rel_name) -2 ;
	*cp = 0 ;
				/* relocation identifier */
#if DEBUG
	dbprt("cmd","cmd - relocation name %s\n",Rel_name) ;
#endif
#endif
	if (signal(SIGINT, SIG_IGN) != SIG_IGN)
		signal(SIGINT,initsig) ;	/* catch the interrupts */
	if (signal(SIGHUP, SIG_IGN) != SIG_IGN)
		signal(SIGHUP,initsig) ;	/* catch the hangups */
	signal(SIGFPE,fpsig) ;

	object = fopen(mktemp(objtemp),"w+");
	no_eof = 1;

	Pass = 1; /* mark as first pass */
	yyparse();

	Pass = 2; /* mark as "second" pass */
#if DEBUG
	dbdo("obj",obj_list) ;		/* list intermediate file */
#endif
	fix_up(Out_name); 		/* cleanup and merge output files */

	fclose(object);
	unlink(objtemp);
	if ( Errors ) {
		fprintf(stderr,"%s: %d errors.\n",Inp_name,Errors);
		exit(2);
	} else
		exit(0);
}


/*
 * Handles assignment statements
 *
 * common code, moved from a86.y to here, since it is used in two
 * places in there.  -Hans Jan 18, 1982
 */
assign(lab, val)
	register struct Sym *lab;
	register struct Expr *val;
{
	if ( lab->sy_flags & F_FIXED )
		error("Cannot redefine symbol ",lab->sy_name);
	else if ( lab == &Dot ) {
		if ( eint(val) < dot && 
		    (etyp(val) != Dot.sy_type || 
		     etyp(val) != A_ABS) )
			error("Illegal location counter assignment","");
		else
		{
			lab->sy_value = eint(val);
			while(new_dot != eint(val))
			{
				putc(A_ABS|R_BYT, object) ;
				putc(0,object);
				new_dot++ ;
			}
		}
	} 
	else
	{
		lab->sy_type = etyp(val);
		lab->sy_value = eint(val);
#ifdef RELFMTI
		if(Fix_it)
			lab->sy_flags |= F_FIXED;
#endif
	}
}



outmove(opcd,dirctn,mem,reg)
struct optab *opcd;
int dirctn;
register int reg;
register struct memstk *mem;
{
	register int word;

	word = ( optyp(opcd) == WORD && reg_len[reg] == 16 );
	chkreg(reg,word);
	dirctn <<= 1;
	if ( reg_num[reg] == AX_AL && mem->m_mod == 6 ) {
		/* accum and unindexed mem */
		outbyte(0242 | word ^ dirctn);
		outw(mem->m_disp);
	} else {
		outbyte(0210 | word | dirctn);
		outaddr(mem,reg_num[reg]);
	}
}

outsmov(type,dirctn,mem,reg,seg)
int type,dirctn,reg,seg;
struct memstk *mem;
{
	dirctn <<= 1;
	outbyte(0214 | dirctn);
	if ( type ) {
		chkreg(reg,1);
		outbyte(0300 | (seg<<3) | reg_num[reg]);
	} else
		outaddr(mem,seg);
}

chkreg(reg,wrdflg)
int reg,wrdflg;
{
	if ( wrdflg ) {
		if ( reg_len[reg] != 16 ) 
			if ( reg_num[reg] > 3 )
				error("Word instruction cannot use ",reg_nam[reg]);
	} else {
		if ( reg_len[reg] != 8 )
			if ( reg_num[reg] > 3 )
				error("Byte instruction cannot use ",reg_nam[reg]);
	}
	return;
}

chkfreg(num)
register int num;
{
	if(num < 0 || num > 7)
		error("Illegal floating point stack register index ","") ;
}

outaddr(ad,reg)
register struct memstk *ad;
register int reg;
{
	register int i;

#ifdef DEBUG
	dbprt("trc","trc - outaddr ad->m_mod %x,ad->m_disp %x,reg %d\n",
	ad->m_mod,ad->m_disp,reg) ;
#endif
	outbyte(ad->m_mod | ((reg&7) << 3));
	if ( (i = ad->m_mod&0300) == 0100)
		outb(ad->m_disp);
	else if ( i == 0200 || ad->m_mod == 6 )
		outw(ad->m_disp);
}

outbyte(value)
register int value;
{
	putc(A_ABS | R_BYT, object);
	putc(value,object);
	core[j++] = value & 255;
	new_dot++;
}

outb(exp)
register struct estak *exp;
{
	register int value;

#if RELFMTI
	if(!Rel_lng && (etyp(exp) & (E_BASE|E_OFFS)))
		{
		if(etyp(exp) & E_BASE) 
			putc(R_PBASE , object) ;
		if(etyp(exp) & E_OFFS)
			putc(R_POFFS , object) ;
		if(Med_opt)
			putc(R_TARG,object); 	/* frame is target */
		else
			putc(R_FTEXT,object) ;	/* frame is text*/
		}
#endif
	switch ( etyp(exp) & 0xff) {
		case A_UND:
		case A_EXPR:
#ifdef RELFMTI
			if(Self_rel)
				putc(R_SELFR,object) ;
#endif
			putc(etyp(exp) | R_BYT, object);
			putw(eint(exp) , object);
			value = 0;
			break;
		case A_ABS:
			putc(etyp(exp) | R_BYT, object);
			putc(eint(exp),object);
			value = eint(exp);
			break;
#if RELFMTI
		case A_TEXT:
		case A_DATA:
		case A_BSS:
			if(Self_rel)
				putc(R_SELFR,object) ;
			putc(etyp(exp) | R_BYT, object);
			putw(eint(exp) , object) ;
			value = 0 ;
			break;
#endif
		default:
			error("Relocatable values not allowed","");
			value = 0;
			break;
	}
	core[j++] = value;
	new_dot++;
#ifdef RELFMTI
	Self_rel = 0;
#endif
}

outw(exp)
register struct estak *exp;
{
	union {
		int integ;
		char bytes[2];
	} spam;

#if RELFMTI
	if(!Rel_lng && (etyp(exp) & (E_BASE|E_OFFS)))
		{
		if(etyp(exp) & E_BASE) 
			putc(R_PBASE , object) ;
		if(etyp(exp) & E_OFFS)
			putc(R_POFFS , object) ;
		if(Med_opt)
			putc(R_TARG,object); 	/* frame is target */
		else
			putc(R_FTEXT,object) ;	/* frame is text*/
		}
#endif
	switch ( etyp(exp) & 0xff ) {
		case A_UND:
		case A_EXPR:
#ifdef RELFMTI
			if(Self_rel)
				putc(R_SELFR,object) ;
#endif
			putc(etyp(exp),object);
			putw( eint(exp) , object);
			spam.integ = 0;
			break;
		case A_ABS:
			putc(etyp(exp),object);
			putw( eint(exp) , object);
			spam.integ = eint(exp);
			break;
		default:
#ifdef RELFMTI
			if(Self_rel)
				putc(R_SELFR,object) ;
#endif
			putc(etyp(exp),object);
			putw( eint(exp) , object);
			spam.integ = eint(exp);
			break;
	}
	core[j++] = spam.bytes[0] & 255;
	core[j++] = spam.bytes[1] & 255;
	new_dot += 2;
#ifdef RELFMTI
	Self_rel = 0;
#endif
}

outwrel(exp)
register struct estak *exp;
{
	union {
		int integ;
		char bytes[2];
	} spam;

	new_dot += 2;
	switch ( etyp(exp) & 0xff ) {
		case A_UND:
		case A_EXPR:
#if DEBUG
			dbprt("loc","unkown relative offset\n") ;
#endif
#ifdef RELFMTI
			putc(R_SELFR,object) ;	/* self relative */
			putc(R_POFFS,object) ;	/* offset */
			putc(R_FTEXT,object) ;	/* frame is text segment */
#endif
			putc(etyp(exp) | R_PC_REL , object);
			putw( eint(exp) , object);
			spam.integ = 0;
			break;
		default:
#if DEBUG
			dbprt("loc","known relative value ") ;
#endif
			putc(A_ABS,object);
			if ( (etyp(exp) & 0xff) != Dot.sy_type ) {
#if DEBUG
				dbprt("loc","ERR %d,%d\n",etyp(exp),Dot.sy_type);
#endif
				spam.integ = 0;
				putw(0,object);
				error("Illegal intersegment jump","");
			} else {
				putw( eint(exp)-new_dot , object);
				spam.integ = eint(exp)-new_dot;
#if DEBUG
				dbprt("loc","offset value %x\n",spam.integ) ;
#endif
			}
			break;
	}
	core[j++] = spam.bytes[0] & 255;
	core[j++] = spam.bytes[1] & 255;
#ifdef RELFMTI
	Self_rel = 0;
#endif
}

outbrel(exp)
register struct estak *exp;
{
	register int value;

	new_dot += 1;
	switch ( etyp(exp) & 0xff ) {
		case A_UND:
		case A_EXPR:
#ifdef RELFMTI
			putc(R_SELFR,object) ;	/* self relative */
			putc(R_FTEXT,object) ;	/* frame is text segment */
#endif
			putc(etyp(exp) | R_BYT | R_PC_REL ,object);
			putw( eint(exp) , object);
			value = 0;
			break;
		default:
			if ( (etyp(exp) & 0xff) != Dot.sy_type ) {
#if DEBUG
				dbprt("loc","ERR %d,%d\n",etyp(exp),Dot.sy_type);
#endif
				value = 0;
				error("Illegal intersegment jump","");
			} else {
				value = eint(exp) - new_dot;
				if ( value < -128 || value > 127 )
					error("Value out of range","");
			}
			putc(A_ABS | R_BYT ,object);
			putc( value , object);
			break;
	}
	core[j++] = value;
#ifdef RELFMTI
	Self_rel = 0;
#endif
}

/* handles .ascii and .asciz strings
   note that the trailing 0 for .asciz is done in the
   yacc file.  -Hans, Nov 5, 1981
 */
outstring()
{
	register char delim, c;
	delim = getnxt();
	while ((c=GETX())!=delim && no_eof)
		switch (c) {
		case '\n':
			putbak(c);
			yyerror("newline in string");
			return;
		case '\\':
			outbkslsh();
			break;
		default:
			outbyte (c);
			break;
		}
	if (no_eof == 0) yyerror("EOF in string");
}

/* outputs the various backslash sequences
   note that this handles the case of \x,
   where x may be the delimeter from above, as
   long as its not in n,t,b,r,f or the digits
   -Hans, Nov 5, 1981
 */
outbkslsh()
{
	register char c;
	register int i;
	int j;
	switch (c = GETX()){
	case 'n':
		outbyte ('\n');
		return;
	case 't':
		outbyte ('\t');
		return;
	case 'b':
		outbyte ('\b');
		return;
	case 'r':
		outbyte ('\r');
		return;
	case 'f':
		outbyte ('\f');
		return;
	case '0': case '1': case '2': case '3':
	case '4': case '5': case '6': case '7':
		i = c-'0';
		j = 3;
		/* allow 3 digits, wont overflow, and its a digit */
		while (--j && i < 040 && (c=GETX()) >= '0' && c<='7')
			i = i*8+c-'0';
		/* if we fell out because its not a digit... */
		if (j && (c<'0' || c>'9')) putbak(c);
		outbyte(i&255);
		return;
	default:
		outbyte (c);
		return;
	}
}

/* handles .float and .double pseudos
   -Hans, Jan 11, 1982
 */
double atof();
outfloat(size)
{
	char text[64];
	register char *p, c;
	char *x;
	unsigned *q, i;
	double doub, hdtotd(), hdtots();
	float sing, dtos();

	c = getnxt();
	p=text;
	*p++ = c;
	while ((c=GETX())=='.' ||
		c== 'e' || c == 'E' ||
		c== '+' || c == '-' ||
		(c>='0'&&c<='9')) *p++ = c;
	/* nul terminate the string, and put back the delimeter */
	c = *p;
	*p=0;
	putbak(c);

	/* from here to the end of the procedure will have to
	   be rewritten when we are generating 8087 format;
	   this is just for microsoft format, which is an interim
	   solution.  -Hans, Jan 11, 82 */
	doub = atof(text);
#ifdef	PDP11
#ifdef IEEE
	/* IEEE format is similar to pdp11. The hidden one bit
	   is considered 1. instead of .1.  The short bias is 127 instead 
	   of 128.  For long the exponent field is larger, so the bias is 1023.
	*/
	if (size > 4) {
		doub = hdtotd(doub) ;   /* host double to target double */
		p = &doub;
		outbyte(p[6]&0377);
		outbyte(p[7]&0377);
		outbyte(p[4]&0377);
		outbyte(p[5]&0377);
	} else {
		doub = hdtots(doub) ;   /* host double to target single */
		p = &doub;
	}
	outbyte(p[2]&0377);
	outbyte(p[3]&0377);
	outbyte(p[0]&0377);
	outbyte(p[1]&0377);
#else
	/* pdp 11 format has sign and exponent in first bytes,
	   microsoft format is word order backwards of this.
	*/
	p= &doub;
	if (size>4)
	{
		outbyte(p[6]&0377);
		outbyte(p[7]&0377);
		outbyte(p[4]&0377);
		outbyte(p[5]&0377);
	}
	outbyte(p[2]&0377);
	outbyte(p[3]&0377);
	/* exponent and sign are flipped */
	outbyte( (p[1]&0200) | (p[0]&0177));
	q= &doub;
	outbyte((q[0]>>7)&0377);
#endif
#else
	if (size == 4) {
		sing = dtos(doub);
		p = &sing;
	} else
		p= &doub;
	for (i=0;i<size;i++) outbyte (*p++);

#endif
}




/***    dtos - convert double to single
 *
 *      dtos converts a double precision IEEE floating point number to
 *      a single-precision one.
 *
 *      This routine exists to facilitate cross-compilation, otherwise
 *      we could just code "singvar = doubvar;"
 */

float
dtos(doub)
    double doub;
{
	register short exp;
	register short dwrd3;
	union {
		double dval;
		short dwrd[4];
	} dbl;
	union {
		float  sval;
		short swrd[2];
	} sng;


	dbl.dval = doub;
	dwrd3 = dbl.dwrd[3];
	exp = ((dwrd3>>4) & 0x07ff);                    /* 11 bits of exponent */
	if (exp == 0) {
		sng.swrd[0] = sng.swrd[1] = 0;
		return (sng.sval);
	}
	sng.swrd[1] = (dwrd3 & 0x8000)         |        /* transfer exponent */
		      ((exp - 0x3ff + 0x7f) << 7) |     /* exponent */
		      ((dwrd3 & 0xf)<<3)       |        /* 1st 4 mantissa */
		      ((dbl.dwrd[2]>>13) & 7);          /* next 3 mantissa */
	sng.swrd[0] = (dbl.dwrd[2]<<3) |
		      ((dbl.dwrd[1]>>13) & 7);
	return(sng.sval);
}




do_stat()
{
	register int i,k;

	mem = MMem; 
	exp = estk;
	strmod = 0; /* reset this just in case of error */

	if ( Listing == 1 ) {
		fprintf(list,"%5d    %04x   ",Linenum,dot);
		for ( i=0 ; j && i < 6 ; j-- )
			fprintf(list,"%02x ",core[i++] & 255);
		fprintf(list,"%s\t%s","                  "+i*3,linebuf);
		while ( j ) {
			fputs("                ",list);
			for ( k=6 ; j && k-- ; j-- )
				fprintf(list,"%02x ",core[i++] & 255);
			putc('\n',list);
		}
	} else {
		i = j;
		j = 0; 
	}
#ifdef DEBUG
	dbprt("trc","dostat - newdot %x,dot %x,t=%x,d=%x,b=%x\n",
	new_dot,dot,Text,Data,Bss) ;
#endif
	if ( Dot.sy_type != A_BSS ) {
		for ( i = (new_dot-dot)-i ; i >= 2 ; i -= 2 ) {
			putc(A_ABS,object);
			putw(0,object);
		}
		if ( i ) {
			putc(A_ABS | R_BYT, object);
			putc(0,object);
		}
	}
	putc(R_LINE,object); /* mark start of next line */
	dot = new_dot;
	++Linenum;
}

struct memstk *
mem_com(exp,reg)
register struct estak *exp; int reg;
{
	register int val;
	register struct memstk *m;
	int disp;

	m = mem++;			/* get next memory stack entry */
	m->m_disp = exp;		/* save the displacement */
	if ( reg ) {			/* one or two registers specified */
		if ( (val = r_m[reg]) == -1 )
			error("Invalid index register combination","");
		else {
			if ( c_len == 0 ) { /* no displacement given */
				if ( val == 6 ) {	/* (bp) */
					val |= 0100;	/* mod 01 */
					eint(exp) = 0;	/* 8 bit disp */
					etyp(exp) = A_ABS;
				}
			} else if ( etyp(exp) == A_ABS ) {
				disp = eint(exp);
				if ( disp >= -128 && disp <= 127 )
					val |= 0100;	/* mod 01,8 bit disp */
				else
					val |= 0200;	/* mod 10,16 bit disp */
			} else   /* for relocatable values ignore "*" and "#" */
				{
				val |= 0200;		/* mod 10,16 bit disp */
				}
		}
	} else  				/* no index, just displacement*/ 
		val = 6;				/* mod 00,direct addr */

	m->m_mod = val;
	return (m);
}

do_glb(sym)
register struct Sym *sym;
{
#if DEBUG
	dbprt("sym","do_glb %s,type %d,flags %d\n",sym->sy_name,
	sym->sy_type,sym->sy_flags) ;
#endif
	if ( sym->sy_type != A_UND )
		error("Illegal redeclaration of symbol ",sym->sy_name);
	else {
		sym->sy_flags = F_EXTRN;
		/* remember the last global in case we blew up
		   -Hans May 11, 83 */
		last_glb = sym->sy_name;
#ifndef RELFMTI
		if ( Ext_only || 
		    (Frm_cmp && (*(sym->sy_name) == 'L' ))) {
			if ( (sym->sy_snum = Globals++) == 0 )
				gl_head = sym;
			else 
				gl_tail->sy_glink = sym;
			gl_tail = sym;
			sym->sy_glink = 0;
		}
#endif
	}
}

loc_swap(newval,newtyp,objcode)
int *newval,newtyp,objcode;
{
#ifdef DEBUG
	dbprt("trc","trc - loc_swap, newval %x,newtype %d,objcode %x\n",
	*newval,newtyp,objcode) ;
#endif
	switch ( Dot.sy_type ) {
		case A_DATA:
			Data = new_dot;
			break;
		case A_BSS:
			Bss = new_dot;
			break;
		case A_TEXT:
			Text = new_dot;
			break;
	}
	dot = new_dot = *newval;
	Dot.sy_type = newtyp;
	putc(objcode,object);
	lcl_tab = 0; /* kill local symbols */
}

do_exp(op,opra,oprb)
int op;
register struct estak *opra,*oprb;
{
	struct Expr temp;
	register struct Expr *expr;
	int ret,value,type;

	temp.E_opr = op;		/* build operation tree */
	temp.E_a = *opra;
	temp.E_b = *oprb;
	if ( (ret = evalexp(&temp,&value,&type)) == -1 ) 
		return (-1);		/* not evaluatable */
	else if ( ret == 1 ) {		/* legal unevaluated 1st pass exprs */
		if ( expr = (struct Expr *)malloc(sizeof(struct Expr)) ) 
			{		/* save expression in stack */
			*expr = temp;
			etyp(opra) = A_EXPR;
			opra->e_v.e_expr = expr;
		} else
			boom("No space for expression work area!!");
	} else {			/* evaluatable expression */
		etyp(opra) = type;	/* save result in stack */
		eint(opra) = value;
	}
	return (0);
}

do_sym(sym)
register struct Sym *sym;
{
	if ( sym->sy_type == A_UND ) {
		sym->sy_value = dot;	/* assign current location counter */
		sym->sy_type = Dot.sy_type;
		sym->sy_flags |= F_FIXED;
#if DEBUG
		dbprt("sym","define symbol %s,value %x,type %d,flags %d\n",
		sym->sy_name,sym->sy_value,sym->sy_type,sym->sy_flags) ;
#endif
	} else
		error("Multilply defined symbol ",sym->sy_name);
}
