/*
 * Copyright (C) Microsoft Corporation, 1982
 */

#include "a86.h"
#include "y.tab.h"

char *reg_nam[] = {
	"ax","bx","cx","dx","sp","bp","si","di",
	"al","ah","bl","bh","cl","ch","dl","dh",
	"cs","ds","es","ss","st" };

int reg_tok[] = {
	REG,REG,REG,REG,REG,REG,REG,REG,
	REG,REG,REG,REG,REG,REG,REG,REG,
	SREG,SREG,SREG,SREG,ST };
char reg_num[] = {
	0,	/* "ax" */
	3,	/* "bx" */
	1,	/* "cx" */
	2,	/* "dx" */
	4,	/* "sp" */
	5,	/* "bp" */
	6,	/* "si" */
	7,	/* "di" */
	0,	/* "al" */
	4,	/* "ah" */
	3,	/* "bl" */
	7,	/* "bh" */
	1,	/* "cl" */
	5,	/* "ch" */
	2,	/* "dl" */
	6,	/* "dh" */
	1,	/* "cs" */
	3,	/* "ds" */
	0,	/* "es" */
	2,	/* "ss" */
	8 };	/* "st" */

char reg_len[] = {
	16,16,16,16,16,16,16,16,
	8,8,8,8,8,8,8,8 };

int arop_rm[] = {
	0x00,0x08,0x10,0x18,0x20,0x28,0x30,0x38,0x84 };
int arop_ai[] = {
	0x04,0x0c,0x14,0x1c,0x24,0x2c,0x34,0x3c,0xa8 };

/*
   the opcode table.
   must be in alphabetical order because we are doing a binary search
 */
struct optab op_codes[] = {
	"aaa",		SPECL,	0x37,	0,
	"aad",		SPECL,	0xD5,	0x0A,
	"aam",		SPECL,	0xD6,	0x0A,
	"aas",		SPECL,	0x3F,	0,
	"adc",		AROP,	WORD,	I_ADC,
	"adcb",		AROP,	BYTE,	I_ADC,
	"add",		AROP,	WORD,	I_ADD,
	"addb",		AROP,	BYTE,	I_ADD,
	"and",		AROP,	WORD,	I_AND,
	"andb",		AROP,	BYTE,	I_AND,
	"bc",		BRNS,	0x72,	0,
	"beq",		BRNS,	0x75,	0,
	"bge",		BRNS,	0x7c,	0,
	"bgt",		BRNS,	0x7e,	0,
	"bhi",		BRNS,	0x76,	0,
	"bhis",		BRNS,	0x72,	0,
	"ble",		BRNS,	0x7f,	0,
	"blo",		BRNS,	0x73,	0,
	"blos",		BRNS,	0x77,	0,
	"blt",		BRNS,	0x7d,	0,
	"bnc",		BRNS,	0x73,	0,
	"bne",		BRNS,	0x74,	0,
	"br",		CTR,	I_JUMP,	0,
	"call",		CTR,	I_CALL,	0,
	"calli",	CTRI,	I_CALL,	0,
	"cbw",		SPECL,	0x98,	0,
	"clc",		SPECL,	0xF8,	0,
	"cld",		SPECL,	0xFC,	0,
	"cli",		SPECL,	0xFA,	0,
	"cmc",		SPECL,	0xF5,	0,
	"cmp",		AROP,	WORD,	I_CMP,
	"cmpb",		AROP,	BYTE,	I_CMP,
	"cmps",		SPECL,	0xA7,	0,
	"cmpsb",	SPECL,	0XA6,	0,
	"cwd",		SPECL,	0x99,	0,
	"daa",		SPECL,	0x27,	0,
	"das",		SPECL,	0x2F,	0,
	"dec",		INCDEC,	WORD,	I_DEC ,
	"decb",		INCDEC,	BYTE,	I_DEC ,
	"div",		SNGOP,	WORD,	6,
	"divb",		SNGOP,	BYTE,	6,	
	"f16add",	STKMEM,	0xf6,	0,
	"f16com",	STKMEM,	0xf6,	2,
	"f16comp",	STKMEM,	0xf6,	3,
	"f16div",	STKMEM,	0xf6,	6,
	"f16divr",	STKMEM,	0xf6,	7,
	"f16ld",	STKMEM,	0xf7,	0,
	"f16mul",	STKMEM,	0xf6,	1,
	"f16st",	STKMEM,	0xf7,	2,
	"f16stp",	STKMEM,	0xf7,	3,
	"f16sub",	STKMEM,	0xf6,	4,
	"f16subr",	STKMEM,	0xf6,	5,
	"f2xm1",	FSPECL,	0xf1,	0xf0,
	"f32add",	STKMEM,	0xf2,	0,
	"f32com",	STKMEM,	0xf2,	2,
	"f32comp",	STKMEM,	0xf2,	2,
	"f32div",	STKMEM,	0xf2,	6,
	"f32divr",	STKMEM,	0xf2,	7,
	"f32ld",	STKMEM,	0xf3,	0,
	"f32mul",	STKMEM,	0xf2,	1,
	"f32st",	STKMEM,	0xf3,	2,
	"f32stp",	STKMEM,	0xf3,	3,
	"f32sub",	STKMEM,	0xf2,	4,
	"f32subr",	STKMEM,	0xf2,	5,
	"f64ld",	STKMEM,	0xf7,	5,
	"f64stp",	STKMEM,	0xf7,	6,
	"fabs",		FSPECL,	0xf1,	0xe1,
	"fadd",		ARSTK,	0xf0,	0xc0,
	"faddd",	STKMEM,	0xf4,	0,
	"faddp",	ARSTKP,	0xf6,	0xc0,
	"fadds",	STKMEM,	0xf0,	0,
	"fbld",		STKMEM,	0xf7,	4,
	"fbstp",	STKMEM,	0xf7,	6,
	"fchs",		FSPECL,	0xf1,	0xe0,
	"fclex",	FSPECL,	0xf3,	0xe2,
	"fcom",		STACKO,	0xf0,	0xd0,
	"fcomd",	STKMEM,	0xf4,	2,
	"fcomdp",	STKMEM,	0xf4,	3,
	"fcomp",	STACKO,	0xf0,	0xd8,
	"fcompp",	FSPECL,	0xf6,	0xd9,
	"fcomsp",	STKMEM,	0xf0,	3,
	"fcoms",	STKMEM,	0xf0,	2,
	"fdecstp",	FSPECL,	0xf1,	0xf6,
	"fdisi",	FSPECL,	0xf3,	0xe1,
	"fdiv",		ARSTK,	0xf0,	0xf8,
	"fdivd",	STKMEM,	0xf4,	6,
	"fdivp",	ARSTKP,	0xf6,	0xf8,
	"fdivr",	ARSTK,	0xf0,	0xf0,
	"fdivrd",	STKMEM,	0xf4,	7,
	"fdivrp",	ARSTKP,	0xf6,	0xf0,
	"fdivrs",	STKMEM,	0xf0,	7,
	"fdivs",	STKMEM,	0xf0,	6,
	"feni",		FSPECL,	0xf3,	0xe0,
	"ffree",	STACK,	0xf5,	0xc0,
	"fincstp",	FSPECL,	0xf1,	0xf7,
	"finit",	FSPECL,	0xf3,	0xe3,
	"fld",		STACK,	0xf1,	0xc0,
	"fld1",		FSPECL,	0xf1,	0xe8,
	"fldcw",	STKMEM,	0xf1,	5,
	"fldd",		STKMEM,	0xf5,	0,
	"fldenv",	STKMEM,	0xf1,	4,
	"fldl2e",	FSPECL,	0xf1,	0xea,
	"fldl2t",	FSPECL,	0xf1,	0xe9,
	"fldlg2",	FSPECL,	0xf1,	0xec,
	"fldln2",	FSPECL,	0xf1,	0xed,
	"fldpi",	FSPECL,	0xf1,	0xeb,
	"flds",		STKMEM,	0xf1,	0,
	"fldt",		STKMEM,	0xf3,	5,
	"fldz",		FSPECL,	0xf1,	0xee,
	"fmul",		ARSTK,	0xf0,	0xc8,
	"fmuld",	STKMEM,	0xf4,	1,
	"fmulp",	ARSTKP,	0xf6,	0xc8,
	"fmuls",	STKMEM,	0xf0,	1,
	"fnclex",	NWFSPECL, 0xf3,	0xe2,
	"fndisi",	NWFSPECL, 0xf3,	0xe1,
	"fneni",	NWFSPECL, 0xf3,	0xe0,
	"fninit",	NWFSPECL, 0xf3,	0xe3,
	"fnop",		FSPECL,	0xf1,	0xd0,
	"fnsave",	NWSTKMEM, 0xf5,	6,
	"fnstcw",	NWSTKMEM, 0xf1,	7,
	"fnstenv",	NWSTKMEM, 0xf1,	6,
	"fnstsw",	NWSTKMEM, 0xf5,	7,
	"fpatan",	FSPECL,	0xf1,	0xf3,
	"fprem",	FSPECL,	0xf1,	0xf8,
	"fptan",	FSPECL,	0xf1,	0xf2,
	"frndint",	FSPECL,	0xf1,	0xfc,
	"frstor",	STKMEM,	0xf5,	4,
	"fsave",	STKMEM,	0xf5,	6,
	"fscale",	FSPECL,	0xf1,	0xfd,
	"fseg",		NWFSPECL, 0xcd,	0xf8,
	"fsqrt",	FSPECL,	0xf1,	0xfa,
	"fst",		STACK,	0xf5,	0xd0,
	"fstcw",	STKMEM,	0xf1,	7,
	"fstd",		STKMEM,	0xf5,	2,
	"fstdp",	STKMEM,	0xf5,	3,
	"fstenv",	STKMEM,	0xf1,	6,
	"fstp",		STACK,	0xf5,	0xd8,
	"fstpt",	STKMEM,	0xf3,	7, 
	"fsts",		STKMEM,	0xf1,	2,
	"fstsp",	STKMEM,	0xf1,	3,
	"fstsw",	STKMEM,	0xf5,	7,
	"fsub",		ARSTK,	0xf0,	0xe8,
	"fsubd",	STKMEM,	0xf4,	4,
	"fsubp",	ARSTKP,	0xf6,	0xe8,
	"fsubr",	ARSTK,	0xf0,	0xe0,
	"fsubrd",	STKMEM,	0xf4,	5,
	"fsubrp",	ARSTKP,	0xf6,	0xe0,
	"fsubrs",	STKMEM,	0xf0,	5,
	"fsubs",	STKMEM,	0xf0,	4,
	"ftst",		FSPECL,	0xf1,	0xe4,
	"fwait",	NWFSPECL, 0xcd,	0xf9,
	"fxam",		FSPECL,	0xf1,	0xe5,
	"fxch",		STACKO,	0xf1,	0xc8,
	"fxtract",	FSPECL,	0xf1,	0xf4,
	"fyl2x",	FSPECL,	0xf1,	0xf1,
	"fyl2xp1",	FSPECL,	0xf1,	0xf9,
	"hlt",		SPECL,	0xF4,	0,
	"idiv",		SNGOP,	WORD,	7,
	"idivb",	SNGOP,	BYTE,	7,
	"imul",		SNGOP,	WORD,	5,
	"imulb",	SNGOP,	BYTE,	5,
	"in",		IN,	BYTE,	0,
	"inc",		INCDEC,	WORD,	I_INC ,
	"incb",		INCDEC,	BYTE,	I_INC ,
	"int",		INT,	0,	0,
	"into",		SPECL,	0xCE,	0,
	"inw",		IN,	WORD,	0,
	"iret",		SPECL,	0xCF,	0,
	"j",		JMPS,	0xeb,	0,
	"ja",		JMPS,	0x77,	0,
	"jae",		JMPS,	0x73,	0,
	"jb",		JMPS,	0x72,	0,
	"jbe",		JMPS,	0x76,	0,
	"jc",		JMPS,	0x72,	0,
	"jcxz",		JMPS,	0xe3,	0,
	"je",		JMPS,	0x74,	0,
	"jg",		JMPS,	0x7f,	0,
	"jge",		JMPS,	0x7d,	0,
	"jl",		JMPS,	0x7c,	0,
	"jle",		JMPS,	0x7e,	0,
	"jmp",		CTR,	I_JUMP,	0,
	"jmpi",		CTRI,	I_JUMP,	0,
	"jna",		JMPS,	0x76,	0,
	"jnae",		JMPS,	0x72,	0,
	"jnb",		JMPS,	0x73,	0,
	"jnbe",		JMPS,	0x77,	0,
	"jnc",		JMPS,	0x73,	0,
	"jne",		JMPS,	0x75,	0,
	"jng",		JMPS,	0x7e,	0,
	"jnge",		JMPS,	0x7c,	0,
	"jnl",		JMPS,	0x7d,	0,
	"jnle",		JMPS,	0x7f,	0,
	"jno",		JMPS,	0x71,	0,
	"jnp",		JMPS,	0x7b,	0,
	"jns",		JMPS,	0x79,	0,
	"jnz",		JMPS,	0x75,	0,
	"jo",		JMPS,	0x70,	0,
	"jp",		JMPS,	0x7a,	0,
	"jpe",		JMPS,	0x7a,	0,
	"jpo",		JMPS,	0x7b,	0,
	"js",		JMPS,	0x78,	0,
	"jz",		JMPS,	0x74,	0,
	"lahf",		SPECL,	0x9F,	0,
	"lds",		LSEGS,	0xc5,	0 ,
	"lea",		LSEGS,	0x8d,	0 ,
	"les",		LSEGS,	0xc4,	0 ,
	"lock",		SPECL,	0xF0,	0,
	"lodb",		SPECL,	0xAC,	0,
	"lodw",		SPECL,	0xAD,	0,
	"loop",		JMPS,	0xe2,	0,
	"loope",	JMPS,	0xe1,	0,
	"loopne",	JMPS,	0xe0,	0,
	"loopnz",	JMPS,	0xe0,	0,
	"loopz",	JMPS,	0xe1,	0,
	"mov",		MOV,	WORD,	0,
	"movb",		MOV,	BYTE,	0,
	"movs",		SPECL,	0xA5,	0,
	"movsb",	SPECL,	0XA4,	0,
	"mul",		SNGOP,	WORD,	4,
	"mulb",		SNGOP,	BYTE,	4,
	"neg",		SNGOP,	WORD,	3,
	"negb",		SNGOP,	BYTE,	3,
	"nop",		SPECL,	0x90,	0,
	"not",		SNGOP,	WORD,	2,
	"notb",		SNGOP,	BYTE,	2,
	"or",		AROP,	WORD,	I_OR,
	"orb",		AROP,	BYTE,	I_OR,
	"out",		OUT,	BYTE,	0,
	"outw",		OUT,	WORD,	0,
	"pop",		POP,	0,	0 ,
	"popf",		SPECL,	0x9D,	0,
	"push",		PUSH,	0,	0 ,
	"pushf",	SPECL,	0x9C,	0,
	"rcl",		SHFTS,	WORD,	2 ,
	"rclb",		SHFTS,	BYTE,	2 ,
	"rcr",		SHFTS,	WORD,	3 ,
	"rcrb",		SHFTS,	BYTE,	3 ,
	"rep",		SPECL,	0xf3,	0,
	"repnz",	SPECL,	0xF2,	0,
	"repz",		SPECL,	0xF3,	0,
	"ret",		RET,	0,	0 ,
	"reti",		RETI,	0,	0 ,
	"rol",		SHFTS,	WORD,	0 ,
	"rolb",		SHFTS,	BYTE,	0 ,
	"ror",		SHFTS,	WORD,	1 ,
	"rorb",		SHFTS,	BYTE,	1 ,
	"sahf",		SPECL,	0x9E,	0,
	"sal",		SHFTS,	WORD,	4 ,
	"salb",		SHFTS,	BYTE,	4 ,
	"sar",		SHFTS,	WORD,	7 ,
	"sarb",		SHFTS,	BYTE,	7 ,
	"sbb",		AROP,	WORD,	I_SBB,
	"sbbb",		AROP,	BYTE,	I_SBB,
	"scab",		SPECL,	0xAE,	0,
	"scaw",		SPECL,	0xAF,	0,
	"seg",		SEG,	0,	0,
	"shl",		SHFTS,	WORD,	4 ,
	"shlb",		SHFTS,	BYTE,	4 ,
	"shr",		SHFTS,	WORD,	5 ,
	"shrb",		SHFTS,	BYTE,	5 ,
	"stc",		SPECL,	0xF9,	0,
	"std",		SPECL,	0xFD,	0,
	"sti",		SPECL,	0xFB,	0,
	"stob",		SPECL,	0xAA,	0,
	"stow",		SPECL,	0xAB,	0,
	"sub",		AROP,	WORD,	I_SUB,
	"subb",		AROP,	BYTE,	I_SUB,
	"test",		AROP,	WORD,	I_TEST,
	"testb",	AROP,	BYTE,	I_TEST,
	"wait",		SPECL,	0x9B,	0,
	"xchg",		XCHG,	WORD,	0,
	"xchgb",	XCHG,	BYTE,	0,
	"xlat",		SPECL,	0xD7,	0,
	"xor",		AROP,	WORD,	I_XOR,
	"xorb",		AROP,	BYTE,	I_XOR,
	"\177",		0,	0,	0 };

struct optab *
oplook(Symbol)
	char *Symbol;
{
	register int offset,k;
	register struct optab *mid;
	struct optab *base;

	base = op_codes;
	offset = sizeof(op_codes)/sizeof(struct optab);
	do {				/* binary search for opcode */
		offset /= 2;
		mid = base+offset;
		if ( (k = strcmp( mid->op_text,Symbol)) == 0 )
			return (mid);	/* return found opcode table entry */
		else if ( k < 0 )
			base = mid+1;
	} while (offset);
	return (NULL);
}
