    /****************************************************************
    *                                                               *
    *                   LINKER INITIALIZATION                       *
    *                                                               *
    ****************************************************************/

#include                "minlit.h"      /* Types and constants */
#include                "bndtrn.h"      /* More of the same */
#include                "bndvrt.h"      /* Virtual memory stuff */
#include                "lnkio.h"       /* Linker I/O definitions */
#if OSXENIX
#include                <signal.h>      /* Signal definitions */
#endif

extern                  InitVm();       /* Initialize virtual memory */
extern                  MainFn();       /* The linker control function */
extern PROPTYPE         PropSymLookup();/* Look up symbol by property */
extern                  SaveInput();    /* Save input file name and address */
#if CMDMSDOS
extern                  UpdateFileParts();
					/* Update parts of file name */
#endif

#if LIBXENIX
extern FTYPE            fLoWordFirst;   /* Ranlib low word first flag */
#endif
#if (OXOUT OR OIAPX286) AND CMDMSDOS
extern FTYPE            fMedium;        /* Medium model */
extern FTYPE            fIandD;         /* Separate code and data */
#endif
int                     carg;           /* Argument count */
char                    **varg;         /* Argument vector */
char                    *lnknam;        /* Argv[0] */
extern RBTYPE           rbMacSyms;      /* Virt addr of free symbol space */
extern RBTYPE           rgrhte[];
extern FTYPE            vfPascalDefaults;
extern FTYPE            vfNoDefaultLibrarySearch;
extern FTYPE            fPauseRun;
extern RECTTYPE         vrectData;
extern char             *vob;
extern FTYPE            fUndefinedExterns;

extern WORD             ggrMac;
extern WORD             gsnMac;
extern WORD             iovMac;
extern RBTYPE           vrhte;
extern RBTYPE           mpgsnrprop[];
extern SEGTYPE          mpgsnseg[];
extern SNTYPE           mpggrgsn[];
extern RBTYPE           vrhteCODEClass;
extern RBTYPE           vrhteRootOverlay;

extern WORD             cbRgRleMac;
extern SEGTYPE          segResLast;
extern SNTYPE           gsnStack;
extern WORD             cbStack;        /* Size of stack segment */
extern RECTTYPE         rect;
extern FTYPE            vfLineNos;
extern WORD             vcln;
extern SNTYPE           vgsnLineNosPrev;
extern WORD             cparMinAlloc;
extern WORD             cparMaxAlloc;
extern FTYPE            vfDSAlloc;
extern SEGTYPE          segLast;
extern SEGTYPE          segStart;
extern RATYPE           raStart;
extern RBTYPE           vrhteFile;
extern FTYPE            vfPass1;
extern FTYPE            vfMap;
extern GRTYPE           ggrDGroup;
extern SATYPE           saOrigin;
extern WORD             cErrors;
extern FTYPE            fDrivePass;
extern FTYPE            fRunFileOpen;
RBTYPE                  rhteFirstObject;

    /****************************************************************
    *                                                               *
    *  InitSym:                                                     *
    *                                                               *
    *  This function takes no  arguments and returns no meaningful  *
    *  value.  It initializes the symbol table handler.             *
    *                                                               *
    ****************************************************************/

LOCAL                   InitSym()       /* Initialize symbol table handler */
  {
    rbMacSyms = 2;                      /* Init. (word-aligned for PDP-11) */
    InitArea(AREASYMS,CPNMAXSYMS);      /* Initialize VM for symbols */
    Setw(RHTENIL,rgrhte,IRHTEMAX);      /* Initialize hash table */
  }

    /****************************************************************
    *                                                               *
    *  InitializeWorld:                                             *
    *                                                               *
    *  This function takes no  arguments and returns no meaningful  *
    *  value.   It  sets  up  virtual  memory,  the  symbol  table  *
    *  Handlers, and it initializes segment structures.             *
    *                                                               *
    ****************************************************************/

LOCAL                   InitializeWorld()
  {
    APROPGROUPTYPE      *papropGroup;
    APROPOVERLAYTYPE    *papropOverlay;

    InitVm();                           /* Initialize virtual memory */
    for(ggrMac = 0; ggrMac < GGRMAX; ++ggrMac) mpggrgsn[ggrMac] = 0;
					/* Map all groups to null segment */
    segLast = 0;
    cErrors = 0;                        /* No errors yet */
    vrectData = RECTNIL;
    segStart = SEGNIL;
    segResLast = SEGNIL;
    raStart = 0;
    vrhteCODEClass = 0;
    cbRgRleMac = 0;
    gsnStack = 0;
    iovMac = 0;
    vgsnLineNosPrev = 0;
    vcln = 0;
    gsnMac = 1;
    ggrMac = 1;
    cparMinAlloc = 0;
    cparMaxAlloc = 0xFFFF;
#if (OXOUT OR OIAPX286) AND CMDMSDOS
    fMedium = FALSE;                    /* Assume small model */
    fIandD = FALSE;                     /* Assume common code and data */
#endif
#if LIBXENIX
    fLoWordFirst = FALSE;               /* Assume high word first in ranlib */
#endif
    vfDSAlloc = FALSE;                  /* Assume no DS allocation */
    vfMap = FALSE;                      /* Assume no map file */
    vfNoDefaultLibrarySearch = FALSE;   /* Assume search of default libs */
    vfPass1 = TRUE;                     /* This is pass 1 */
    fUndefinedExterns = FALSE;          /* Assume no undefined externals */
    fDrivePass = FALSE;                 /* Not executing DrivePass() */
    fRunFileOpen = FALSE;               /* No run file yet */
    cbStack = 0;                        /* Assume default stack segment */
    InitSym();                          /* Initialize symbol table handler */
    InitArea(AREARGRLE,CPNMAXRLE);      /* Initialize reloc area handler */
    /* Set up root overlay definition */
    vob = PropSymLookup("\0",ATTROVL,TRUE);
					/* Install nil sym. as root overlay */
    papropOverlay = (APROPOVERLAYTYPE *) vob;
    DEBUGVALUE(vrhte);                  /* Debug info */
    vrhteRootOverlay = vrhte;
    papropOverlay->ao_iov = iovMac++;
    /* Definition of DGROUP */
    vob = PropSymLookup("\006DGROUP",ATTRGRP,TRUE);
    papropGroup = (APROPGROUPTYPE *) vob;
    papropGroup->ag_ggr = ggrMac;
    ggrDGroup = ggrMac++;
  }

    /****************************************************************
    *                                                               *
    *  ProcObject:                                                  *
    *                                                               *
    *  This function takes as its  argument a pointer to a length-  *
    *  prefixed string  containing the name of an object file.  It  *
    *  processes that file name.  It does not  return a meaningful  *
    *  value.                                                       *
    *                                                               *
    ****************************************************************/

ProcObject(psbObj)                      /* Process object file name */
BYTE                    *psbObj;        /* Object file name */
  {
    SBTYPE              sbFile;         /* File name */

    ENTER(ProcObject);                  /* Debug info */
#if DEBUG                               /* If debugging on */
    fprintf(stderr,"  Object ");        /* Message */
    OutSb(stderr,psbObj);               /* File name */
    NEWLINE(stderr);                    /* Newline */
    fflush(stderr);                     /* Force message to screen */
#endif                                  /* End debugging code */
#if CMDMSDOS
    PeelFlags(psbObj);                  /* Process flags, if any */
#endif
    if(psbObj[0])                       /* If object name length not zero */
      {
#if CMDMSDOS
	Move(5,sbDotObj,sbFile);        /* Put extension in sbFile */
	UpdateFileParts(sbFile,psbObj); /* Add extension to name */
#endif
#if CMDXENIX
	Move(B2W(psbObj[0]) + 1,psbObj,sbFile);
					/* Leave name unchanged */
#endif
#if LIBMSDOS
	SaveInput(sbFile,0,FHNIL);      /* Save input file name and address */
#endif
#if LIBXENIX
	SaveInput(sbFile,0L,FHNIL);     /* Save input file name and address */
#endif
	if(rhteFirstObject == RHTENIL) rhteFirstObject = vrhte;
					/* Save hash addr of first entry */
      }
    LEAVE(ProcObject,0xFFFF);           /* Debug info */
  }

#if OSXENIX
    /****************************************************************
    *                                                               *
    *  UserKill:                                                    *
    *                                                               *
    *  Clean up if linker killed by user.                           *
    *                                                               *
    ****************************************************************/

LOCAL                   UserKill()
  {
    Fatal("terminated by user.",1);     /* Die gracefully */
  }
#endif

    /****************************************************************
    *                                                               *
    *  main:                                                        *
    *                                                               *
    ****************************************************************/

main(argc,argv)                         /* The real main routine */
int                     argc;           /* Argument count */
char                    **argv;         /* Argument list */
  {
#if NOT FIXEDSTACK AND NOT OSTOPS20
    InitStack();                        /* Initialize stack */
#endif
    lnknam = argv[0];                   /* Save ptr to linker name */
    varg = &argv[1];                    /* Save ptr to arg vector */
    carg = argc - 1;                    /* Save arg count */
    InitializeWorld();                  /* Initialize the linker */
#if OSXENIX
    signal(SIGHUP,UserKill);            /* Trap hangup signal */
    signal(SIGINT,UserKill);            /* Trap user interrupts */
    signal(SIGTERM,UserKill);           /* Trap software termination */
#endif
    MainFn();                           /* The main control function */
  }
