    /****************************************************************
    *                                                               *
    *                           NEWMAP.C                            *
    *                                                               *
    ****************************************************************/

#include                "minlit.h"      /* Types and constants */
#include                "bndtrn.h"      /* Types and constants */
#include                "bndvrt.h"      /* Virtual memory stuff */
#include                "lnkio.h"       /* Linker I/O definitions */

extern                  OutError();     /* Error message routine */
extern                  Fatal();        /* Fatal error routine */
extern                  EnSyms();       /* Symbol enumeration routine */
extern                  PropSymLookup();/* Symbol lookup routine */

extern FTYPE            rgfDirty[];     /* Used by DirtyBp() macro */
extern WORD             cErrors;        /* Count of non-fatal errors */
extern SEGTYPE          segLast;
extern WORD             gsnMac;
extern RBTYPE           vrhteCODEClass;
extern FTYPE            vfCreated;
extern RBTYPE           mpgsnrprop[];
extern RBTYPE           vrprop;
extern SNTYPE           gsnStack;
extern GRTYPE           ggrDGroup;
extern WORD             vbp;            /* Virtual page pointer */
extern FTYPE            vfDSAlloc;      /* TRUE if DS-allocating */
extern IOVTYPE          mpsegiov[];     /* Map segment no. to overlay no. */
extern WORD             iovMac;         /* Number of overlays */
extern FTYPE            fOverlays;      /* TRUE if there are overlays */
extern SEGTYPE          mpgsnseg[];     /* Map global SEGDEF to segment no. */
extern RATYPE           mpgsndra[];
extern SNTYPE           mpggrgsn[];
extern WORD             csegsAbs;       /* Number of absolute LSEGs */
extern SEGTYPE          segResLast;     /* Last segment in root overlay */
extern WORD             cbStack;        /* Size of stack segment */
extern SATYPE           mpsegsa[];      /* Map segment no. to base value */
extern RATYPE           mpsegraFirst[]; /* Map seg no. to offset of 1st byte */
extern WORD             mpsegcb[];      /* Map seg. no. to seg. size */
extern ALIGNTYPE        mpsegalign[];   /* Map seg. no. to alignment type */
extern FTYPE            mpsegFlags[];   /* Map seg. no. to flags */
extern WORD             bsLst;          /* File handle for map file */
static SEGTYPE          seg;            /* Current seg number */
#if OEXE
extern FTYPE            fLstFileOpen;   /* True if list file is open */
#endif
#if OXOUT OR OIAPX286
extern FTYPE            fIandD;         /* True if separate I & D */
extern FTYPE            fMedium;        /* True if middle model */
#endif
#if CALLOCATION
static WORD             cbCommon = 0;   /* Count of bytes in COMMON */
static SEGTYPE          gsnCommon;
extern RBTYPE           vrhte;
#endif
#if OIAPX286
extern WORD             mpstcb[];       /* Map seg. table no. to seg. size */
extern BYTE             mpstacc[];      /* Map seg. table no. to access code */
extern WORD             stMac;          /* No. of seg. table entries */
extern SATYPE           saData;         /* Virtual base of data segment */
#endif

    /****************************************************************
    *                                                               *
    *  FixPublicSymRaS:                                             *
    *                                                               *
    *  Fix public symbol offset when segments are combined.         *
    *                                                               *
    ****************************************************************/

FixPublicSymRaS(hte,papropName,rhte,rprop,fNewHte)
HTETYPE                 *hte;
APROPNAMETYPE           *papropName;
RBTYPE                  rhte;
RBTYPE                  rprop;
FTYPE                   fNewHte;
  {
    SNTYPE              gsn;

    if(!(gsn = papropName->an_gsn)) return;
    papropName->an_ra += mpgsndra[gsn];
    DirtyBp(vbp);
  }

#if CALLOCATION
    /****************************************************************
    *                                                               *
    *  AllocateCommon:                                              *
    *                                                               *
    *  Allocate space for C common variables.                       *
    *                                                               *
    ****************************************************************/

LOCAL                   AllocateCommon(pahte,papropName,rhte,rprop,fNewHte)
AHTETYPE                *pahte;
APROPNAMETYPE           *papropName;
RBTYPE                  rhte;
RBTYPE                  rprop;
FTYPE                   fNewHte;
  {
    APROPUNDEFTYPE      *papropUndef = (APROPUNDEFTYPE *) papropName;
    register WORD       len;

    ASSERT(papropName->an_attr == ATTRUND);
					/* Should always be undefined */
    if(len = papropUndef->au_len)       /* If symbol not zero length */
      {
	papropName->an_attr = ATTRPNM;  /* Give it the public attribute */
	papropName->an_ra = cbCommon;   /* Assign an offset */
	papropName->an_gsn = gsnCommon; /* Assign to c-common segment */
	if(65535 - cbCommon >= len) cbCommon += len;
					/* Update byte count */
	else Fatal("Common area longer than 65535 bytes.",0x42);
					/* Fatal if too much common */
	DirtyBp(vbp);                   /* Mark virtual page as dirty */
      }
  }
#endif

    /****************************************************************
    *                                                               *
    *  AssignAddresses:                                             *
    *                                                               *
    *  This  function  scans  the  set  of  segments, given  their  *
    *  ordering, and assigns segment registers and addresses.       *
    *                                                               *
    ****************************************************************/

AssignAddresses()                       /* Assign addresses, perhaps */
  {
    SNTYPE              gsn;
    ALIGNTYPE           align;
    SNTYPE              gsn1;
    GRTYPE              ggr;
    RBTYPE              rhteClass;
    APROPSNTYPE         *papropSn;
    SEGTYPE             segTopDGroup;
    SNTYPE              gsnBottomDGroup;
    SEGTYPE             segBottomDGroup;
    SEGTYPE             segOverlay;
    SEGTYPE             gsnOverlay;
    SEGTYPE             segPrev;
    IOVTYPE             iov;
    ALIGNTYPE           alignOverlay;
    SEGTYPE             mpiovsegLast[IOVMAX];
    WORD                cbOverlay;
#if OEXE
    BSTYPE              bsTmp;          /* Used to save bsLst */
#endif
    SATYPE              saMaxDGroup;
#if OXOUT OR OIAPX286
    FTYPE               fDoCode;
#endif
#if CALLOCATION
    RBTYPE              vrhteBssClass;
#endif
#if OIAPX286
    SATYPE              sa;             /* Segment base */
    WORD                incr;           /* Increment to set alignment */
    WORD                mask;           /* Mask to set alignment */
    WORD                stcbPrev;       /* Prev. value of mpstcb[stMac] */

    stMac = 0;                          /* Initialize counter */
#endif
    /* Set up stack allocation */
    segTopDGroup = 0;
#if OEXE
    if(!gsnStack)
      {
	++cErrors;                      /* Increment error count */
	bsTmp = bsLst;                  /* Save file descriptor */
	bsLst = stdout;                 /* Force message to screen */
	for(;;)                         /* Loop to write error message */
	  {
	    OutError(bsTmp,"Warning: no STACK segment\n");
					/* Print message if no stack seg */
	    if(!fLstFileOpen || bsTmp == stdout) break;
					/* Exit loop */
	    bsLst = bsTmp;              /* Restore file descriptor */
	    bsTmp = stdout;             /* Insure loop termination */
	  }
      }
    else                                /* Else if stack segment */
      {
#endif
	papropSn =
	  (APROPSNTYPE *) ObGetVa(AREASYMS,mpgsnrprop[gsnStack],TRUE);
#if OEXE
	if(!cbStack) cbStack = papropSn->as_cbMx;
	else papropSn->as_cbMx = cbStack;
      }
#endif
#if OXOUT OR OIAPX286
    papropSn->as_cbMx = 0;
#endif
    /* If we have overlays, set up an overlay area */
    if(fOverlays)
      {
	papropSn =
	  (APROPSNTYPE *) PropSymLookup("\014OVERLAY_AREA",ATTRPSN,TRUE);
	papropSn->as_tysn =
	  (papropSn->as_tysn & MASKTYSNCOMBINE) | (ALGNPAR << 5);
	if(vfCreated)
	  {
	    papropSn->as_tysn = PARAPUBSEG;
					/* (0x2C) para-aligned public seg */
	    papropSn->as_rCla = vrhteCODEClass;
	    mpgsnrprop[gsnMac] = vrprop;
	    papropSn->as_gsn = gsnMac++;
#if OXOUT OR OIAPX286
	    papropSn->as_fCod = !fIandD;
#endif
	  }
	gsnOverlay = papropSn->as_gsn;
      }
#if CALLOCATION
    /* Get BSS class name pointer */
    papropSn = (APROPSNTYPE *) PropSymLookup("\003BSS",ATTRNIL,TRUE);
    vrhteBssClass = vrhte;
    /* Get segment name for c_common data */
    papropSn = (APROPSNTYPE *) PropSymLookup("\010c_common",ATTRPSN,TRUE);
    papropSn->as_tysn =
      (papropSn->as_tysn & MASKTYSNCOMBINE) | (ALGNPAR << 5);
    /* If it is a new segment, give it some default values */
    if(vfCreated)
      {
	papropSn->as_tysn = PARAPUBSEG; /* (0x2C) para-aligned public seg */
	papropSn->as_rCla = vrhteBssClass;
	mpgsnrprop[gsnMac] = vrprop;
	papropSn->as_gsn = gsnMac++;
	papropSn->as_ggr = ggrDGroup;
#if OXOUT OR OIAPX286
	papropSn->as_fCod = FALSE;      /* Bss is NOT code */
#endif
      }
    gsnCommon = papropSn->as_gsn;
    cbCommon = papropSn->as_cbMx;
    EnSyms(AllocateCommon,ATTRUND);
    papropSn = (APROPSNTYPE *) PropSymLookup("\010c_common",ATTRPSN,FALSE);
    papropSn->as_cbMx = cbCommon;
#endif
    /* Assign the final segment ordering to the gsn's.  Find an
    *  unassigned segment and then assign all others in its class.
    *  Repeat this process for non-code segments.  Include all
    *  overlays.
    */
#if OXOUT OR OIAPX286
    fDoCode = FCODE;                    /* Do text before data */
    for(;;)                             /* Loop to assign addresses */
      {
#endif
	for(gsn = 1; gsn < gsnMac; ++gsn)
	  {                             /* Examine each segment */
	    if(mpgsnseg[gsn] == SEGNIL) /* If gsn unassigned */
	      {
		papropSn =
		  (APROPSNTYPE *) ObGetVa(AREASYMS,mpgsnrprop[gsn],FALSE);
					/* Get symbol table entry */
		if(papropSn->as_tysn == TYSNABS) continue;
					/* Skip absolute LSEGs */
#if OXOUT OR OIAPX286
		if(fIandD && (papropSn->as_fCod & FCODE) != fDoCode) continue;
					/* If separate I & D, skip if not
					*  type we're doing this iteration
					*/
#endif
		mpgsnseg[gsn] = ++segLast;
					/* Assign an ordering number */
#if OXOUT OR OIAPX286
		mpsegFlags[segLast] = papropSn->as_fCod & FCODE;
					/* Set code flag appropriately */
#endif
#if OIAPX286
		if(fIandD && (mpsegFlags[segLast] & FCODE))
		  sa = (stMac << 3) + STBIAS;
					/* Calc. code segment virtual base */
		else sa = saData;       /* Else use data segment base */
		mpsegsa[segLast] = sa;  /* Save it in table */
		mpsegraFirst[segLast] = 0;
					/* First byte in seg. at offset 0 */
		mpstcb[stMac] = papropSn->as_cbMx;
					/* Store initial segment size */
		mpsegcb[segLast] = mpstcb[stMac];
					/* Save length of this piece of grp */
		stcbPrev = 0;           /* No length previously */
		if(!fIandD || !(papropSn->as_fCod & FCODE))
		  mpstacc[stMac] = '\222';
					/* Set access code for DATA (92H) */
		else mpstacc[stMac] = '\232';
					/* Set access code for CODE */
		ggr = papropSn->as_ggr; /* Get global GRPDEF index */
		if(ggr != GRNIL)        /* If segment is a group member */
		  {
		    /* Find all others in same group and assign them */
#endif
#if NOT OIAPX286
		    rhteClass = papropSn->as_rCla;
					/* Get virt. addr. of class hte */
		    /* Find all others in a like class and assign them */
#endif
		    gsn1 = gsn + 1;     /* Examine only following gsns */
		    while(gsn1 < gsnMac)
		      {                 /* Loop to examine other gsns */
			papropSn = (APROPSNTYPE *)
			  ObGetVa(AREASYMS,mpgsnrprop[gsn1],FALSE);
					/* Fetch SEGDEF from virt. mem. */
#if OIAPX286
			if(papropSn->as_ggr == ggr &&
#endif
#if NOT OIAPX286
			if(papropSn->as_rCla == rhteClass &&
#endif
#if OXOUT OR OIAPX286
			  (!fIandD ||
			    (papropSn->as_fCod & FCODE) == fDoCode) &&
#endif
			  papropSn->as_tysn != TYSNABS)
			  {
			    ++segLast;
#if OXOUT OR OIAPX286
			    mpsegFlags[segLast] = papropSn->as_fCod & FCODE;
					/* Set whether text or data */
#endif
			    mpgsnseg[gsn1] = segLast;
					/* Save ordering number */
#if OIAPX286
			    mpsegsa[segLast] = sa;
					/* Set virtual segment base */
			    /* Fix alignment */
			    align = (papropSn->as_tysn >> 5) & 7;
					/* Get alignment */
			    switch(align)
			      {         /* Switch on alignment type */
				case ALGNWRD:
				  incr = 1;
				  mask = 0xFFFE;
				  break;

				case ALGNPAR:
				  incr = 15;
				  mask = 0xFFF0;
				  break;

				case ALGNPAG:
				  incr = 0x3FF;
				  mask = 0xFC00;
				  break;
			      }
			    if(align != ALGNBYT)
			      {         /* If not a byte-aligned segment */
				if(0xFFFF - mpstcb[stMac] < incr)
				  Fatal("Group too large.",0xB0);
					/* Check for overflow */
				mpstcb[stMac] = (mpstcb[stMac] + incr) & mask;
					/* Round size up to next boundary */
			      }
			    mpsegcb[segLast - 1] = mpstcb[stMac] - stcbPrev;
					/* Correct segment size */
			    stcbPrev = mpstcb[stMac];
					/* Save group size */
			    mpsegraFirst[segLast] = mpstcb[stMac];
					/* Save offset of first byte in seg. */
			    if(0xFFFF - mpstcb[stMac] < papropSn->as_cbMx)
			      Fatal("Group too large.",0xB1);
					/* Check for overflow */
			    mpstcb[stMac] += papropSn->as_cbMx;
					/* Update group length */
#endif
			  }
			++gsn1;         /* Increment counter */
		      }
#if OIAPX286
		  }
		if(sa != saData) ++stMac;
					/* Increment counter */
#endif
	      }
	  }
#if OXOUT OR OIAPX286
	if(fDoCode != FCODE || !fIandD) break;
					/* Quit if text and data both done */
	fDoCode = 0;                    /* Do data next */
      }
#endif
    /* We haven't yet assigned absolute segments (it is assumed
    *  they are empty and are used only for addressing purposes),
    *  but now we must assign them somewhere.
    */
    csegsAbs = 0;                       /* Assume there are no absolute segs */
    for(gsn = 1; gsn < gsnMac; ++gsn)   /* Loop to initialize absolute segs */
      {
	if(mpgsnseg[gsn] == SEGNIL)     /* If we have an absolute segment */
	  {
	    ++csegsAbs;                 /* Increment counter */
	    mpgsnseg[gsn] = ++segLast;  /* Assign a segment order number */
	  }
      }
    if(vfDSAlloc)                       /* If doing DS allocation */
      {
	gsnBottomDGroup = gsnMac;       /* Fix the bottom of DGROUP */
	++csegsAbs;                     /* Inc absolute seg counter */
	segBottomDGroup = ++segLast;    /* Bottom segment in DGROUP */
	mpgsnseg[gsnBottomDGroup] = segLast;
					/* Store entry in table */
      }
    /* Initialize addressing tables */
    for(seg = 0; seg <= segLast; ++seg) /* Loop to initialize addr tabs */
      {
#if NOT OIAPX286
	mpsegsa[seg] = 0;               /* Zero the base */
	mpsegraFirst[seg] = 0;          /* Zero offset of 1st byte */
	mpsegcb[seg] = 0;               /* Zero size */
#endif
	mpsegalign[seg] = ALGNNIL;      /* No alignment type yet */
      }
    for(iov = 0; iov < IOVMAX; ++iov) mpiovsegLast[iov] = SEGNIL;
    alignOverlay = ALGNPAR;             /* Overlays are para-aligned */
    segLast -= csegsAbs;                /* Get no. of last non-abs seg */
    /* Find lowest segment in groups */
    for(gsn = 1; gsn < gsnMac; ++gsn)   /* Loop to find lowest segs */
      {
	seg = mpgsnseg[gsn];            /* Get segment number */
	papropSn = (APROPSNTYPE *) ObGetVa(AREASYMS,mpgsnrprop[gsn],TRUE);
					/* Get symbol table entry */
	/* Assign absolute segs their location */
	if(papropSn->as_tysn == TYSNABS) mpsegsa[seg] = papropSn->as_cbMx;
	ggr = papropSn->as_ggr;         /* Get GRPDEF number */
	DEBUGVALUE(ggr);                /* Debug info */
	if(ggr != GRNIL)                /* If segment is group member */
	  {
	    DEBUGVALUE(seg);            /* Debug info */
	    DEBUGVALUE(ggrDGroup);      /* Debug info */
	    if(!mpggrgsn[ggr] || mpgsnseg[mpggrgsn[ggr]] > seg)
	      mpggrgsn[ggr] = gsn;
	    if(ggr == ggrDGroup && seg > segTopDGroup) segTopDGroup = seg;
	  }
	align = papropSn->as_tysn >> 5;
	if((papropSn->as_tysn & MASKTYSNCOMBINE) == TYSNSTACK) align = ALGNPAR;
	if(align > mpsegalign[seg]) mpsegalign[seg] = align;
	if(mpsegiov[seg] != IOVROOT &&
	  mpiovsegLast[mpsegiov[seg]] == SEGNIL &&
	  align > alignOverlay)
	  {
	    mpiovsegLast[mpsegiov[seg]] = SEGNIL + 1;
	    alignOverlay = align;
	  }
	mpgsndra[gsn] = 0;
	mpsegiov[seg] = papropSn->as_iov;
#if NOT OIAPX286
	mpsegcb[seg] = papropSn->as_cbMx;
#endif
      }
    /* Now we assign actual addresses.  The procedure is as follows:
    *  For each code segment
    *  (1) Assign all addresses of the root up to OVERLAY_AREA.
    *  (2) Assign all addresses of the overlays.
    *  (3) Set the start of the segment after OVERLAY_AREA to be
    *      the greatest of all the overlays including the root
    *      OVERLAY_AREA.
    *  (4) Assign the rest of the root segments.
    *  Repeat steps one through four for all remaining segments.
    *
    *  Set limit of part (1): up to OVERLAY_AREA(if there are overlays)
    *  or the end of the segment list.  Do not assign OVERLAY_AREA until
    *  after all the overlays have been taken care of.
    */
    if(fOverlays)                       /* If there are overlays */
      {
	segOverlay = mpgsnseg[gsnOverlay];
					/* Set limit at 1st overlay */
	mpsegalign[segOverlay] = alignOverlay;
      }
    else segOverlay = segLast;          /* Else look at all segments */
#if OXOUT OR OIAPX286
    fDoCode = FCODE;                    /* Do text, then data */
    for(;;)
      {
#endif
	/* "Set the sizes of all of the root up until the OVERLAY_AREA." */
	segPrev = 0;                    /* Nothing to remember yet */
	seg = 1;                        /* Start with 1st seg */
	while(seg <= segOverlay)        /* While beneath limit */
	  {
#if OXOUT OR OIAPX286
	    if(!fIandD || (mpsegFlags[seg] & FCODE) == fDoCode)
	      {
#endif
		if(mpsegiov[seg] == IOVROOT)
		  {
		    SetSizes(segPrev,TRUE);
		    segPrev = seg;      /* Save segment number */
		  }
#if OXOUT OR OIAPX286
	      }
#endif
	    ++seg;                      /* Increment segment number */
	  }
	/* If there are no overlays, then we have assigned all
	*  segments.  Otherwise, the previous segment of the
	*  beginning of the overlays is the OVERLAY_AREA in the
	*  root.
	*/
	if(fOverlays)                   /* If there are overlays */
	  {
	    for(iov = IOVROOT + 1; iov < iovMac; ++iov)
	      mpiovsegLast[iov] = segOverlay;
	    /* Assign addresses to the overlays.  We do not assign the
	    *  rest of the root because we may have to expand the size of
	    *  OVERLAY_AREA to accommodate a large overlay.
	    */
	    cbOverlay = mpsegcb[segOverlay];
					/* Save size of overlay segment??? */
	    mpsegcb[segOverlay] = 0;    /* Zero the size field??? */
	    for(seg = 1; seg <= segLast; ++seg)
	      {
#if OXOUT OR OIAPX286
		if(!fIandD || (mpsegFlags[seg] & FCODE) == fDoCode)
		  {
#endif
		    if(mpsegiov[seg] != IOVROOT)
		      {
			SetSizes(mpiovsegLast[mpsegiov[seg]],TRUE);
			mpiovsegLast[mpsegiov[seg]] = seg;
		      }
#if OXOUT OR OIAPX286
		  }
#endif
	      }
	    mpsegcb[segOverlay] = cbOverlay;
	    /* Determine first segment in root after OVERLAY_AREA */
	    seg = segOverlay + 1;
	    while(seg <= segLast &&
#if OXOUT OR OIAPX286
	      (!fIandD || ((mpsegFlags[seg] & FCODE) == fDoCode)) &&
#endif
	      mpsegiov[seg] != IOVROOT) ++seg;
	    /* "If there is a segment in the root after the overlays,
	    *  then go through all of the overlays as previous segments
	    *  and set its size with the previous one being the last seg
	    *  of each overlay.  We won't initialize the Vm for that
	    *  segment because we won't know the maximum placement until
	    *  afterward."
	    */
	    if(seg <= segLast)
	      {
		for(iov = IOVROOT + 1; iov < iovMac; ++iov)
		  SetSizes(mpiovsegLast[iov],FALSE);
		/* Assign the rest of the root */
		segPrev = segOverlay;
		while(seg <= segLast)
		  {
#if OXOUT OR OIAPX286
		    if(!fIandD || (mpsegFlags[seg] & FCODE) == fDoCode)
		      {
#endif
			if(mpsegiov[seg] == IOVROOT)
			  {
			    SetSizes(segPrev,TRUE);
			    segPrev = seg;
			  }
#if OXOUT OR OIAPX286
		      }
#endif
		    ++seg;
		  }
	      }
	  }
#if OXOUT OR OIAPX286
	if(fDoCode != FCODE || !fIandD) break;
					/* Break if text and data done */
	fDoCode = 0;                    /* Do data next */
      }
#endif
#if NOT OIAPX286
    if(vfDSAlloc)                       /* If doing DS allocation */
      {
	DEBUGVALUE(segTopDGroup);       /* Debug info */
	DEBUGVALUE(mpsegsa[segTopDGroup]);
					/* Debug info */
	DEBUGVALUE(mpsegraFirst[segTopDGroup]);
					/* Debug info */
	DEBUGVALUE(mpsegcb[segTopDGroup]);
					/* Debug info */
	saMaxDGroup = mpsegsa[segTopDGroup] +
	  ((mpsegraFirst[segTopDGroup] + mpsegcb[segTopDGroup] + 0xF) >> 4);
	DEBUGVALUE(saMaxDGroup);        /* Debug info */
	mpggrgsn[ggrDGroup] = gsnBottomDGroup;
	mpsegsa[segBottomDGroup] = (saMaxDGroup - 0x1000) & ~(~0 << WORDLN);
	DEBUGVALUE(mpsegsa[segBottomDGroup]);
					/* Debug info */
	mpgsndra[gsnBottomDGroup] = 0;
      }
#endif
    segResLast = segLast;
    for(gsn = 1; gsn < gsnMac; ++gsn)
      mpgsndra[gsn] += mpsegraFirst[mpgsnseg[gsn]];
    /* Set all absolute segs to the root overlay */
    seg = segLast + 1;
    while(seg < segLast + csegsAbs) mpsegiov[seg++] = IOVROOT;
    /* "Remember those absolute symbols, too !" */
    mpsegiov[0] = IOVROOT;
    EnSyms(FixPublicSymRaS,ATTRPNM);
  }

#if OXOUT
    /****************************************************************
    *                                                               *
    *  MaxError:                                                    *
    *                                                               *
    *  This function  is called to print error  information when a  *
    *  segment has exceeded 64K.                                    *
    *                                                               *
    ****************************************************************/

LOCAL                   MaxError(seg)
SEGTYPE                 seg;
  {
    if(!(mpsegFlags[seg] & FCODE)) Fatal("Cannot handle > 64K data.",0x43);
					/* Too much data */
    if(!fIandD) Fatal("Cannot handle > 64K impure.",0x44);
					/* Too much code */
    if(!fMedium) Fatal("Cannot handle > 64K code in small model.",0x45);
					/* Too much code */
  }
#endif

    /****************************************************************
    *                                                               *
    *  SetSizes:                                                    *
    *                                                               *
    ****************************************************************/

LOCAL                   SetSizes(segLast,fInitVm)
SEGTYPE                 segLast;
FTYPE                   fInitVm;
  {
    ALIGNTYPE           align;
    RATYPE              raFirst;
    RATYPE              raFirstNew;
    SATYPE              saFirst;
    long                cbSeg;          /* No. of bytes of VM for segment */

    ENTER(SetSizes);                    /* Debug info */
#if OIAPX286
    raFirst = mpsegraFirst[seg] & 15;
#endif
#if NOT OIAPX286
    align = mpsegalign[seg];
    /* Get address of end of previous segment */
    raFirst = mpsegraFirst[segLast] + (mpsegcb[segLast] & 0xF);
    saFirst = mpsegsa[segLast] + (mpsegcb[segLast] >> 4);
    if(raFirst >= 0x10)
      {
	++saFirst;
	raFirst &= 0xF;
      }
    if(align == ALGNWRD) raFirstNew = (raFirst + 1) & 0xFFFE;
					/* Round up to word offset */
    else if(align == ALGNPAR) raFirstNew = (raFirst + 0xF) & 0xFFF0;
					/* Round up to paragraph offset */
    else if(align == ALGNPAG)           /* If page-aligned */
      {
	raFirst += ((saFirst & 0xF) << 4);
	saFirst &= 0xFFF0;
	raFirstNew = (raFirst + 0xFF) & 0xFF00;
					/* Round up to page offset */
      }
    else raFirstNew = raFirst;          /* Else use byte alignment */
    /* Pad previous segment */
    raFirst = raFirstNew;
    /* Keep proper 20-bit address (canonical address???) */
    saFirst += raFirst >> 4;
    raFirst &= 0xF;
    /* Assign beginning of this segment */
    if(mpsegsa[seg] < saFirst ||
      (mpsegsa[seg] == saFirst && mpsegraFirst[seg] < raFirst))
      {
	DEBUGVALUE(saFirst);            /* Debug info */
	mpsegsa[seg] = saFirst;         /* ??? */
	mpsegraFirst[seg] = raFirst;    /* ??? */
      }
#if OXOUT
    if(saFirst + ((mpsegcb[seg] + raFirst + 0xF) >> 4) >= 0x1000)
      MaxError(seg);
#endif
#endif
    /* Set up VM for this segment */
    if(fInitVm)                         /* If VM should be initialized */
      {
	cbSeg = (long) raFirst + mpsegcb[seg] + CBVBF - 1;
					/* No. of bytes of VM needed */
	InitArea(AREAFSG + seg,(WORD)(cbSeg >> 9));
					/* Initialize VM area */
      }
    LEAVE(SetSizes,0xFFFF);             /* Debug info */
  }
