    /****************************************************************
    *                                                               *
    *                          NEWOUT.C                             *
    *                                                               *
    ****************************************************************/

#include                "minlit.h"      /* Types and constants */
#include                "bndtrn.h"      /* Types and constants */
#include                "bndrel.h"      /* Relocation definitions */
#include                "xenfmt.h"      /* Executable file header def.s */
#include                "lnkio.h"       /* Linker I/O definitions */

extern char             *ObGetVa();     /* Fetch from virtual memory */
extern                  ReadPvs();      /* Read pages from virtual memory */
extern                  Fatal();        /* Fatal error routine */
#if SYMBOL
extern                  EnSyms();       /* Apply procedure to symbols */

extern FTYPE            fSymbol;        /* Generate symbol output if TRUE */
#endif
#if OXOUT OR OIAPX286
extern FTYPE            fLoWordFirst;   /* Low word first in longs */
#endif
extern FTYPE            fOverlays;      /* True if there are overlays */
extern FTYPE            fRunFileOpen;   /* True if run file open */
extern WORD             cparMinAlloc;
extern WORD             cparMaxAlloc;
extern SEGTYPE          segLast;
extern SEGTYPE          segStart;
extern RATYPE           raStart;
extern WORD             cbRgRleMac;
extern SEGTYPE          mpgsnseg[];
extern FTYPE            mpsegFlags[];
extern SNTYPE           gsnStack;
extern WORD             cbStack;        /* Size of stack segment */
extern SATYPE           mpsegsa[];
extern RATYPE           mpsegraFirst[];
extern WORD             mpsegcb[];
extern IOVTYPE          mpsegiov[];
#if NOT OIAPX286
extern WORD             mpiovcRle[];    /* Count of long relocs per overlay */
#endif
extern BSTYPE           bsRunfile;      /* File handle for runfile */
#if OEXE
static WORD             vchksum;        /* EXE file word checksum */
#endif
#if OSMSDOS OR OSPCDOS OR OSCPM86
extern BYTE             chRunFile;
#endif
#if OXOUT OR OIAPX286
extern FTYPE            fIandD;         /* TRUE if pure and separate I & D */
extern FTYPE            fMedium;        /* Medium model flag */
extern FTYPE            fUndefinedExterns;
					/* TRUE if there are undef. externs */
#endif
#if OIAPX286
extern WORD             stMac;          /* Count of seg table entries */
extern WORD             mpstcb[];       /* Map seg. table no. to size */
extern BYTE             mpstacc[];      /* Map seg. table no. to access code */
extern LFATYPE          mpstlfa[];      /* Map seg. table no. to lfa */
#endif
static BYTE             *psbRun;        /* Run file name */

#define LOWORD(x)       ((WORD) (x) & ~(~0 << WORDLN))
					/* Low-order word of x */
#define HIWORD(x)       ((WORD) ((x) >> WORDLN) & ~(~0 << WORDLN))
					/* High-order word of x */

    /****************************************************************
    *                                                               *
    *  KillRunfile:                                                 *
    *                                                               *
    *  Delete the runfile if it is open.                            *
    *                                                               *
    ****************************************************************/

KillRunfile()
  {
    if(fRunFileOpen)                    /* If the run file is open */
      {
	fclose(bsRunfile);              /* Close it */
	psbRun[B2W(psbRun[0]) + 1] = '\0';
					/* Null-terminate name */
	unlink(&psbRun[1]);             /* Delete it */
	fRunFileOpen = FALSE;           /* Reset flag */
      }
  }


    /****************************************************************
    *                                                               *
    *  ChkRunErr:                                                   *
    *                                                               *
    *  Check for i/o errors on run file stream.                     *
    *                                                               *
    ****************************************************************/

LOCAL                   ChkRunErr()
  {
    register WORD       erc;            /* Error code */

    if(erc = ferror(bsRunfile))         /* If an error occurred */
      {
#if SYMBOL AND OEXE
	if(fRunFileOpen)                /* If runfile open */
	  {
#endif
	    Fatal("I/O error on run file",erc);
					/* Write error message */
#if SYMBOL AND OEXE
	  }
	else Fatal("I/O error on symbol file",erc);
					/* Write error message */
#endif
      }
  }

#if OEXE
    /****************************************************************
    *                                                               *
    *  ChkSum:                                                      *
    *                                                               *
    ****************************************************************/

LOCAL                   ChkSum(cb,pb,fOdd)
WORD                    cb;             /* Byte count */
register BYTE           *pb;            /* Byte pointer */
register FTYPE          fOdd;           /* True if starting address is odd */
  {
    while(cb--)                         /* While bytes remain */
      {
	if(fOdd) vchksum += B2W(*pb++) << BYTELN;
					/* If high order byte of word */
	else vchksum += B2W(*pb++);     /* Else if low order byte */
	fOdd = (!fOdd);                 /* Toggle flag */
      }
    vchksum &= ~(~0 << WORDLN);         /* Use only low 16 bits */
  }
#endif

    /****************************************************************
    *                                                               *
    *  OutArea:                                                     *
    *                                                               *
    *  This function moves an area  from virtual memory to the run  *
    *  file.  It must  handle  output of "correct  section  of rle  *
    *  table."                                                      *
    *                                                               *
    ****************************************************************/

LOCAL                   OutArea(area,cb,raFirst,
#if OEXE
			  fChkSum,
#endif
			  iov)
AREATYPE                area;
WORD                    cb;
RATYPE                  raFirst;
#if OEXE
FTYPE                   fChkSum;
#endif
IOVTYPE                 iov;
  {
    RATYPE              ra;
    char                *pb;
    WORD                cbChunk;
    WORD                pv;
    WORD                cpv;
    BYTE                rgbBufferOut[CBBSIZE];

    ITYPE               irleOffset;
    ITYPE               irleLast;
    ITYPE               ibRle;

    ENTER(OutArea);                     /* Debug info */
#if DEBUG
    fprintf(stderr,"Area %02x  size %04x  raFirst %04x\n",
      B2W(area),cb,raFirst);            /* Debug message */
#endif
    irleOffset = 0;
    pv = raFirst >> 9;
    ra = raFirst;
    raFirst &= CBVBF - 1;
    while(cb)
      {
	if(cb + raFirst > CBBOUT) cbChunk = CBBOUT;
	else cbChunk = cb + raFirst;
	cpv = (cbChunk + CBVBF - 1)/CBVBF;
	DEBUGVALUE(area);               /* Debug info */
	DEBUGVALUE(pv);                 /* Debug info */
	DEBUGVALUE(cpv);                /* Debug info */
	ReadPvs(&rgbBufferOut[irleOffset],area,pv,cpv);
	pb = &rgbBufferOut[raFirst];
	cbChunk -= raFirst;
	raFirst = 0;
#if NOT OIAPX286
	if(area != AREARGRLE)
	  {
#endif
#if OEXE
	    ChkSum(cbChunk,pb,ra & 1);
#endif
	    fwrite(pb,1,cbChunk,bsRunfile);
	    ChkRunErr();
#if NOT OIAPX286
	  }
	else
	  {
	    irleLast = cbChunk + irleOffset;
	    ibRle = 0;
	    while(ibRle <= irleLast - CBRLE)
	      {
		ASSERT(fOverlays || rgbBufferOut[ibRle] == iov);
					/* If no overlays,iov=rgbBufferOut=0 */
		if(rgbBufferOut[ibRle] == iov)
		  {
		    pb = &rgbBufferOut[ibRle + 1];
#if OEXE
		    if(fChkSum) ChkSum(CBRLE - 1,pb,0);
#endif
		    fwrite(pb,1,CBRLE - 1,bsRunfile);
		    ChkRunErr();        /* Check for error */
		    if(!--mpiovcRle[iov])
		      {                 /* If reloc table empty */
			LEAVE(OutArea,0xFFFE);
					/* Debug info */
			return;
		      }
		  }
		ibRle += CBRLE;
	      }
	    irleOffset = 0;
	    while(ibRle < irleLast)
	      rgbBufferOut[irleOffset++] = rgbBufferOut[ibRle++];
	  }
#endif
	ra += cbChunk;
	cb -= cbChunk;
	pv += cpv;
      }
    LEAVE(OutArea,0xFFFF);              /* Debug info */
  }

#if OXOUT
    /****************************************************************
    *                                                               *
    *  Get32BitSize:                                                *
    *                                                               *
    ****************************************************************/

LOCAL                   Get32BitSize(seglo,seghi,psize)
SEGTYPE                 seglo;
SEGTYPE                 seghi;
long                    *psize;
  {
    if(seglo <= seghi)
      {
	*psize = (long) mpsegraFirst[seghi] + ((long) mpsegsa[seghi] << 4) -
	  (long) mpsegraFirst[seglo] - ((long) mpsegsa[seglo] << 4) +
	  (long) mpsegcb[seghi];
      }
    else *psize = 0L;
  }
#endif

#if SYMBOL
#if OEXE
    /****************************************************************
    *                                                               *
    *  DumpSymTab:                                                  *
    *                                                               *
    *  Output the symbol table.                                     *
    *                                                               *
    ****************************************************************/

DumpSymTab(hte,prop,rhte,rprop,fNewHte)
AHTETYPE                *hte;           /* Hash table entry address */
APROPNAMETYPE           *prop;          /* Name address */
RBTYPE                  rhte;           /* Hash tab entry virtual addr */
RBTYPE                  rprop;          /* Overlay prop virt address */
FTYPE                   fNewHte;        /* Unused flag */
  {
    OutWord(bsRunfile,mpsegsa[mpgsnseg[prop->an_gsn]]);
					/* Output symbol seg base */
    OutWord(bsRunfile,prop->an_ra);     /* Output symbol seg offset */
    OutByte(bsRunfile,mpsegFlags[mpgsnseg[prop->an_gsn]]);
					/* Output symbol type flags */
    fwrite(hte->cch,1,B2W(hte->cch[0]) + 1,bsRunfile);
					/* Output symbol name */
    ChkRunErr();                        /* Check for error */
  }
#endif
#if OXOUT OR OIAPX286
    /****************************************************************
    *                                                               *
    *  DumpSymTab:                                                  *
    *                                                               *
    *  Output the symbol table.                                     *
    *                                                               *
    ****************************************************************/

DumpSymTab(hte,prop,rhte,rprop,fNewHte)
AHTETYPE                *hte;           /* Hash table entry address */
APROPNAMETYPE           *prop;          /* Name address */
RBTYPE                  rhte;           /* Hash tab entry virtual addr */
RBTYPE                  rprop;          /* Overlay prop virt address */
FTYPE                   fNewHte;        /* Unused flag */
  {
    register SEGTYPE    seg;            /* Global segment number */
    register WORD       stype;          /* Symbol type */

    seg = mpgsnseg[prop->an_gsn];       /* Get SEGDEF number */
    stype = S_EXTERN;                   /* Initialize */
    if(mpsegFlags[seg] & FNOTEMPTY)     /* If segment not empty */
      {
	if(mpsegFlags[seg] & FCODE) stype |= S_TEXT;
					/* If code, set text field */
	else stype |= S_DATA;           /* Else set data field */
      }
    else                                /* Else if segment empty */
      {
	if(mpsegFlags[seg] & FCODE) stype |= S_TEXT;
					/* If code, set text field */
	else stype |= S_BSS;            /* Else set BSS field */
      }
    OutWord(bsRunfile,stype);           /* Write the type field */
    OutWord(bsRunfile,0);               /* Write the pad field */
    if(fLoWordFirst)                    /* If lo word first in longs */
      {
	OutWord(bsRunfile,prop->an_ra); /* Output the offset */
	OutWord(bsRunfile,mpsegsa[seg]);/* Output the segment base */
      }
    else                                /* Else if hi word first */
      {
	OutWord(bsRunfile,mpsegsa[seg]);/* Output the segment base */
	OutWord(bsRunfile,prop->an_ra); /* Output the offset */
      }
    fwrite(&hte->cch[1],1,B2W(hte->cch[0]),bsRunfile);
					/* Output the public name */
    OutByte(bsRunfile,'\0');            /* Null-terminate the name */
    ChkRunErr();                        /* Check for error */
  }
#endif
#endif

    /****************************************************************
    *                                                               *
    *  OutHeader:                                                   *
    *                                                               *
    *  Thus function  outputs  an  overlay  header.   It  takes as  *
    *  its argument a pointer to a RUNTYPE.   This  function  does  *
    *  not return a meaningful value.                               *
    *                                                               *
    ****************************************************************/

LOCAL                   OutHeader(prun)
RUNTYPE                 *prun;          /* Header to write out */
  {
    ENTER(OutHeader);                   /* Debug info */
#if CPU8086 OR CPU286
    fwrite(prun,sizeof(RUNTYPE),1,bsRunfile);  /* Write the header */
#else
#if OEXE
    vchksum = 0;                        /* Initialize check sum */
    OutWord(bsRunfile,prun->wSignature);/* Output "magic number" */
    vchksum += prun->wSignature;        /* Checksum */
    OutWord(bsRunfile,prun->cbLastp);
    vchksum += prun->cbLastp;           /* Checksum */
    OutWord(bsRunfile,prun->cpnRes);
    vchksum += prun->cpnRes;            /* Checksum */
    OutWord(bsRunfile,prun->irleMax);
    vchksum += prun->irleMax;           /* Checksum */
    OutWord(bsRunfile,prun->cparDirectory);

    vchksum += prun->cparDirectory;     /* Checksum */
    OutWord(bsRunfile,prun->cparMinAlloc);

    vchksum += prun->cparMinAlloc;      /* Checksum */
    OutWord(bsRunfile,prun->cparMaxAlloc);

    vchksum += prun->cparMaxAlloc;      /* Checksum */
    OutWord(bsRunfile,prun->saStack);   /* Output SS value */
    vchksum += prun->saStack;           /* Checksum */
    OutWord(bsRunfile,prun->raStackInit);
					/* Output starting SP value */
    vchksum += prun->raStackInit;       /* Checksum */
    OutWord(bsRunfile,prun->wchksum);   /* Output word checksum */
    vchksum += prun->wchksum;           /* Checksum */
    OutWord(bsRunfile,prun->raStart);   /* Output starting IP value */
    vchksum += prun->raStart;           /* Checksum */
    OutWord(bsRunfile,prun->saStart);   /* Output starting CS value */
    vchksum += prun->saStart;           /* Checksum */
    OutWord(bsRunfile,prun->rbrgrle);   /* Output pointer to reloc table */
    vchksum += prun->rbrgrle;           /* Checksum */
    OutWord(bsRunfile,prun->iovMax);    /* Output count of overlays */
    vchksum += prun->iovMax;            /* Checksum */
#if ODDWORDLN
    vchksum = LO16BITS(vchksum);        /* Use only 16 bits of checksum */
#endif
#endif
#if OXOUT OR OIAPX286
    OutWord(bsRunfile,prun->x_magic);   /* Output the magic number */
    OutWord(bsRunfile,prun->x_ext);     /* Output header extension size */
    if(fLoWordFirst)                    /* If low word first in longs */
      {
	OutWord(bsRunfile,LOWORD(prun->x_text));
					/* Low-order word of text size */
	OutWord(bsRunfile,HIWORD(prun->x_text));
					/* High-order word of text size */
	OutWord(bsRunfile,LOWORD(prun->x_data));
					/* Low-order word of data size */
	OutWord(bsRunfile,HIWORD(prun->x_data));
					/* High-order word of data size */
	OutWord(bsRunfile,LOWORD(prun->x_bss));
					/* Low-order word of bss size */
	OutWord(bsRunfile,HIWORD(prun->x_bss));
					/* High-order word of bss size */
	OutWord(bsRunfile,LOWORD(prun->x_syms));
					/* Low-order word of symbol size */
	OutWord(bsRunfile,HIWORD(prun->x_syms));
					/* High-order word of symbol size */
	OutWord(bsRunfile,LOWORD(prun->x_reloc));
					/* Low-order word of reloc size */
	OutWord(bsRunfile,HIWORD(prun->x_reloc));
					/* High-order word of reloc size */
	OutWord(bsRunfile,LOWORD(prun->x_entry));
					/* Initial CS value */
	OutWord(bsRunfile,HIWORD(prun->x_entry));
					/* Initial IP value */
      }
    else                                /* Else if bullshit ordering */
      {
	OutWord(bsRunfile,HIWORD(prun->x_text));
					/* High-order word of text size */
	OutWord(bsRunfile,LOWORD(prun->x_text));
					/* Low-order word of text size */
	OutWord(bsRunfile,HIWORD(prun->x_data));
					/* High-order word of data size */
	OutWord(bsRunfile,LOWORD(prun->x_data));
					/* Low-order word of data size */
	OutWord(bsRunfile,HIWORD(prun->x_bss));
					/* High-order word of bss size */
	OutWord(bsRunfile,LOWORD(prun->x_bss));
					/* Low-order word of bss size */
	OutWord(bsRunfile,HIWORD(prun->x_syms));
					/* High-order word of symbol size */
	OutWord(bsRunfile,LOWORD(prun->x_syms));
					/* Low-order word of symbol size */
	OutWord(bsRunfile,HIWORD(prun->x_reloc));
					/* High-order word of reloc size */
	OutWord(bsRunfile,LOWORD(prun->x_reloc));
					/* Low-order word of reloc size */
	OutWord(bsRunfile,HIWORD(prun->x_entry));
					/* Initial IP value */
	OutWord(bsRunfile,LOWORD(prun->x_entry));
					/* Initial CS value */
      }
    OutByte(bsRunfile,prun->x_cpu);     /* Output cpu spec */
    OutByte(bsRunfile,prun->x_relsym);  /* Output symbol style */
    OutWord(bsRunfile,prun->x_renv);    /* Output runtime environment */
#endif
#endif
    ChkRunErr();                        /* Check for errors */
    LEAVE(OutHeader,0xFFFF);            /* Debug info */
  }

#if OXOUT OR OIAPX286
    /****************************************************************
    *                                                               *
    *  OutExtHdr:                                                   *
    *                                                               *
    *  Thus function outputs a XENIX extended header.  It takes as  *
    *  its argument a  pointer  to  a  RUNEXTTYPE.   The  function  *
    *  does not return a meaningful value.                          *
    *                                                               *
    ****************************************************************/

LOCAL                   OutExtHdr(pre)
RUNEXTTYPE              *pre;           /* Extended header to write out */
  {
    ENTER(OutExtHdr);                   /* Debug info */
#if CPU8086 OR CPU286
    fwrite(pre,sizeof(RUNEXTTYPE),1,bsRunfile);
					/* Write extended header */
#else
    if(fLoWordFirst)                    /* If low word first in longs */
      {
	OutWord(bsRunfile,LOWORD(pre->xe_trsize));
					/* Low-order word of text size */
	OutWord(bsRunfile,HIWORD(pre->xe_trsize));
					/* High-order word of text size */
	OutWord(bsRunfile,LOWORD(pre->xe_drsize));
					/* Low-order word of data size */
	OutWord(bsRunfile,HIWORD(pre->xe_drsize));
					/* High-order word of data size */
	OutWord(bsRunfile,LOWORD(pre->xe_tbase));
					/* Low-order word of text base */
	OutWord(bsRunfile,HIWORD(pre->xe_tbase));
					/* High-order word of text base */
	OutWord(bsRunfile,LOWORD(pre->xe_dbase));
					/* Low-order word of data base */
	OutWord(bsRunfile,HIWORD(pre->xe_dbase));
					/* High-order word of data base */
	OutWord(bsRunfile,LOWORD(pre->xe_stksize));
					/* Low-order word of stack size */
	OutWord(bsRunfile,HIWORD(pre->xe_stksize));
					/* High-order word of stack size */
      }
    else                                /* Else if bullshit ordering */
      {
	OutWord(bsRunfile,HIWORD(pre->xe_trsize));
					/* High-order word of text size */
	OutWord(bsRunfile,LOWORD(pre->xe_trsize));
					/* Low-order word of text size */
	OutWord(bsRunfile,HIWORD(pre->xe_drsize));
					/* High-order word of data size */
	OutWord(bsRunfile,LOWORD(pre->xe_drsize));
					/* Low-order word of data size */
	OutWord(bsRunfile,HIWORD(pre->xe_tbase));
					/* High-order word of text base */
	OutWord(bsRunfile,LOWORD(pre->xe_tbase));
					/* Low-order word of text base */
	OutWord(bsRunfile,HIWORD(pre->xe_dbase));
					/* High-order word of data base */
	OutWord(bsRunfile,LOWORD(pre->xe_dbase));
					/* Low-order word of data base */
	OutWord(bsRunfile,HIWORD(pre->xe_stksize));
					/* High-order word of stack size */
	OutWord(bsRunfile,LOWORD(pre->xe_stksize));
					/* Low-order word of stack size */
      }
    #endif
    ChkRunErr();                        /* Check for errors */
    LEAVE(OutExtHdr,0xFFFF);            /* Debug info */
  }
#endif

#if NOT OIAPX286
    /****************************************************************
    *                                                               *
    *  OutRunFile:                                                  *
    *                                                               *
    *  This function  outputs an  overlay  to disk.  It is  called  *
    *  from EnSyms() during  an  enumeration  of all overlays.  It  *
    *  takes  the  following inputs: (1) hte, hash table  entry of  *
    *  overlay  name (a real address); (2) prop, overlay  property  *
    *  (a real address); (3) rhte, virtual address of overlay name  *
    *  hash  table entry; (4) rprop, virtual  address of property;  *
    *  and (5) fNewHte, a flag  that is ignored  by this function.  *
    *                                                               *
    ****************************************************************/

OutRunFile(hte,prop,rhte,rprop,fNewHte)
AHTETYPE                *hte;           /* Hash table entry address */
APROPOVERLAYTYPE        *prop;          /* Overlay property address */
RBTYPE                  rhte;           /* Hash tab entry virtual addr */
RBTYPE                  rprop;          /* Overlay prop virt address */
FTYPE                   fNewHte;        /* Unused flag */
  {
    SEGTYPE             seg;
    RUNTYPE             run;
    WORD                cbPadding;
    WORD                cbDirectory;
    WORD                cparDirectory;
    WORD                cb;
    WORD                pn;
    SEGTYPE             segStack;
    WORD                cbRunt;
    long                lfa;
#if SYMBOL
    long                lfaSym;
#endif
    IOVTYPE             iov;
    SBTYPE              sbOverlay;
    RATYPE              ra;
    SATYPE              sa;
#if OXOUT
    FTYPE               fDoCode;
    WORD                cbPadData;
    SEGTYPE             segLastCodeiov; /* Last segment in class CODE */
    SEGTYPE             segFirstDataiov;/* First segment in class DATA */
    RUNEXTTYPE          runext;         /* X.OUT extended header */
#if SYMBOL
    long                cbRle;          /* Updated reloc table size */
#endif
#endif
    SEGTYPE             segIovFirst;    /* First segment in overlay */
    SEGTYPE             segFinaliov;    /* Last seg in overlay to output */
    SEGTYPE             segIovLast;     /* Last segment in overlay */
    WORD                erc;            /* Error code */

    ENTER(OutRunFile);                  /* Debugging info */
    if(!hte->cch[0])                    /* If no entry */
      {
	DEBUGMSG("hte->cch[0] == 0");   /* Message */
	return;                         /* Return */
      }
    iov = prop->ao_iov;
    psbRun = sbOverlay;
    /* Get size of overlay */
    cb = 0;
    pn = 0;
    for(seg = 1; seg <= segLast; ++seg) /* Find lowest seg in overlay */
      {
	if(mpsegiov[seg] == iov) break; /* Break when found */
      }
    if(seg > segLast)                   /* If no overlay to output */
      {
	DEBUGMSG("seg > segLast");      /* Message */
	return;                         /* Return */
      }
    /* Get starting address of lowest segment */
    segIovFirst = seg;                  /* Save the segment number */
    ra = mpsegraFirst[seg];             /* Get the lowest offset??? */
    sa = mpsegsa[seg];                  /* Get the segment value */
    /* Find the last segment in the overlay */
    segIovLast = SEGNIL;
    seg = segLast;                      /* Search from end of list */
    DEBUGVALUE(segLast);                /* Debug info */
    while(seg)                          /* Find highest seg in overlay */
      {
	if(mpsegiov[seg] == iov)        /* If seg. no. equals overlay no. */
	  {
	    if(segIovLast == SEGNIL) segIovLast = seg;
#if OEXE
	    if(!cparMaxAlloc) break;
#endif
	    if((mpsegFlags[seg] & FNOTEMPTY) == FNOTEMPTY) break;
	  }
	--seg;                          /* Decrement counter */
      }
    if(!seg)                            /* If no data in overlay */
      {
	DEBUGMSG("seg == 0");           /* Debug message */
	return;                         /* Return */
      }
    /* "Get size in between" */
    segFinaliov = seg;                  /* Save number of last segment */
    sa = mpsegsa[seg] - sa - 1;
    ra = mpsegraFirst[seg] - ra + 16;
    sa += ra >> 4;
    ra &= 0xF;
    /* Take into account size of last segment */
    if(0xFFFF - ra >= mpsegcb[seg]) ra += mpsegcb[seg];
    else
      {
	ra -= 0xFFFF - mpsegcb[seg];
	--ra;
	sa += 0x1000;                   /* Add in carry */
      }
    /* Normalize */
    sa += ra >> 4;
    ra &= 0xF;
    pn = sa >> 5;
    cb = ((sa << 4) + ra) & MASKRB;
#if OEXE
    run.cbLastp = cb;
#endif
    if(cb)
      {
	cb = 0;
	++pn;
      }
    /* "Empty and overlays don't get created..." */
    if(iov != IOVROOT && !pn)           /* If empty overlay */
      {
	DEBUGMSG("iov != IOVROOT && pn == 0");
					/* Debug message */
	return;                         /* Return */
      }
#if OSMSDOS OR OSPCDOS OR OSCPM86
    if(hte->cch[2] != ':')
      {
	sbOverlay[1] = chRunFile;
	sbOverlay[2] = ':';
	sbOverlay[0] = '\002';
      }
    else
#endif
    sbOverlay[0] = '\0';
    Move(B2W(hte->cch[0]),&hte->cch[1],&sbOverlay[B2W(sbOverlay[0]) + 1]);
    sbOverlay[0] += B2W(hte->cch[0]);
    if(iov != IOVROOT)
      {
#if CMDMSDOS
	UpdateFileParts(sbOverlay,"\004.OVL");
					/* Update file name */
#endif
#if CMDXENIX
	Move(4,".ovl",&sbOverlay[B2W(sbOverlay[0]) + 1]);
					/* Add extension ".ovl" */
	sbOverlay[0] += (BYTE) 4;       /* Update length of name */
#endif
      }
    sbOverlay[B2W(sbOverlay[0]) + 1] = '\0';
					/* Null-terminate name */
    if((bsRunfile = fopen(&sbOverlay[1],WRBIN)) == NULL)
      {                                 /* If create failed */
	if(iov != IOVROOT) Fatal("Cannot open overlay.",1);
					/* Die */
	Fatal("Cannot open run file.",1);
					/* Die */
      }
    SETRAW(bsRunfile);                  /* Binary file */
#if OEXE
    cbRunt = CBRUN;
    run.wSignature = WSIGNATURERUN;
    run.saStart = mpsegsa[segStart];
    run.raStart = raStart;
    if(segFinaliov = segIovLast) run.cparMinAlloc = 0;
    else
      {
	/* sa:ra is end of overlay being output.  Find empty area size */
	sa = mpsegsa[segIovLast] - sa - 1;
	ra = mpsegraFirst[segIovLast] - ra + 0x10;
	/* Add in last segment size */
	if(0xFFFF - ra >= mpsegcb[segIovLast]) ra += mpsegcb[segIovLast];
	else
	  {
	    ra -= 0xFFFF - mpsegcb[segIovLast];
	    --ra;
	    sa += 0x1000;               /* Add in carry */
	  }
	sa += ra >> 4;
	ra &= 0xF;
	run.cparMinAlloc = sa + ((ra + 0xF) >> 4);
      }
    run.cparMaxAlloc = cparMaxAlloc;
    run.irleMax = mpiovcRle[iov];
    segStack = mpgsnseg[gsnStack];
    run.saStack = mpsegsa[segStack];
    run.raStackInit = cbStack + mpsegraFirst[segStack];
    run.wchksum = 0;
    run.rbrgrle = cbRunt;
    run.cpnRes = pn;
    cbDirectory = cbRunt + (mpiovcRle[iov] << 2);
    cbPadding = (0x200 - (cbDirectory & 0x1FF)) & 0x1FF;
    pn = (cbDirectory + 0x1FF) >> 9;
    cparDirectory = pn << SHPNTOPAR;
    run.cparDirectory = cparDirectory;
    run.cpnRes += pn;
    run.iovMax = iov;
    OutHeader(&run);                    /* Output run file header */
    OutArea(AREARGRLE,cbRgRleMac,0,TRUE,iov);
    DEBUGVALUE(iov);                    /* Debug info */
    DEBUGVALUE(mpiovcRle[iov]);         /* Debug info */
    if(mpiovcRle[iov]) Fatal("Bad internal reloc table\n",0x6900);
    while(cbPadding--) OutByte(bsRunfile,'\0');
    ChkRunErr();                        /* Check for write error */
#endif
#if OXOUT
    run.x_magic = X_MAGIC;              /* Install magic number */
    run.x_renv = 0;                     /* Initialize */
    if(cbStack)                         /* If fixed stack */
      {
	run.x_renv |= XE_FS;            /* Set the fixed stack bit */
	run.x_ext = CBRUNEXT;           /* Set size of extended header */
	runext.xe_trsize = 0L;          /* No text reloc table */
	runext.xe_drsize = 0L;          /* No data reloc table */
	runext.xe_tbase = 0L;           /* No text reloc base */
	runext.xe_dbase = 0L;           /* No data reloc base */
	runext.xe_stksize = (long) cbStack;
					/* Set size of fixed stack */
      }
    else run.x_ext = 0;                 /* Else no extended header */
    run.x_cpu = XC_8086;                /* Set CPU code for 8086 */
    run.x_relsym = XR_R86ABS | XR_S86ABS;
					/* Set symbol format for 8086 */
    run.x_renv |= XE_V2;                /* NOT System 3 */
    if(fIandD) run.x_renv |= XE_PURE | XE_SEP;
					/* Set flags for sep. I&D */
    if(fMedium) run.x_renv |= XE_LTEXT; /* Set long text bit if medium model */
    if(!fUndefinedExterns) run.x_renv |= XE_EXEC;
					/* File is executable if no undef.s */
    /* Find last code segment in overlay(assume ordering is TEXT, DATA, BSS) */
    seg = segIovLast;
    while(seg >= segIovFirst)
      {
	if(mpsegiov[seg] == iov &&
	  (!fIandD || (mpsegFlags[seg] & FCODE) == FCODE)) break;
					/* If in overlay and impure or code */
	--seg;
      }
    segLastCodeiov = seg;
    /* Find first data segment in overlay */
    if(!fIandD) seg = segIovLast + 1;   /* No "data" if impure */
    else                                /* Else if separate code and data */
      {
	while(seg <= segIovLast)
	  {
	    if(mpsegiov[seg] == iov &&
	      (mpsegFlags[seg] & FCODE) != FCODE) break;
	    ++seg;
	  }
      }
    segFirstDataiov = seg;
    /* Determine size of TEXT */
    /* Output only non-empty code segments */
    if((seg = segLastCodeiov) > segFinaliov) seg = segFinaliov;
    Get32BitSize(segIovFirst,seg,&(run.x_text));
    cbPadding = 0;
    while(run.x_text & 0xF)
      {
	++cbPadding;
	++run.x_text;
      }
    /* Get size of DATA */
    Get32BitSize(segFirstDataiov,segFinaliov,&(run.x_data));
    cbPadData = 0;
    while(run.x_data & 0xF)
      {
	++cbPadData;
	++run.x_data;
      }
    /* Get size of BSS */
    for(seg = segFinaliov; seg <= segIovLast; ++seg)
      {
	if(mpsegiov[seg] == iov && (mpsegFlags[seg] & FNOTEMPTY) != FNOTEMPTY)
	  break;
      }
    Get32BitSize(seg,segIovLast,&(run.x_bss));
    run.x_syms = 0L;                    /* Assume no symbol table */
    run.x_entry = ((long) raStart << WORDLN) | mpsegsa[segStart];
					/* Load entry pt. (IP-CS) */
    run.x_reloc = (long) mpiovcRle[iov] << 2;
					/* Size of reloc table */
    OutHeader(&run);                    /* Output header */
    if(run.x_ext) OutExtHdr(&runext);   /* Output extended header, if any */
#endif
    ra = mpsegraFirst[segIovFirst];
    sa = mpsegsa[segIovFirst];
#if OXOUT
    fDoCode = FCODE;                    /* Do text first */
    for(;;)                             /* Loop to output text and data */
      {
#endif
	seg = segIovFirst;
	while(seg <= segFinaliov)
	  {
	    if(mpsegiov[seg] == iov)
	      {
#if OXOUT
		if(!fIandD || (mpsegFlags[seg] & FCODE) == fDoCode)
		  {
#endif
		    while(ra != mpsegraFirst[seg] || sa != mpsegsa[seg])
		      {
			OutByte(bsRunfile,'\0');
			if(++ra > 0xF)
			  {
			    ra &= 0xF;
			    ++sa;
			  }
		      }
		    ChkRunErr();        /* Check for error */
		    DEBUGVALUE(seg + AREAFSG);
					/* Degug info */
		    DEBUGVALUE(mpsegcb[seg]);
					/* Degug info */
		    DEBUGVALUE(mpsegraFirst[seg]);
					/* Degug info */
		    DEBUGVALUE(iov);    /* Debug info */
		    OutArea((AREATYPE) seg + AREAFSG,mpsegcb[seg],
		      mpsegraFirst[seg],
#if OEXE
		      TRUE,
#endif
		      iov);
		    sa += mpsegcb[seg] >> 4;
		    ra += mpsegcb[seg] & 0xF;
		    if(ra > 0xF)
		      {
			ra &= 0xF;
			++sa;
		      }
#if OXOUT
		  }
#endif
	      }
	    ++seg;
	  }
#if OXOUT
	while(cbPadding--) OutByte(bsRunfile,'\0');
					/* Pad to paragraph boundary */
	ChkRunErr();                    /* Check for error */
	if(fDoCode != FCODE || !fIandD) break;
					/* If second pass or impure, break */
	ra = 0;
	sa = 0;
	cbPadding = cbPadData;          /* Pad data next time around */
	fDoCode = 0;
      }
#if SYMBOL
    if(fSymbol)                         /* If symbol table requested */
      {
	lfaSym = ftell(bsRunfile);      /* Save lfa before writing symbols */
	EnSyms(DumpSymTab,ATTRPNM);     /* Dump the symbol table */
	lfa = ftell(bsRunfile);         /* Get offset after dumping symbols */
	lfaSym = lfa - lfaSym;          /* Get size of symbol table */
	if(mpiovcRle[iov])              /* If there is reloc info */
	  {
	    if(cbPadding = (WORD)(lfa & 3)) cbPadding = 4 - cbPadding;
					/* Start relocs on 4-byte boundary */
	    cbRle = ((long) mpiovcRle[iov] << 2) + cbPadding;
					/* Get updated reloc table size */
	  }
	fseek(bsRunfile,IBWSYM,0);      /* Seek to header entry */
#if ASSRTON
	ChkRunErr();                    /* Check for error */
#endif
	if(fLoWordFirst)                /* If lo word first in longs */
	  {
	    OutWord(bsRunfile,LOWORD(lfaSym));
					/* Write lo word of symbol length */
	    OutWord(bsRunfile,HIWORD(lfaSym));
					/* Write hi word of symbol length */
	    if(mpiovcRle[iov])          /* If there is reloc info */
	      {
		OutWord(bsRunfile,LOWORD(cbRle));
					/* Write lo word of reloc length */
		OutWord(bsRunfile,HIWORD(cbRle));
					/* Write hi word of reloc length */
	      }
	  }
	else                            /* If hi word first in longs */
	  {
	    OutWord(bsRunfile,HIWORD(lfaSym));
					/* Write hi word of symbol length */
	    OutWord(bsRunfile,LOWORD(lfaSym));
					/* Write lo word of symbol length */
	    if(mpiovcRle[iov])          /* If there is reloc info */
	      {
		OutWord(bsRunfile,HIWORD(cbRle));
					/* Write hi word of reloc length */
		OutWord(bsRunfile,LOWORD(cbRle));
					/* Write lo word of reloc length */
	      }
	  }
	ChkRunErr();                    /* Check for error */
	fseek(bsRunfile,lfa,0);         /* Seek to end of symbols */
#if ASSRTON
	ChkRunErr();                    /* Check for error */
#endif
	if(mpiovcRle[iov])              /* If there is a reloc table */
	  {
	    while(cbPadding--) OutByte(bsRunfile,'\0');
					/* Pad start of reloc table */
	    ChkRunErr();                /* Check for error */
	  }
      }
#endif
    OutArea(AREARGRLE,cbRgRleMac,0,iov);/* Output relocation table */
    if(mpiovcRle[iov]) Fatal("Bad internal reloc table.",0x6900);
					/* Check for inconsistency */
#endif
#if OEXE
#if SYMBOL
    if(iov != IOVROOT)
      {
	lfaSym = ftell(bsRunfile);
	EnSyms(DumpSymTab,ATTRPNM);
	lfa = (long) IBWSYM;
	fseek(bsRunfile,lfa,0);
	OutWord(bsRunfile,LOWORD(lfaSym));
					/* Low-order word of offset */
	OutWord(bsRunfile,HIWORD(lfaSym));
					/* High-order word of offset */
      }
#endif
    vchksum = (~vchksum) & ~(~0 << WORDLN);
					/* Complement lo 16 bits of checksum */
    fseek(bsRunfile,IBWCHKSUM,0);       /* Seek to checksum field in header */
    OutWord(bsRunfile,vchksum);         /* Fill in the checksum */
    ChkRunErr();                        /* Check for error */
#endif
    fclose(bsRunfile);                  /* Close the run file */
    fRunFileOpen = FALSE;               /* Run file is closed */
#if OSXENIX
    if(!fUndefinedExterns) chmod(&sbOverlay[1],0755);
					/* Set protection executable */
#endif
#if OEXE AND SYMBOL
    if(!fSymbol) return;                /* Quit if symbols not requested */
    /* "Generate those symbol tables!!!!" */
    if(iov != IOVROOT) return;
    hte = (AHTETYPE *) ObGetVa(AREASYMS,rhte,FALSE);
					/* Fetch symbol file name from VM */
    Move(B2W(hte->cch[0]) + 1,hte->cch,sbOverlay);
					/* Copy file name */
    Move(4,".sym",&sbOverlay[B2W(sbOverlay[0]) + 1]);
					/* Append ".sym" extension */
    sbOverlay[0] += (BYTE) 4;           /* Increment name length */
    sbOverlay[B2W(sbOverlay[0]) + 1] = '\0';
					/* Null-terminate name */
    if((bsRunfile = fopen(&sbOverlay[1],WRBIN)) == NULL)
      {                                 /* If error opening symbol file */
	Fatal("Cannot open symbol file.",0x52);
					/* Fatal error */
      }
    SETRAW(bsRunfile);                  /* Binary file */
    EnSyms(DumpSymTab,ATTRPNM);         /* Dump symbol table to file */
    fclose(bsRunfile);                  /* Close symbol file */
#endif
    LEAVE(OutRunFile,0xFFFF);           /* Debugging info */
#endif
  }
#if OIAPX286
    /****************************************************************
    *                                                               *
    *  OutRunFile:                                                  *
    *                                                               *
    *  This function  outputs an  overlay  to disk.  It is  called  *
    *  from EnSyms() during  an  enumeration  of all overlays.  It  *
    *  takes  the  following inputs: (1) hte, hash table  entry of  *
    *  overlay  name (a real address); (2) prop, overlay  property  *
    *  (a real address); (3) rhte, virtual address of overlay name  *
    *  hash  table entry; (4) rprop, virtual  address of property;  *
    *  and (5) fNewHte, a flag  that is ignored  by this function.  *
    *                                                               *
    ****************************************************************/

OutRunFile(hte,prop,rhte,rprop,fNewHte)
AHTETYPE                *hte;           /* Hash table entry address */
APROPOVERLAYTYPE        *prop;          /* Overlay property address */
RBTYPE                  rhte;           /* Hash tab entry virtual addr */
RBTYPE                  rprop;          /* Overlay prop virt address */
FTYPE                   fNewHte;        /* Unused flag */
  {
    SEGTYPE             seg;
    RUNTYPE             run;            /* X.OUT header */
    RUNEXTTYPE          runext;         /* X.OUT extended header */
    long                lfa;
#if SYMBOL
    long                lfaSym;
    long                cbRle;          /* Size of reloc area */
#endif
    IOVTYPE             iov;
    SBTYPE              sbOverlay;
    WORD                erc;            /* Error code */
    SEGTYPE             segFullLast;    /* Last non-empty segment */
    WORD                st;             /* Segment table number */
    SATYPE              saPrev;         /* Previous segment's virtual base */
    LFATYPE             stLfa;          /* Segment table lfa */
    WORD                cbPadding;      /* No. of bytes of padding */

    ENTER(OutRunFile);                  /* Debugging info */
    if(!hte->cch[0])                    /* If no entry */
      {
	DEBUGMSG("hte->cch[0] == 0");   /* Debug message */
	return;                         /* Return */
      }
    iov = prop->ao_iov;
    psbRun = sbOverlay;
#if OSMSDOS OR OSPCDOS OR OSCPM86
    if(hte->cch[2] != ':')
      {
	sbOverlay[1] = chRunFile;
	sbOverlay[2] = ':';
	sbOverlay[0] = '\002';
      }
    else
#endif
    sbOverlay[0] = '\0';
    Move(B2W(hte->cch[0]),&hte->cch[1],&sbOverlay[B2W(sbOverlay[0]) + 1]);
    sbOverlay[0] += B2W(hte->cch[0]);
    if(iov != IOVROOT)
      {
#if CMDMSDOS
	UpdateFileParts(sbOverlay,"\004.OVL");
					/* Update file name */
#endif
#if CMDXENIX
	Move(4,".ovl",&sbOverlay[B2W(sbOverlay[0]) + 1]);
					/* Add extension ".ovl" */
	sbOverlay[0] += (BYTE) 4;       /* Update length of name */
#endif
      }
    sbOverlay[B2W(sbOverlay[0]) + 1] = '\0';
					/* Null-terminate name */
    if((bsRunfile = fopen(&sbOverlay[1],WRBIN)) == NULL)
      {                                 /* If create failed */
	if(iov != IOVROOT) Fatal("Cannot open overlay.",1);
					/* Spit blood and die */
	Fatal("Cannot open run file.",1);
					/* Die */
      }
    SETRAW(bsRunfile);                  /* Binary file */
    run.x_magic = X_MAGIC;              /* Install magic number??? */
    run.x_renv = 0;                     /* Initialize */
    if(cbStack)                         /* If fixed stack */
      {
	run.x_renv |= XE_FS;            /* Set the fixed stack bit */
	run.x_ext = CBRUNEXT;           /* Set size of extended header */
	runext.xe_trsize = 0L;          /* No text reloc table */
	runext.xe_drsize = 0L;          /* No data reloc table */
	runext.xe_tbase = 0L;           /* No text reloc base */
	runext.xe_dbase = 0L;           /* No data reloc base */
	runext.xe_stksize = (long) cbStack;
					/* Set size of fixed stack */
      }
    else run.x_ext = 0;                 /* Else no extended header */
    run.x_cpu = XC_286;                 /* 286 CPU code */
    run.x_relsym = XR_R286ABS | XR_S86ABS;
					/* Set symbol style for 286 */
    run.x_renv |= XE_V2;                /* NOT System 3 */
    if(fIandD) run.x_renv |= XE_PURE | XE_SEP;
					/* Set bits for sep. code and data */
    if(fMedium) run.x_renv |= XE_LTEXT;  /* Set long text bit if medium model */
    if(!fUndefinedExterns) run.x_renv |= XE_EXEC;
					/* File is executable if no undef.s */
    run.x_text = 0L;                    /* Initialize */
    run.x_data = 0L;                    /* Initialize */
    run.x_bss = 0L;                     /* Initialize */
    segFullLast = 0;                    /* Initialize */
    seg = segLast;                      /* Start at end and work back */
    while(seg >= 1)                     /* While not at the beginning */
      {
	if(mpsegiov[seg] == iov)        /* If member of this overlay */
	  {
	    if((!fIandD || !(mpsegFlags[seg] & FCODE)) &&
	      !(mpsegFlags[seg] & FNOTEMPTY))
	      {                         /* If BSS segment */
		run.x_bss += (long) mpsegcb[seg];
					/* Add length of segment */
	      }
	    else                        /* Else if not BSS */
	      {
		segFullLast = seg;      /* Save segment number */
		break;                  /* Exit the loop */
	      }
	  }
	--seg;                          /* Decrement index */
      }
    stLfa = (long) CBRUN + run.x_ext;   /* Init. as header size plus ext. */
    for(seg = 1; seg <= segFullLast; ++seg)
      {                                 /* Loop to get size of TEXT and DATA */
	if(mpsegiov[seg] == iov)        /* If member of this overlay */
	  {
	    stLfa += (long) mpsegcb[seg];
					/* Add segment size to lfa */
	    if(!fIandD || (mpsegFlags[seg] & FCODE))
	      {                         /* If code segment */
		run.x_text += (long) mpsegcb[seg];
					/* Accumulate size of code */
	      }
	    else run.x_data += (long) mpsegcb[seg];
					/* Accumulate size of data */
	  }
      }
    if(run.x_data + run.x_bss + (long) cbStack > 0x10000L)
      Fatal("DATA + STACK + BSS exceeds 64K.",0x8F);
					/* Check for overflow */
    run.x_syms = 0L;                    /* Assume no symbol table */
    run.x_entry = ((long) raStart << WORDLN) | mpsegsa[segStart];
					/* Load entry pt. (IP-CS) */
    run.x_reloc = 0L;                   /* Will fix later */
    OutHeader(&run);                    /* Output the header */
    if(run.x_ext) OutExtHdr(&runext);   /* Output the extended header */
    st = 0;                             /* Initialize index */
    saPrev = 0;                         /* No previous value */
    for(seg = 1; seg <= segFullLast; ++seg)
      {                                 /* Loop thru segments */
	if(mpsegiov[seg] == iov)        /* If segment member of this overlay */
	  {
	    if(mpsegsa[seg] != saPrev)  /* If we have a new base value */
	      {
		saPrev = mpsegsa[seg];  /* Save it */
		mpstlfa[st++] = ftell(bsRunfile);
					/* Get current file position */
		ASSERT(st <= stMac);    /* Index should never exceed max */
	      }
	    OutArea(AREAFSG + seg,mpsegcb[seg],mpsegraFirst[seg] & 15,iov);
					/* Output the segment */
	  }
      }
#if SYMBOL
    if(fSymbol)                         /* If symbol table requested */
      {
	lfaSym = ftell(bsRunfile);      /* Save lfa before writing symbols */
	EnSyms(DumpSymTab,ATTRPNM);     /* Dump the symbol table */
	lfa = ftell(bsRunfile);         /* Get offset after dumping symbols */
	lfaSym = lfa - lfaSym;          /* Get size of symbol table */
	fseek(bsRunfile,IBWSYM,0);      /* Seek to header entry */
	ChkRunErr();                    /* Check for error */
	if(fLoWordFirst)                /* If lo word first in longs */
	  {
	    OutWord(bsRunfile,LOWORD(lfaSym));
					/* Write lo word of symbol length */
	    OutWord(bsRunfile,HIWORD(lfaSym));
					/* Write hi word of symbol length */
	  }
	else                            /* If hi word first in longs */
	  {
	    OutWord(bsRunfile,HIWORD(lfaSym));
					/* Write hi word of symbol length */
	    OutWord(bsRunfile,LOWORD(lfaSym));
					/* Write lo word of symbol length */
	  }
	ChkRunErr();                    /* Check for error */
	fseek(bsRunfile,lfa,0);         /* Seek to end of symbols */
	ChkRunErr();                    /* Check for error */
      }
#endif
    lfa = ftell(bsRunfile);             /* Get offset */
    if(cbPadding = (WORD)(lfa & 7)) cbPadding = 8 - cbPadding;
					/* Start relocs on 8-byte boundary */
    cbRle = ((long) stMac << 3) + cbPadding;
					/* Get updated reloc table size */
    fseek(bsRunfile,IBWRLE,0);          /* Seek to header entry */
    ChkRunErr();                        /* Check for error */
    if(fLoWordFirst)                    /* If lo word first in longs */
      {
	OutWord(bsRunfile,LOWORD(cbRle));
					/* Write lo word of reloc length */
	OutWord(bsRunfile,HIWORD(cbRle));
					/* Write hi word of reloc length */
      }
    else                                /* Else if hi word first in longs */
      {
	OutWord(bsRunfile,HIWORD(cbRle));
					/* Write hi word of reloc length */
	OutWord(bsRunfile,LOWORD(cbRle));
					/* Write lo word of reloc length */
      }
    ChkRunErr();                        /* Check for error */
    fseek(bsRunfile,lfa,0);             /* Seek to end of symbols */
    ChkRunErr();                        /* Check for error */
    while(cbPadding--) OutByte(bsRunfile,'\0');
					/* Add padding */
    ChkRunErr();                        /* Check for error */
    for(st = 0; st < stMac; ++st)       /* Loop to write segment table */
      {
	OutWord(bsRunfile,(mpstcb[st]? mpstcb[st] - 1: 0));
					/* Output segment length */
	OutWord(bsRunfile,LOWORD(mpstlfa[st]));
					/* Output low word of lfa */
	OutByte(bsRunfile,(BYTE) HIWORD(mpstlfa[st]));
					/* Output high byte of lfa */
	OutByte(bsRunfile,mpstacc[st]); /* Output access code */
	OutWord(bsRunfile,0);           /* Unused field */
	ChkRunErr();                    /* Check for error */
      }
    fclose(bsRunfile);                  /* Close run file */
    fRunFileOpen = FALSE;               /* Run file closed */
#if OSXENIX
    if(!fUndefinedExterns) chmod(&sbOverlay[1],0755);
					/* Set protection executable */
#endif
    LEAVE(OutRunFile,0xFFFF);           /* Debugging info */
  }
#endif
