    /****************************************************************
    *                                                               *
    *                           NEWREL.C                            *
    *                                                               *
    ****************************************************************/

#include                "minlit.h"      /* Types and constants */
#include                "bndtrn.h"      /* Basic types and constants */
#include                "bndvrt.h"      /* Virtual memory stuff */
#include                "bndrel.h"      /* Relocation definitions */
#include                "bndmi.h"       /* Machine instruction op codes */
#include                "lnkio.h"       /* Linker I/O definitions */

extern BYTE             Gets();         /* Get a byte from input file */
extern WORD             WGets();        /* Get a word from input file */
extern                  OutError();     /* Output an error message */
extern                  OutFileCur();   /* Output name and pos in input file */
extern WORD             GetIndex();     /* Read in an index value */
extern                  InvalidObject();/* Fatal error routine */
extern                  Fatal();        /* Fatal error message routine */

extern FTYPE            fLstFileOpen;   /* Map-file-open flag */
extern FTYPE            fOverlays;      /* True if there are overlays */
extern WORD             cErrors;        /* Count of non-fatal errors */
extern WORD             cbRec;          /* Record size in bytes */
extern MITYPE           rgmi[];         /* Array of code */
extern SEGTYPE          segLast;        /* Number of final segment */
extern WORD             snMac;          /* Count of local SEGDEFs */
extern WORD             extMac;         /* Count of publics */
extern SEGTYPE          segStart;       /* Number of first segment */
extern RATYPE           raStart;        /* Rel. addr. of 1st byte in seg */
extern FTYPE            vfDSAlloc;      /* Do-DS-Allocation flag */
extern WORD             csegsAbs;       /* Count of absolute segments */
extern SEGTYPE          mpgsnseg[];     /* Map segdef no. to segment */
extern RATYPE           mpgsndra[];
extern SNTYPE           mpggrgsn[];
extern RATYPE           mpextra[];      /* Map public no. to its rel addr */
extern SNTYPE           mpextgsn[];     /* Map public no. to its SEGDEF */
extern SNTYPE           mpsngsn[];      /* Map local SEGDEF to global */
extern FTYPE            mpsegFlags[];   /* Map segment to flags */
extern SATYPE           mpsegsa[];      /* Map segment to base value */
extern IOVTYPE          mpsegiov[];     /* Map segment to its overlay no. */
#if NOT OIAPX286
extern WORD             mpiovcRle[];    /* Count of long reloc.s per overlay */
#endif
extern BYTE             mprlcb[];
extern KINDTYPE         mprlckind[];
extern KINDTYPE         mprlcGroupKind[];
extern BYTE             mprlcbGroup[];
extern WORD             cbRgRleMac;
extern GRTYPE           mpgrggr[];      /* Map GRPDEF no. to global GRPDEF */
extern WORD             grMac;          /* Count of local GRPDEFs */
extern RECTTYPE         vrectData;      /* Type of data record */
extern SEGTYPE          vsegCur;        /* Current segment */
extern RATYPE           vraCur;         /* Current rel addr */
extern SNTYPE           vgsnCur;        /* Current global SEGDEF */
extern BSTYPE           bsLst;          /* File handle for listing file */
extern char             *ompimisegDstIdata;
					/* Try this one in hangman sometime! */
#if OXOUT
extern FTYPE            fMedium;        /* Medium model flag */
#endif

    /****************************************************************
    *                                                               *
    *  Getgsn:                                                      *
    *                                                               *
    *  This  function processes  a partially decoded  FixDat byte,  *
    *  returning  gsn and ra  if they  are specified.  It takes as  *
    *  its inputs the kind of index  being used, the index itself,  *
    *  a return address  for the gsn and a return  address for the  *
    *  ra ("useful for extern/seg").  The function does not return  *
    *  a meaningful value.                                          *
    *                                                               *
    ****************************************************************/

LOCAL                   Getgsn(kind,idx,pgsn,pra)
KINDTYPE                kind;           /* Kind of index */
WORD                    idx;            /* The index */
SEGTYPE                 *pgsn;          /* gsn (ref) */
RATYPE                  *pra;           /* ra (ref) */
  {
    ENTER(Getgsn);                      /* Debug info */
    DEBUGVALUE(kind);                   /* Debug info */
    DEBUGVALUE(idx);                    /* Debug info */
    switch(kind)                        /* Decide what to do */
      {
	case KINDSEG:                   /* Segment index */
	  if(idx >= snMac) InvalidObject();
					/* Make sure index not too big */
	  *pgsn = mpsngsn[idx];         /* Get gsn */
	  *pra = mpgsndra[*pgsn];       /* Get ra */
	  break;

	case KINDGROUP:                 /* Group index */
	  if(idx >= grMac) InvalidObject();
					/* Make sure index not too big */
	  *pgsn = mpggrgsn[mpgrggr[idx]];
					/* Get gsn */
	  *pra = mpgsndra[*pgsn];       /* Get ra */
	  break;

	case KINDEXT:                   /* External index */
	  if(idx >= extMac) InvalidObject();
					/* Make sure index not too big */
	  *pgsn = mpextgsn[idx];        /* Get gsn */
	  *pra = mpextra[idx];          /* Get ra */
	  break;

	default:                        /* All other kinds */
	  *pgsn = SEGNIL;               /* No gsn */
	  *pra = 0;                     /* No ra */
	  break;
      }
    DEBUGVALUE(*pgsn);                  /* Debug value */
    DEBUGVALUE(*pra);                   /* Debug value */
    LEAVE(Getgsn,0xFFFF);               /* Debug info */
  }

    /****************************************************************
    *                                                               *
    *  GetFrameTarget:                                              *
    *                                                               *
    *  This function reads and  processes a FixDat byte, returning  *
    *  the frame, target, and dra(???).  It takes  as  its  inputs  *
    *  the  return  address  for  the  gsnFrame(???),  the  return  *
    *  address of  the gsnTarget(???), the return  address  of the  *
    *  raTarget(???), and  the  return  address  of the "dra to be  *
    *  applied  to  this  fixup."  The function does  not return a  *
    *  meaningful value.                                            *
    *  See pp. 72-73 in "8086 Object Module Formats EPS."           *
    *                                                               *
    ****************************************************************/

LOCAL                   GetFrameTarget(pgsnFrame,pgsnTarget,praTarget,pdra)
SEGTYPE                 *pgsnFrame;     /* gsnFrame (ref) */
SEGTYPE                 *pgsnTarget;    /* gsnTarget (ref) */
RATYPE                  *praTarget;     /* raTarget (ref) */
RATYPE                  *pdra;          /* raType (ref) */
  {
    BYTE                fixdat;         /* The fixdat byte */
    KINDTYPE            kindFrame;
    KINDTYPE            kindTarget;
    WORD                idxFrame;       /* Frame index */
    WORD                idxTarget;      /* Target index */
    register WORD       i;              /* Index variable */

    ENTER(GetFrameTarget);              /* Debug info */
    fixdat = Gets();                    /* Read in the fixdat byte */
    DEBUGVALUE(B2W(fixdat));            /* Print debug info */
    if(fixdat & '\200')                 /* If frame given by thread */
      {
	i = ((WORD) fixdat >> 4) & 3;   /* Calculate the thread number */
#if DEBUG                               /* If debugging on */
	fprintf(stderr,"Frame thread number %x\n",i);
					/* Message */
#endif                                  /* End debugging code */
	idxFrame = mprlcbGroup[i];      /* Get the frame index */
	kindFrame = mprlcGroupKind[i];  /* Get the frame kind */
      }
    else                                /* Else if frame given explicitly */
      {
	DEBUGMSG("Frame explicit");     /* Debug message */
	kindFrame = ((WORD) fixdat >> 4) & 7;
					/* Calculate method number */
	/* Method no:   Frame specification:
	*  0            segment index
	*  1            group index
	*  2            external index
	*  3            frame number
	*  4            implicit (local)
	*  5            implicit (target)
	*  6            none
	*  7            invalid
	*/
	ASSERT(kindFrame != 3);         /* Unimplemented by this linker */
	ASSERT(kindFrame != 7);         /* Invalid */
	if(kindFrame <= 2) idxFrame = GetIndex(0,IMAX);
					/* If specified by index, get it */
      }
    DEBUGVALUE(kindFrame);              /* Debug info */
    DEBUGVALUE(idxFrame);               /* Debug info */
    if(fixdat & '\010')                 /* If target specified by thread ref */
      {
	i = (WORD) fixdat & 3;          /* Get thread number */
#if DEBUG                               /* If debugging on */
	fprintf(stderr,"Target thread number %x\n",i);
					/* Message */
#endif                                  /* End debugging code */
	idxTarget = mprlcb[i];          /* Get target index */
	kindTarget = mprlckind[i];      /* Get target kind */
      }
    else                                /* Else if specified explicitly */
      {
	DEBUGMSG("Target explicit");    /* Debug message */
	kindTarget = (WORD) fixdat & 3; /* Get method number */
	/* Method no.   Target specification:
	*  0            segment index
	*  1            group index
	*  2            external index
	*  3            frame number
	*/
	ASSERT(kindTarget != 3);        /* Unimplemented by this linker */
	idxTarget = GetIndex(0,IMAX);   /* Get index */
      }
    DEBUGVALUE(kindTarget);             /* Debug info */
    DEBUGVALUE(idxTarget);              /* Debug info */
    if(kindFrame == 4 || kindFrame == 5)/* If frame spec implicit */
      {
	kindFrame = kindTarget;         /* Use target frame kind */
	idxFrame = idxTarget;           /* Use target index */
	DEBUGVALUE(kindFrame);          /* Debug info */
	DEBUGVALUE(idxFrame);           /* Debug info */
      }
    Getgsn(kindFrame,idxFrame,pgsnFrame,pdra);
					/* Get gsn and ra, if any */
    Getgsn(kindTarget,idxTarget,pgsnTarget,praTarget);
					/* Get gsn and ra, if any */
    if(!(fixdat & '\004')) *pdra = WGets();
					/* If target spec secondary, */
    else *pdra = 0;                     /* get the target displacement */
    DEBUGVALUE(*pgsnFrame);             /* Debug info */
    DEBUGVALUE(*pgsnTarget);            /* Debug info */
    DEBUGVALUE(*praTarget);             /* Debug info */
    DEBUGVALUE(*pdra);                  /* Debug info */
    LEAVE(GetFrameTarget,0xFFFF);       /* Debug info */
  }

    /****************************************************************
    *                                                               *
    *  EndRec:                                                      *
    *                                                               *
    *  This   function  is  called  to   process  the  information  *
    *  contained  in  a  MODEND  (type 8AH) record  concerning the  *
    *  program  starting address.  The function  does not return a  *
    *  meaningful value.                                            *
    *  See pp. 80-81 in "8086 Object Module Formats EPS."           *
    *                                                               *
    ****************************************************************/

EndRec()
  {
    BYTE                modtyp;         /* MODEND record modtyp byte */
    SEGTYPE             gsnTarget;
    SEGTYPE             gsnFrame;
    RATYPE              dra;
    RATYPE              raTarget;       /* Fixup target offset */
    RATYPE              ra;
    SATYPE              dsa;
    SEGTYPE             segTarget;      /* Target segment */
    SEGTYPE             segFrame;

    ENTER(EndRec);                      /* Debug info */
    modtyp = Gets();                    /* Read modtyp byte */
    /* if(modtyp & 0x80)                /* If main module found */
    /*   { */
    /*     if(fMain) Fatal("Multiple main modules specified.",0x143); */
    /*                                  /* Check if main already defined */
    /*     fMain = TRUE;                /* Set flag */
    /*   } */
    if(modtyp & FSTARTADDRESS)          /* If execution start address given */
      {
	ASSERT(modtyp & 1);             /* Must have logical start address */
	GetFrameTarget(&gsnFrame,&gsnTarget,&raTarget,&dra);
					/* Get fixup information */
	if(gsnFrame == SEGNIL) gsnFrame = gsnTarget;
					/* Use target val. if none given */
	segFrame = mpgsnseg[gsnFrame];  /* Get frame segment */
	segTarget = mpgsnseg[gsnTarget];/* Get target segment */
	dsa = LO16BITS(mpsegsa[segTarget] - mpsegsa[segFrame]);
					/* Calculate base delta */
	DEBUGVALUE(dsa);                /* Give debug info */
	DEBUGVALUE(raTarget);           /* Give debug info */
	DEBUGVALUE(dra);                /* Give debug info */
#if NOT OIAPX286
	if(dsa > 0x1000) FixupOverflow(raTarget + dra);
					/* Delta > 64Kbytes */
	ra = LO16BITS(dsa << 4);
	if(0xFFFF - ra < raTarget)      /* If addition would overflow */
	  {
	    ra = LO16BITS(ra - 0xFFFF + raTarget);
					/* Fix up addition */
	    --ra;
	    FixupOverflow(raTarget + dra);
	  }
	else ra = LO16BITS(ra + raTarget);
					/* Else perform addition */
#endif
#if OIAPX286
	if(dsa) FixupOverflow(raTarget + dra);
					/* No intersegment fixups */
	ra = raTarget;                  /* Use target offset */
#endif
	if(0xFFFF - ra < dra)           /* If addition would overflow */
	  {
	    ra = LO16BITS(ra - 0xFFFF + dra);
					/* Fix up addition */
	    --ra;
	    FixupOverflow(raTarget + dra);
	  }
	else ra = LO16BITS(ra + dra);   /* Else perform addition */
	if(segStart == SEGNIL)
	  {
	    segStart = segFrame;
	    raStart = ra;
	    if(fLstFileOpen)            /* If there is a listing file */
	      {
		NEWLINE(bsLst);         /* Skip a line */
		fprintf(bsLst,"Program entry point at %04x:%04x",
		  mpsegsa[segStart],raStart);
					/* Print entry point */
		NEWLINE(bsLst);         /* Newline */
	      }
	  }
      }
    LEAVE(EndRec,0xFFFF);               /* Debug info */
  }

    /****************************************************************
    *                                                               *
    *  SetRlc:                                                      *
    *                                                               *
    *  See pp. 70-71 in "8086 Object Module Formats EPS."           *
    *                                                               *
    ****************************************************************/

LOCAL                   SetRlc(trddat,fFrame)
RLCTYPE                 trddat;         /* trddat byte from fixup record */
FTYPE                   fFrame;         /* 'D' subfield from trddat byte */
  {
    WORD                idx = 0;        /* Index */
    KINDTYPE            kind;           /* Index kind */
    register WORD       threadno;       /* Thread number (0 - 3) */

    ENTER(SetRlc);                      /* Debug info */
    kind = ((WORD) trddat >> 2) & 7;    /* Get method subfield */
    threadno = (WORD) trddat & 3;       /* Get the thread number */
    if(fFrame)                          /* If frame thread definition */
      {
	ASSERT(kind != 3);              /* Frame number unimplemented */
	if(kind <= 2) idx = GetIndex(1,IMAX);
					/* Get index if specified */
	mprlcGroupKind[threadno] = kind;/* Save the index kind */
	mprlcbGroup[threadno] = idx;    /* Save the index value */
      }
    else                                /* Else if target thread definition */
      {
	kind &= 3;                      /* Get proper method subfield */
	ASSERT(kind != 3);              /* Frame number unimplemented */
	idx = GetIndex(1,IMAX);         /* Get index */
	mprlckind[threadno] = kind;     /* Save the index kind */
	mprlcb[threadno] = idx;         /* Save the index value */
      }
    LEAVE(SetRlc,0xFFFF);               /* Debug info */
  }

#if NOT OIAPX286
    /****************************************************************
    *                                                               *
    *  RecordSegmentReference:                                      *
    *                                                               *
    ****************************************************************/

RecordSegmentReference(seg,ra,segDst)
SEGTYPE                 seg;
RATYPE                  ra;
SEGTYPE                 segDst;
  {
    SEGTYPE             segAbsLast;     /* Last absolute segemnt??? */
    RLETYPE             rle;
#if OXOUT
    long                xxaddr;         /* Twenty bit address */
#endif

    ENTER(RecordSegmentReference);      /* Debug info */
    segAbsLast = segLast + csegsAbs;    /* Calc. last absolute seg no. */
    if(vfDSAlloc) --segAbsLast;
    if(segDst > segLast && segDst <= segAbsLast) return;
					/* Don't bother if absolute segment */
    if(vrectData == LIDATA) ompimisegDstIdata[ra - vraCur] = segDst;
    else                                /* Else if not iterated data */
      {
	rle.iov = mpsegiov[seg];        /* Get overlay number */
#if OEXE
	rle.rgbrasa[RLEOSA] = (BYTE)(mpsegsa[seg] & 0377);
	rle.rgbrasa[RLEOSA + 1] = (BYTE)((mpsegsa[seg] >> BYTELN) & 0377);
					/* Get segment address */
	rle.rgbrasa[RLEORA] = (BYTE)(ra & 0377);
	rle.rgbrasa[RLEORA + 1] = (BYTE)((ra >> BYTELN) & 0377);
					/* Save relative address */
#endif
#if OXOUT
	xxaddr = ((long) mpsegsa[seg] << 4) + (long) ra;
					/* Form 20-bit address */
	rle.bhigh = (BYTE)((xxaddr >> WORDLN) & 0377);
					/* High 8 bits of address */
	rle.wlow[1] = (BYTE)((xxaddr >> BYTELN) & 0377);
					/* Middle 8 bits of address */
	rle.wlow[0] = (BYTE)(xxaddr & 0377);
					/* Low 8 bits of address */
	if(mpsegFlags[seg] & FCODE == FCODE)
	  {
	    if(mpsegFlags[segDst] & FCODE == FCODE)
	      {
		if(!fMedium)
		  Fatal("Long code to code reference; use -M.",0xB0);
					/* Must be medium model */
		else rle.ref = FCODETOCODE;
					/* Log reference type */
	      }
	    else rle.ref = FCODETODATA; /* Log reference type */
	  }
	else if(mpsegFlags[segDst] & FCODE == FCODE) rle.ref = FDATATOCODE;
					/* Log reference type */
	else rle.ref = FDATATODATA;     /* Log reference type */
#endif
	ASSERT(fOverlays || rle.iov == IOVROOT);
					/* If no overlays then iov = IOVROOT */
	++mpiovcRle[rle.iov];
	MoveToVm(CBRLE,&rle,AREARGRLE,cbRgRleMac);
					/* Save in VM */
	cbRgRleMac += CBRLE;            /* Update byte counter */
      }
    LEAVE(RecordSegmentReference,0xFFFF);
					/* Debug info */
  }
#endif

    /****************************************************************
    *                                                               *
    *  RelocErr:                                                    *
    *                                                               *
    *  Error message routine.                                       *
    *                                                               *
    ****************************************************************/

RelocErr(msg,ra)
char                    *msg;           /* Error message */
RATYPE                  ra;             /* Relative addr of error */
  {
    BSTYPE              bsTmp;          /* Save file handle */

    ++cErrors;                          /* Increment error count */
    bsTmp = stdout;                     /* Send message to screen first */
    for(;;)                             /* Loop to give message */
      {
	fprintf(bsTmp,"%s in ",msg);    /* Output message */
	OutFileCur(bsTmp);              /* Current file name */
	fprintf(bsTmp," near %04x",ra - mpgsndra[vgsnCur]);
					/* Offset */
	NEWLINE(bsTmp);                 /* Newline */
	fflush(bsTmp);                  /* Force message out */
	if(!fLstFileOpen || bsTmp == bsLst) break;
					/* Exit loop */
	bsTmp = bsLst;                  /* Insure loop exit */
      }
  }

    /****************************************************************
    *                                                               *
    *  FixupOverflow:                                               *
    *                                                               *
    *  Error message routine.                                       *
    *                                                               *
    ****************************************************************/

LOCAL                   FixupOverflow(ra)
RATYPE                  ra;             /* Relative addr of error */
  {
    RelocErr("Fixup offset exceeds field width",ra);
  }

    /****************************************************************
    *                                                               *
    *  DoRelocItem:                                                 *
    *                                                               *
    *  See pp. 71-73 in "8086 Object Module Formats EPS."           *
    *                                                               *
    ****************************************************************/

LOCAL                   DoRelocItem(loc,imiHigh,fSelfRel)
WORD                    loc;            /* LOC subfield in LOCAT lobyte */
WORD                    imiHigh;        /* Hi bits of data record offset */
FTYPE                   fSelfRel;       /* True if self-relative fixup */
  {
    register WORD       imi;            /* Data record offset subfield */
    SEGTYPE             segTarget;      /* Fixup target segment */
    SEGTYPE             segFrame;       /* Fixup frame segment */
    SEGTYPE             gsnTarget;
    SEGTYPE             gsnFrame;
    RATYPE              raTarget;       /* Fixup target rel. addr. */
    RATYPE              raTmp;
    RATYPE              ra;
    WORD                dra;
    WORD                draFinal;
    WORD                dsa;

    ENTER(DoRelocItem);                 /* Debug info */
    imi = B2W(Gets()) | (imiHigh << BYTELN);
					/* Calculate data record offset */
    ra = vraCur + imi;                  /* Get rel. addr. of fixup */
    GetFrameTarget(&gsnFrame,&gsnTarget,&raTarget,&draFinal);
					/* Process the FIXDAT byte */
    segTarget = mpgsnseg[gsnTarget];    /* Get target segment */
    if(gsnFrame != SEGNIL) segFrame = mpgsnseg[gsnFrame];

    else segFrame = SEGNIL;
    if(vsegCur == SEGNIL) return;
    if(gsnTarget == SEGNIL)
      {
	if(loc == FIXULONG)             /* If "pointer" (4 byte) fixup */
	  {
	    if(imi > 1)
	      {
		if(rgmi[imi - 1] == CALLDIRECTFAR)
		  rgmi[imi - 1] = BREAKPOINT;
					/* Replace long call w/ breakpoint */
	      }
	    return;
	  }
	if(loc == FIXUSEG) return;      /* Return if "base" (2 byte) fixup */
      }
    else
      {
	if(segFrame != SEGNIL)
	  {
	    dsa = LO16BITS(mpsegsa[segTarget] - mpsegsa[segFrame]);
#if NOT OIAPX286
	    dra = (dsa & 0xFFF) << 4;
	    raTmp = LO16BITS(raTarget + dra);
	    DEBUGVALUE(segTarget);      /* Debug info */
	    DEBUGVALUE(mpsegsa[segTarget]);
	    DEBUGVALUE(segFrame);       /* Debug info */
	    DEBUGVALUE(mpsegsa[segFrame]);
	    DEBUGVALUE(raTarget);       /* Debug info */
	    if(dsa >= 0x1000 || raTmp < raTarget) FixupOverflow(ra);
					/* Check for error */
	    raTarget = raTmp;
#endif
#if OIAPX286
	    if(dsa) FixupOverflow(ra);  /* No intersegment fixups */
#endif
	    segTarget = segFrame;
	    DEBUGVALUE(raTarget);       /* Debug info */
	    DEBUGVALUE(segTarget);      /* Debug info */
	  }
	if(fSelfRel)                    /* If self-relative fixup */
	  {
	    if(segFrame != SEGNIL)
	      dsa = LO16BITS(mpsegsa[vsegCur] - mpsegsa[segFrame]);
	    else dsa = LO16BITS(mpsegsa[vsegCur] - mpsegsa[segTarget]);
	    dra = (dsa & 0xFFF) << 4;
	    DEBUGVALUE(dsa);            /* Debug info */
	    DEBUGVALUE(ra);             /* Debug info */
#if NOT OIAPX286
	    if(dsa >= 0x1000 || 0xFFFF - ra < dra) FixupOverflow(ra);
					/* Check for error */
#endif
#if OIAPX286
	    if(dsa) FixupOverflow(ra);  /* No intersegment fixups */
#endif
	    if(loc == FIXUWORD)
	      raTarget = LO16BITS(raTarget - dra - ra - 2);
	    else raTarget = LO16BITS(raTarget - dra - ra - 1);
	    DEBUGVALUE(raTarget);       /* Debug info */
	  }
      }
    raTmp = raTarget;
    DEBUGVALUE(draFinal);               /* Debug info */
    raTarget += draFinal + B2W(rgmi[imi]) + (B2W(rgmi[imi + 1]) << BYTELN);
#if ODDWORDLN                           /* If machine word length > 16 */
    raTarget = LO16BITS(raTarget);      /* Take only lo order 16 bits */
#endif
    DEBUGVALUE(raTarget + 0);           /* Debug info */
    switch(loc)                         /* Switch on fixup type */
      {
	case 0:                         /* 8-bit "lobyte" fixup */
	  raTarget = raTmp + B2W(rgmi[imi]) + draFinal;
#if ODDWORDLN                           /* If machine word length > 16 */
	  raTarget = LO16BITS(raTarget);/* Take only lo order 16 bits */
#endif
	  rgmi[imi] = (BYTE) raTarget;
	  if(raTarget >= 0x100 && (!fSelfRel || raTarget < 0xff80))
	    FixupOverflow(ra);
	  break;

	case 1:                         /* 16-bit "offset" fixup */
	  rgmi[imi] = (BYTE)(raTarget & 0377);
	  rgmi[imi + 1] = (BYTE)((raTarget >> BYTELN) & 0377);
					/* Perform fixup */
	  break;

	case 2:                         /* 16-bit "base" fixup */
	  rgmi[imi] = (BYTE)(mpsegsa[segTarget] & 0377);
	  rgmi[imi + 1] = (BYTE)((mpsegsa[segTarget] >> BYTELN) & 0377);
					/* Perform fixup */
#if NOT OIAPX286
	  RecordSegmentReference(vsegCur,ra,segTarget);
					/* Record reference */
#endif
	  break;

	case 3:                         /* 32-bit "pointer" fixup */
	  rgmi[imi] = (BYTE)(raTarget & 0377);
	  rgmi[imi + 1] = (BYTE)((raTarget >> BYTELN) & 0377);
					/* Fix up offset */
	  imi += 2;                     /* Advance to next word */
	  rgmi[imi] = (BYTE)(mpsegsa[segTarget] & 0377);
	  rgmi[imi + 1] = (BYTE)((mpsegsa[segTarget] >> BYTELN) & 0377);
					/* Fix up base */
#if NOT OIAPX286
	  RecordSegmentReference(vsegCur,ra + 2,segTarget);
					/* Record reference */
#endif
	  break;
      }
    LEAVE(DoRelocItem,0xFFFF);          /* Debug info */
  }

    /****************************************************************
    *                                                               *
    *  RelocRec:                                                    *
    *                                                               *
    *  See pp. 70-73 in "8086 Object Module Formats EPS."           *
    *                                                               *
    ****************************************************************/

RelocRec()
  {
    WORD                b;
    WORD                bOp1;
    WORD                bOp2;
    WORD                bData;

    ENTER(RelocRec);                    /* Debug info */
    while(cbRec > 1)                    /* While not at end of record */
      {
	b = B2W(Gets());                /* Read a byte */
	bOp1 = b >> 6;                  /* Get high two bits */
	bOp2 = (b >> 2) & 3;            /* Get METHOD/LOC subfield */
	bData = b & 3;                  /* Get data offset bits */
	switch(bOp1)                    /* Process thread or fixup */
	  {
	    case 0:                     /* Target thread definition */
	      /* "Set relocation counter" */
	      SetRlc((RLCTYPE) b & 0xF,FALSE);
					/* Process thread */
	      break;

	    case 1:                     /* Frame thread definition */
	      /* "Set group relocation counter" */
	      SetRlc((RLCTYPE) b & 0xF,TRUE);
					/* Process thread */
	      break;

	    case 2:                     /* Self-relative fixup */
	      /* "Self relative relocation" */
	      DoRelocItem(bOp2,bData,TRUE);
					/* Do fixup */
	      break;

	    case 3:                     /* Segment-relative fixup */
	      /* "Vanilla relocation" */
	      DoRelocItem(bOp2,bData,FALSE);
					/* Do fixup */
	      break;
	  }
      }
    LEAVE(RelocRec,0xFFFF);             /* Debug info */
  }
