    /****************************************************************
    *                                                               *
    *                           BNDTP1                              *
    *                                                               *
    *  Pass 1 object module processing routines.                    *
    *                                                               *
    ****************************************************************/

#include                "minlit.h"      /* Basic types and constants */
#include                "bndtrn.h"      /* More types and constants */
#include                "bndvrt.h"      /* Virt. mem. system declarations */
#include                "xenfmt.h"      /* Executable file header def.s */
#include                "lnkio.h"       /* Linker I/O definitions */

extern                  SetPascalDefaults();
					/* What it says! */
extern                  OutFileCur();
extern BYTE             Gets();         /* Read in a byte */
extern WORD             WGets();        /* Read in a word */
extern WORD             GetIndex();     /* Read 1- or 2-byte index */
extern                  OutError();     /* Output an error message */
extern                  Fatal();        /* Fatal error routine */
extern                  InvalidObject();/* Invalid object error routine */
extern                  SkipBytes();    /* Skip n bytes on input */
extern PROPTYPE         PropSymLookup();/* Look up symbol/property pair */
extern PROPTYPE         PropAdd();      /* Add property to a symbol's list */
extern PROPTYPE         PropRhteLookup();
					/* Look up property given rhte */
extern                  AddLibrary();   /* Add library to search list */
extern short            SbCompare();    /* Cmp two length-prefixed strings */

extern FTYPE            rgfDirty[];     /* Used by DirtyBp() macro */
extern FTYPE            fLstFileOpen;   /* True if list file open */
extern WORD             cErrors;        /* Count of non-fatal errors */
extern WORD             cbRec;          /* Size of object record in bytes */
extern RECTTYPE         rect;           /* Object record type */
extern WORD             vbp;            /* Virtual block number */
extern char             *vob;
extern FTYPE            vfNoDefaultLibraySearch;
					/* How could I be more eloquent?!! */
extern FTYPE            vfCreated;      /* New hash table entry created flag */
extern FTYPE            fOverlays;      /* Overlays exist flag */
extern RBTYPE           vrhte;          /* Address of hash table entry */
extern RBTYPE           vrprop;         /* Address of property cell */
extern RBTYPE           vrpropFile;
extern RATYPE           mpgsndra[];     /* f(gsn) = dra */
extern RBTYPE           mplnamerhte[];  /* f(LNAME no.) = hash table addr */
extern WORD             lnameMac;       /* Count of LNAMEs */
extern WORD             extMac;         /* Count of PUBDEFs */
extern WORD             snMac;          /* Count of SEGDEFs */
extern WORD             iovMac;         /* Count of overlays */
extern RBTYPE           mpiovrhte[];    /* f(overlay no.) = hte addr. */
extern SNTYPE           mpsngsn[];      /* f(SEGDEF no.) = gsn */
extern WORD             gsnMac;         /* Count of gsn's */
extern WORD             ggrMac;         /* Count of ggr's */
extern RBTYPE           mpgsnrprop[];   /* f(gsn) = prop. cell addr. */
extern RATYPE           mpextra[];      /* f(PUBDEF no.) = segment offset */
extern SNTYPE           mpextsn[];      /* f(PUBDEF no.) = SEGDEF no. */
extern RBTYPE           vrhteCODEClass; /* Addr. of hte for class CODE */
extern RBTYPE           vrhteRootOverlay;
					/* Addr. of hte for root overlay */
extern GRTYPE           grMac;          /* Count of GRPDEFs */
extern SNTYPE           gsnStack;       /* Stack segment number */
#if CALLOCATION
extern WORD             mpitypelen[];   /* f(TYPDEF no.) = type length */
extern WORD             typMac;         /* Count of TYPDEFs */
#endif
#if OXOUT OR OIAPX286
extern FTYPE            fIandD;         /* TRUE if separate I & D */
#endif
extern BSTYPE           bsLst;          /* List file handle */
extern BSTYPE           bsInput;        /* Input file handle */

    /****************************************************************
    *                                                               *
    *  ModRc1:                                                      *
    *                                                               *
    *  This function reads the name from a THEADR record and makes  *
    *  an entry containing the name in the hash table.              *
    *  See p. 26 in "8086 Object Module Formats EPS."               *
    *                                                               *
    ****************************************************************/

ModRc1()
  {
    SBTYPE              sbModule;       /* The module name */
    APROPFILETYPE       *apropFile;

    ENTER(ModRc1);                      /* Debug info */
    sbModule[0] = Gets();               /* Read in the length byte */
    GetBytes(&sbModule[1],B2W(sbModule[0]));
					/* Read in the name */
    apropFile = (APROPFILETYPE *) ObGetVa(AREASYMS,vrpropFile,TRUE);
					/* Allocate space in virt mem */
    vob = PropSymLookup(sbModule,ATTRNIL,TRUE);
					/* Make entry in hash table */
    apropFile->af_rMod = vrhte;         /* Load pointer into hash table */
    LEAVE(ModRc1,0xFFFF);               /* Debug info */
  }

#if CALLOCATION
    /****************************************************************
    *                                                               *
    *  TypRc1:                                                      *
    *                                                               *
    *  This function processes  TYPDEF records.  These records are  *
    *  difficult  to  understand.  They are (poorly) described  on  *
    *  pp. 40-43 of "8086  Object  Module  Formats EPS," with some  *
    *  additional information on pp. 89-90.  These records contain  *
    *  debugging  information  only,   and  they   are  not  fully  *
    *  supported by this linker.                                    *           
    *                                                               *
    ****************************************************************/

TypRc1()
  {
    long                l;
    WORD                b;

    ENTER(TypRc1);                      /* Debug info */
    SkipBytes(B2W(Gets()));             /* Skip the name field */
    Gets();                             /* Skip the EN byte */
    l = 0;                              /* Initialize */
    if(cbRec >= 2)                      /* If at least two bytes left */
      {
	if(B2W(Gets()) == 0x62)         /* Numeric leaf 98 */
	  {
	    if(cbRec < 2) InvalidObject();
					/* Make sure record long enough */
	    b = B2W(Gets());            /* Get next leaf */
	    if(b == 0x7B || b == 0x79 || b == 0x77)
	      {                         /* Scalar, structure, or array */
		if(cbRec < 2) InvalidObject();
					/* Make sure record long enough */
		b = B2W(Gets());        /* Get length byte */
		if(b < 128) l = B2L(b); /* One byte length field */
		else if(b == 129)       /* Else if two byte length */
		  {
		    if(cbRec < 3) InvalidObject();
					/* Make sure record long enough */
		    l = (long) WGets(); /* Get the length */
		  }
		else if(b == 132)       /* Else if three byte length */
		  {
		    if(cbRec < 4) InvalidObject();
					/* Make sure record long enough */
		    l = (long) WGets(); /* Get the length */
		    l += B2L(Gets()) << WORDLN;
					/* Get the length */
		  }
		l = (l + 7) >> 3;       /* Round to nearest 8-byte bound */
	      }
	  }
      }
    if(typMac >= TYPMAX) Fatal("Too many TYPDEFs.",0x91);
					/* Check table limit */
    mpitypelen[typMac++] = (WORD) l;    /* Store length */
    SkipBytes(cbRec - 1);               /* Skip all but the checksum */
    LEAVE(TypRc1,0xFFFF);               /* Debug info */
  }
#endif

    /****************************************************************
    *                                                               *
    *  LNmRc1:                                                      *
    *                                                               *
    *  This function  reads LNAME records and  stores the names in  *
    *  the  hash  table.  In  addition, when  it  finds  the  name  *
    *  "CODE," it saves the address of the hash table entry in the  *
    *  the global variable  vrhteCODEClass.  The function does not  *
    *  return a meaningful value.                                   *
    *  See p. 31 in "8086 Object Module Formats EPS."               *
    *                                                               *
    ****************************************************************/

LNmRc1()
  {
    SBTYPE              lname;          /* Buffer for lnames */

    ENTER(LNmRc1);                      /* Debug info */
    while(cbRec > 1)                    /* While not at end of record */
      {
	if(lnameMac >= LNAMEMAX)
	  Fatal("Too many groups,segments,classes.",0x92);
					/* Check for table overflow */
	lname[0] = Gets();              /* Get name length */
	DEBUGVALUE(B2W(lname[0]));      /* Length of name */
	GetBytes(&lname[1],B2W(lname[0]));
					/* Read in text of name */
	DEBUGSB(lname);                 /* The name itself */
	vob = PropSymLookup(lname,ATTRNIL,TRUE);
					/* Insert symbol in hash table */
	mplnamerhte[lnameMac++] = vrhte;/* Save address of hash table entry */
	if(!SbCompare(lname,"\004CODE",1) && !vrhteCODEClass)
	  vrhteCODEClass = vrhte;       /* Save addr of class "CODE" */
      }
    LEAVE(LNmRc1,0xFFFF);               /* Debug info */
  }

    /****************************************************************
    *                                                               *
    *  SegRc1:                                                      *
    *                                                               *
    *  This function processes SEGDEF records.                      *
    *  See pp. 32-35 in "8086 Object Module Formats EPS."           *
    *                                                               *
    ****************************************************************/

SegRc1()
  {
    TYSNTYPE            tysn;
    LNAMETYPE           lname;          /* Index into mplnamerhte??? */
    LNAMETYPE           lnameClass;     /* Index into mplnamerhte??? */
    WORD                seglen;         /* Segment length */
    APROPSNTYPE         *apropSn;
    APROPOVERLAYTYPE    *apropOverlay;
    SNTYPE              gsn;
    ALIGNTYPE           align;          /* Segment alignment */
    WORD                comb;           /* Segment combinality */
    WORD                cbMaxPrev;
    AHTETYPE            *ahte;          /* Pointer to hash table entry */
    RBTYPE              rhteOverlayName;/* Addr. of ovl name hash tab entry */
    SATYPE              saAbs;
    LNAMETYPE           lnameOverlay;   /* Index into mplnamerhte */
    IOVTYPE             iov;            /* Overlay number */
#if OXOUT OR OIAPX286
    FTYPE               fIsCode;        /* Distinguish text from data */
#endif

    ENTER(SegRc1);                      /* Debug info */
    if(gsnMac >= GSNMAX || snMac >= SNMAX) Fatal("Too many segments.",0x93);
					/* Check for table overflow */
    tysn = Gets();                      /* Read in the ACBP byte */
    align = (ALIGNTYPE)(tysn >> 5) & 7; /* Get the alignment field */
    ASSERT(align != 5);                 /* Not supported by this linker */
    ASSERT(align != 6);                 /* Not supported by this linker */
    if(align == ALGNABS)                /* If absolute LSEG */
      {
	saAbs = (SATYPE) WGets();       /* Get the frame number */
	Gets();                         /* Punt the frame offset */
      }
    seglen = WGets();                   /* Get segment length */
    lname = (LNAMETYPE) GetIndex(1,lnameMac - 1);
					/* Get segment name index */
    lnameClass = (LNAMETYPE) GetIndex(1,lnameMac - 1);
					/* Get class name index */
    lnameOverlay = (LNAMETYPE) GetIndex(0,lnameMac - 1);
					/* Get overlay name index */
    DEBUGVALUE(seglen);                 /* Debug info */
    DEBUGVALUE(lname);                  /* Debug info */
    DEBUGVALUE(lnameClass);             /* Debug info */
    DEBUGVALUE(lnameOverlay);           /* Debug info */
    if(!lnameOverlay) rhteOverlayName = vrhteRootOverlay;
					/* If index zero, use root overlay */
    else                                /* Else if index not zero */
      {
	DEBUGVALUE(mplnamerhte[lnameOverlay]);
					/* Debug info */
	DEBUGVALUE(vrhteRootOverlay);   /* Debug info */
	if(mplnamerhte[lnameOverlay] != vrhteRootOverlay) fOverlays = TRUE;
					/* Overlays if f(name) != root */
	rhteOverlayName = mplnamerhte[lnameOverlay];
					/* Save pointer to hash table entry */
      }
    apropOverlay = (APROPOVERLAYTYPE *)
      PropRhteLookup(rhteOverlayName,ATTROVL,TRUE);
					/* Look up overlay in hash table */
    if(vfCreated)                       /* If it wasn't there before */
      {
	if(iovMac >= IOVMAX) Fatal("Too many overlays.",0x94);
					/* Check for table overflow */
	apropOverlay->ao_iov = iovMac;  /* Save the overlay number */
	mpiovrhte[iovMac++] = rhteOverlayName;
					/* Map iov no. to hash table entry */
      }
    iov = apropOverlay->ao_iov;         /* Get the overlay number */
#if OXOUT OR OIAPX286
    ahte = (AHTETYPE *) ObGetVa(AREASYMS,mplnamerhte[lnameClass],FALSE);
					/* Fetch hash table entry from VM */
    fIsCode = FALSE;                    /* Assume this is not a code segment */
    if(!SbCompare(ahte->cch,"\004TEXT",1) ||
      !SbCompare(ahte->cch,"\004CODE",1)) fIsCode = FCODE;
					/* Check if segment contains code */
#endif
    switch(align)                       /* Switch on alignment type */
      {
	case ALGNABS:                   /* Absolute LSEG */
	case ALGNBYT:                   /* Relocatable byte-aligned LSEG */
	case ALGNWRD:                   /* Relocatable word-aligned LSEG */
	case ALGNPAR:                   /* Relocatable para-aligned LSEG */
	case ALGNPAG:                   /* Relocatable page-aligned LSEG */
	  break;

	default:                        /* ABSMAS, LTL LSEG, or error */
	  ASSERT(0 != 0);               /* We should never get here! */
	  mpsngsn[snMac++] = 0;
	  LEAVE(SegRc1,0xFFFE);         /* Debug info */
	  return;
      }
    if(comb = ((WORD) tysn >> 2) & 7)   /* If C-field of ACBP not zero */
      {
	apropSn = (APROPSNTYPE *)
	  PropRhteLookup(mplnamerhte[lname],ATTRPSN,TRUE);
					/* Look up symbol table entry */
	if(!vfCreated)                  /* If it was already there */
	  {
	    while(apropSn->as_rCla != mplnamerhte[lnameClass] &&
	      apropSn->as_attr != ATTRNIL)
	      {                         /* Look for class match or list end */
		apropSn = (APROPSNTYPE *)
		  ObGetVa(AREASYMS,apropSn->as_next,FALSE);
					/* Advance down the list */
	      }
	    if(apropSn->as_attr != ATTRPSN)
	      {                         /* If attribute not public */
		vfCreated = TRUE;       /* New cell */
		apropSn = (APROPSNTYPE *)
		  PropAdd(mplnamerhte[lname],ATTRPSN);
					/* Segment is public */
	      }
	  }
      }
    else                                /* Else if new entry made */
      {
	vfCreated = TRUE;               /* New cell */
	apropSn = (APROPSNTYPE *) PropAdd(mplnamerhte[lname],ATTRLSN);
					/* Segment is local */
      }
    if(vfCreated)                       /* If a new cell was created */
      {
	apropSn->as_gsn = gsnMac;       /* Assign new global SEGDEF number */
	mpgsnrprop[gsnMac++] = vrprop;  /* Save address of property list */
	apropSn->as_rCla = mplnamerhte[lnameClass];
					/* Save ptr to class hash tab ent */
	DEBUGVALUE(apropSn);            /* Debug info */
	DEBUGVALUE(apropSn->as_rCla);   /* Debug info */
	apropSn->as_tysn = tysn;        /* Save ACBP field */
	apropSn->as_rFil = vrpropFile;
	apropSn->as_iov = iov;          /* Save overlay number */
#if OXOUT OR OIAPX286
	apropSn->as_fCod = fIsCode;     /* Save whether code or not */
#endif
      }
    gsn = apropSn->as_gsn;              /* Save global SEGDEF no. */
    if(comb == COMBSTK)                 /* If segment combines like stack */
      {
	gsnStack = gsn;                 /* Set stack global SEGDEF number */
	align = ALGNBYT;                /* Force to byte alignment */
      }
    else if(comb == COMBCOM)            /* If segment combines like common */
      {
	cbMaxPrev = apropSn->as_cbMx;   /* Get previous segment size */
	apropSn->as_cbMx = 0;           /* Set size to zero */
	if(seglen < cbMaxPrev) seglen = cbMaxPrev;
					/* Take the larger of the two sizes */
      }
    if(align == ALGNWRD) cbMaxPrev = 0xFFFE & (apropSn->as_cbMx + 1);
					/* Round size up to word boundary */
    else if(align == ALGNPAR) cbMaxPrev = 0xFFF0 & (apropSn->as_cbMx + 0xF);
					/* Round size up to para. boundary */
    else if(align == ALGNPAG) cbMaxPrev = 0xFC00 & (apropSn->as_cbMx + 0x3FF);
					/* Round size up to word boundary */
    else cbMaxPrev = apropSn->as_cbMx;  /* Else it's byte-aligned */
    if(align > (apropSn->as_tysn >> 5) & 7)
      {                                 /* If align greater than prev. val. */
	apropSn->as_tysn = (apropSn->as_tysn & 31) | (align << 5);
					/* Use new value */
      }
    if(align != ALGNABS)                /* If not an absolute LSEG */
      {
	if(0xFFFF - seglen < cbMaxPrev)
	  Fatal("Segment size exceeds 64K.",0x95);
					/* Check for segment overflow */
	apropSn->as_cbMx = cbMaxPrev + seglen;
					/* Save new segment size */
      }
    else apropSn->as_cbMx = saAbs;      /* "Hack to save origin of abs seg" */
    mpgsndra[gsn] = cbMaxPrev;          /* Save previous size */
    mpsngsn[snMac++] = gsn;             /* Map SEGDEF no to global SEGDEF no */
    DirtyBp(vbp);                       /* Virtual page has changed */
    LEAVE(SegRc1,0xFFFF);               /* Debug info */
  }

    /****************************************************************
    *                                                               *
    *  GrpRc1:                                                      *
    *                                                               *
    *  This function processes GRPDEF records on pass 1.            *
    *  See pp. 36-39 in "8086 Object Module Formats EPS."           *
    *                                                               *
    ****************************************************************/

GrpRc1()
  {
    LNAMETYPE           lnameGroup;     /* Group LNAME number */
    SNTYPE              sn;             /* Group (local) segment number */
    APROPSNTYPE         *apropSn;
    APROPGROUPTYPE      *apropGroup;
    GRTYPE              ggr;            /* Global GRPDEF number */
    WORD                gcdesc;         /* GRPDEF component descriptor */

    ENTER(GrpRc1);                      /* Debug info */
    if(grMac++ >= GRMAX) Fatal("Too many GRPDEFs in module.",0x96);
					/* Check for table overflow */
    lnameGroup = GetIndex(1,lnameMac - 1);
					/* Read in group name index */
    apropGroup = (APROPGROUPTYPE *)
      PropRhteLookup(mplnamerhte[lnameGroup],ATTRGRP,TRUE);
					/* Look up cell in hash table */
    if(vfCreated)                       /* If entry did not exist before */
      {
	if(ggrMac >= GGRMAX) Fatal("Too many groups.",0x96);
					/* Check for table overflow */
	apropGroup->ag_ggr = ggrMac++;  /* Save global GRPDEF no. */
      }
    ggr = apropGroup->ag_ggr;           /* Get global GRPDEF no. */
    while(cbRec > 1)                    /* While not at end of record */
      {
	gcdesc = B2W(Gets());           /* Read in the descriptor */
	ASSERT(gcdesc == 0xFF);         /* Linker doesn't handle others */
	sn = GetIndex(1,snMac);         /* Get local SEGDEF index */
	apropSn = (APROPSNTYPE *)
	  ObGetVa(AREASYMS,mpgsnrprop[mpsngsn[sn]],TRUE);
					/* Fetch/create from virt mem */
	if(apropSn->as_ggr == GRNIL) apropSn->as_ggr = ggr;
					/* Store global GRPDEF no. if none */
	ASSERT(apropSn->as_ggr == ggr); /* Cannot belong to multiple grps */
      }
    LEAVE(GrpRc1,0xFFFF);               /* Debug info */
  }

    /****************************************************************
    *                                                               *
    *  PubRc1:                                                      *
    *                                                               *
    *  This function processes PUBDEF records on pass 1.            *
    *  See pp. 44-46 in "8086 Object Module Formats EPS."           *
    *                                                               *
    ****************************************************************/

PubRc1()
  {
    WORD                grpind;         /* Group index in base */
    SNTYPE              sn;             /* Local SEGDEF no. */
    SNTYPE              gsn;            /* Global SEGDEF no. */
    RATYPE              dra;
    SBTYPE              sb;             /* Public symbol */
    RATYPE              ra;             /* Public symbol offset */
    APROPNAMETYPE       *apropName;     /* Table entry for symbol name */
#if OXOUT OR OIAPX286
    ATTRTYPE            attr;           /* Symbol attribute */
    APROPUNDEFTYPE      *apropUndef;    /* Undefined symbol entry */
    APROPSNTYPE         *apropSn;
    WORD                len;            /* Symbol length */
#endif
    BSTYPE              bsTmp;          /* Used to save file handle */

    ENTER(PubRc1);                      /* Debug info */
    DEBUGVALUE(B2W(grMac - 1));         /* Debug info */
    grpind = GetIndex(0,B2W(grMac - 1));/* Get group index */
    DEBUGVALUE(grpind);                 /* Debug info */
    if(!(sn = GetIndex(0,snMac - 1)))   /* If frame number present */
      {
	gsn = 0;                        /* No global SEGDEF no. */
	dra = 0;
	SkipBytes(2);                   /* Skip the frame number */
      }
    else                                /* Else if local SEGDEF no. given */
      {
	gsn = mpsngsn[sn];              /* Get global SEGDEF no. */
	dra = mpgsndra[gsn];
      }
    DEBUGVALUE(cbRec);                  /* Debug info */
    while(cbRec > 1)                    /* While there are symbols left */
      {
	sb[0] = Gets();                 /* Get symbol length */
	GetBytes(&sb[1],B2W(sb[0]));    /* Read in symbol text */
	ra = WGets();                   /* Get symbol segment offset */
	GetIndex(0,0x7FFF);             /* Skip type index */
	apropName = (APROPNAMETYPE *) PropSymLookup(sb,ATTRUND,FALSE);
					/* Look for symbol among undefined */
	if(apropName != PROPNIL) vfCreated = TRUE;
					/* Symbol known to be undefined */
	else apropName = (APROPNAMETYPE *)
	  PropSymLookup(sb,ATTRPNM,TRUE);
					/* Else try to create new entry */
	if(vfCreated)                   /* If new PUBNAM entry created or */
	  {                             /* old UNDEF entry to modify */
	    if(extMac >= EXTMAX)
	      Fatal("Too many public symbols in one module.",0x97);
					/* Check for table overflow */
	    mpextsn[extMac] = sn;       /* Map PUBDEF no. to SEGDEF no. */
	    mpextra[extMac++] = ra;     /* Map PUBDEF no. to offset */
#if OXOUT OR OIAPX286
	    attr = apropName->an_attr;  /* Get symbol's attribute */
	    if(attr == ATTRUND)         /* If undefined symbol */
	      {
		apropUndef = (APROPUNDEFTYPE *) apropName;
					/* Recast pointer */
		len = apropUndef->au_len;
					/* Save length of undefined object */
	      }
	    else len = 0;               /* Else set length of object to zero */
#endif
	    apropName->an_attr = ATTRPNM;
					/* Symbol is a public name */
	    apropName->an_ra = ra + dra;/* Give symbol its adjusted offset */
	    apropName->an_gsn = gsn;    /* Save its global SEGDEF no. */
	    DirtyBp(vbp);               /* Mark virtual page as changed */
	  }
	else                            /* Else if symbol already defined */
	  {
	    ++cErrors;                  /* Increment error count */
	    bsTmp = stdout;             /* Error to screen first */
	    for(;;)                     /* Hack! */
	      {
		OutError(bsTmp,"symbol defined more than once: ");
		OutSb(bsTmp,sb);        /* Output the symbol */
		fputs(" in file ",bsTmp);
		OutFileCur(bsTmp);      /* Output current file name */
		NEWLINE(bsTmp);         /* Newline */
		fflush(bsTmp);          /* Force message out */
		if(!fLstFileOpen || bsTmp == bsLst) break;
					/* Exit condition */
		bsTmp = bsLst;          /* Send message to list file */
	      }
	  }
      }
    LEAVE(PubRc1,0xFFFF);               /* Debug info */
  }

    /****************************************************************
    *                                                               *
    *  ExtRc1:                                                      *
    *                                                               *
    *  This function processes EXTDEF records on pass 1.            *
    *  See pp. 47-48 in "8086 Object Module Formats EPS."           *
    *                                                               *
    ****************************************************************/

ExtRc1()
  {
    SBTYPE              sb;             /* EXTDEF symbol */
    APROPUNDEFTYPE      *apropUndef;    /* Pointer to symbol entry */
#if CALLOCATION
    WORD                itype;          /* Type index */
#endif

    ENTER(ExtRc1);                      /* Debug info */
    while(cbRec > 1)                    /* While there are symbols left */
      {
	sb[0] = Gets();                 /* Get symbol length */
	GetBytes(&sb[1],B2W(sb[0]));    /* Read in text of symbol */
#if CALLOCATION
	itype = GetIndex(0,typMac - 1); /* Get type index */
#else
	GetIndex(0,0x7FFF);             /* Punt type index */
#endif
	apropUndef = (APROPUNDEFTYPE *) PropSymLookup(sb,ATTRPNM,FALSE);
					/* Look for a matching PUBDEF */
	if(apropUndef == PROPNIL)       /* If there isn't one */
	  {
	    apropUndef = (APROPUNDEFTYPE *)
	      PropSymLookup(sb,ATTRUND,TRUE);
					/* Insert as undefined symbol */
#if CALLOCATION
	    if(apropUndef->au_len < mpitypelen[itype])
	      apropUndef->au_len = mpitypelen[itype];
					/* Save longest length specification */
#endif
	  }
      }
    LEAVE(ExtRc1,0xFFFF);               /* Debug info */
  }

    /****************************************************************
    *                                                               *
    *  ComRc1:                                                      *
    *                                                               *
    *  This function processes COMENT records on pass 1.            *
    *  See pp. 86-87 in "8086 Object Module Formats EPS."           *
    *                                                               *
    ****************************************************************/

ComRc1()
  {
    WORD                class;          /* Comment class */
    SBTYPE              text;           /* Comment text */

    ENTER(ComRc1);                      /* Debug info */
    Gets();                             /* Skip byte 1 of comment type field */
    class = B2W(Gets());                /* Get comment class byte */
    text[0] = (BYTE) cbRec - 1;         /* Get length of comment */
    GetBytes(&text[1],cbRec - 1);       /* Read in text of comment */
#if LIBMSDOS
    if(!class)                          /* If translator comment */
      {
	if(!SbCompare(text,"\011MS PASCAL",1))
	  {                             /* If compiled by MS-Pascal */
	    if(!vfNoDefaultLibrarySearch) AddLibrary(sbPascalLib);
					/* Add Pascal library to search list */
	    SetPascalDefaults();        /* We are linking a Pascal program */
	  }
	else if(!SbCompare(text,"\012FORTRAN 77",1))
	  {                             /* If compiled by MS-FORTRAN77 */
	    SetPascalDefaults();        /* Use Pascal defaults */
	  }
      }
    else                                /* Else if class is not zero */
#endif
    if(class == 0x81 && !vfNoDefaultLibrarySearch) AddLibrary(text);
					/* Add name to search list */
    LEAVE(ComRc1,0xFFFF);               /* Debug info */
  }

    /****************************************************************
    *                                                               *
    *  ProcP1:                                                      *
    *                                                               *
    *  This function  controls the processing of an  object module  *
    *  on pass 1.                                                   *
    *                                                               *
    ****************************************************************/

ProcP1()
  {
#if LIBXENIX
    LFATYPE             lfa;
#endif

    ENTER(ProcP1);                      /* Debug info */
    snMac = 1;                          /* Initialize counter */
    grMac = 1;                          /* Initialize */
    extMac = 1;                         /* Initialize counter */
    lnameMac = 1;                       /* Initialize counter */
#if CALLOCATION
    typMac = 1;                         /* Initialize counter */
#endif
    DEBUGVALUE(gsnMac);                 /* Debug info */
    DEBUGVALUE(ggrMac);                 /* Debug info */
#if LIBXENIX
    lfa = ftell(bsInput);               /* Save initial file position */
    cbRec = WGets();                    /* Read a word */
    if(cbRec == X_MAGIC)                /* If magic number found */
      {
	cbRec = 30;                     /* Size of x.out header minus 2 */
	SkipBytes(26);                  /* Skip to cpu specification */
	if(Gets() != XC_8086) InvalidObject();
					/* Bad if not 8086 */
	if(Gets() != XR_R86REL + XR_S86REL) InvalidObject();
	if(WGets() != XE_V2) InvalidObject();
      }
    else fseek(bsInput,lfa,0);          /* Else return to start */
#endif
    for(;;)                             /* Loop to process object module */
      {
	rect = Gets();                  /* Read record type */
	cbRec = WGets();                /* Read record length */
	DEBUGVALUE(rect);               /* Debug info */
	DEBUGVALUE(cbRec);              /* Debug info */
	if(!cbRec || B2W(rect) < 0x6E || B2W(rect) > 0xAA || (rect & 1)) break;
					/* Break if invalid object */
	switch(rect)                    /* Switch on record type */
	  {
#if CALLOCATION
	    case TYPDEF:                /* rect = 8Eh */
	      TypRc1();
	      break;
#endif

	    case SEGDEF:
	      SegRc1();
	      break;

	    case THEADR:
	      ModRc1();
	      break;

	    case COMENT:
	      ComRc1();
	      break;

	    case LHEADR:
	      ModRc1();
	      break;

	    case GRPDEF:
	      GrpRc1();
	      break;

	    case EXTDEF:
	      ExtRc1();
	      break;

	    case LNAMES:
	      LNmRc1();
	      break;

	    case PUBDEF:
	      PubRc1();
	      break;

	    case MODEND:                /* End of module */
	      LEAVE(ProcP1,0xFFFF);     /* Debug info */
	      return;                   /* And return */

	    default:
	      SkipBytes(cbRec - 1);     /* Skip to checksum byte */
	      break;
	  }
	if(cbRec != 1) break;           /* If record length bad */
	Gets();                         /* Eat the checksum byte */
      }
    InvalidObject();                    /* Invalid module */
  }
