    /****************************************************************
    *                                                               *
    *                           XENIN2.C                            *
    *                                                               *
    *  This module contains the XENIX command line processor.       *
    *                                                               *
    ****************************************************************/

    /****************************************************************
    *                                                               *
    *  The command form is assumed to be:                           *
    *                                                               *
    *  ld [options]    where options are:                           *
    *  <file>          a file to be included in the link            *
    *  -l <file>       search library file                          *
    *  -m <file>       create a memory map and send it to file      *
    *                  this will include line numbers and symbols   *
    *  -o <file>       the output will be file, default is x.out    *
    *  -i              Separate I and D space                       *
    *  -C              case is significant                          *
    *  -D <hex no>     (286 only) Set value for virtual base of     *
    *                  data segment                                 *
    *  -F <hex no>     fixed stack size                             *
    *  -L              low word first in long integers              *
    *  -M              allow medium model (>64K code)               *
    *  -S              dump out symbol table                        *
    *                                                               *
    ****************************************************************/

#include                "minlit.h"      /* Types and constants */
#include                "bndtrn.h"      /* More types and constants */
#include                "lnkio.h"       /* I/O definitions */

extern                  MainFn();       /* Main control function */
extern BYTE             *ObGetVa();     /* Fetch from virtual memory */
extern PROPTYPE         PropSymLookup();/* Symbol table lookup */
extern                  Fatal();        /* Fatal error procedure */
extern                  AddLibrary();   /* Add library to search list */
extern                  ProcObject();   /* Process input object file */
#define IsSwitch(s)     ((s)[0] == '-') /* True if string is a switch */

extern FTYPE            vfNoDefaultLibrarySearch;
					/* Self-explanatory!! */
extern FTYPE            vfLineNos;      /* True if line numbers requested */
extern FTYPE            vfMap;          /* True if map file requested */
extern FTYPE            fMedium;        /* True if program is medium model */
extern FTYPE            fIandD;         /* True if separate code and data */
extern FTYPE            fIgnoreCase;    /* True if upper and lower case same */
extern FTYPE            fSymbol;        /* True if symbol file requested */
extern FTYPE            fLstFileOpen;   /* True if list file is open */
extern FTYPE            fLoWordFirst;   /* True if lo word first in ranlib */
extern BSTYPE           bsLst;          /* List file descriptor */
extern IOVTYPE          iovMac;         /* Number of overlays */
extern WORD             cbStack;        /* Stack size */
extern WORD             carg;           /* Argument count */
extern char             **varg;         /* Argument vector */
extern char             *lnknam;        /* Linker name */
#if OIAPX286
extern SATYPE           saData;         /* Virtual base of data segment */
#endif

    /****************************************************************
    *                                                               *
    *  MakeSb:                                                      *
    *                                                               *
    *  Make a length-prefixed  string given a null-terminated one.  *
    *                                                               *
    ****************************************************************/

MakeSb(sz,sb)
char                    *sz;            /* Null-terminated string */
char                    *sb;            /* Length-prefixed string */
  {
    register WORD       i;              /* Index variable */

    i = 0;                              /* Initialize index */
    while(*sz) sb[++i] = *sz++;         /* Copy string */
    sb[0] = (char) i;                   /* Store length */
  }

    /****************************************************************
    *                                                               *
    *  S2H:                                                         *
    *                                                               *
    *  This function converts a string to an integer in hex.        *
    *                                                               *
    ****************************************************************/

LOCAL WORD              S2H(s)          /* Convert string to hex integer */
char                    *s;             /* Null-terminated string */
  {
    WORD                h;              /* Hex integer */

    h = 0;                              /* Initialize */
    while(*s)                           /* While not at end of string */
      {
	if(UPPER(*s) >= 'A' && UPPER(*s) <= 'F')
	  h = h*16 + UPPER(*s) - 'A' + 10;
					/* If letter, treat accordingly */
	else if(*s >= '0' && *s <= '9') h = h*16 + *s - '0';
					/* Else if number */
	++s;                            /* Increment pointer */
      }
    return(h);                          /* Return integer */
  }

    /****************************************************************
    *                                                               *
    *  ParseCmdLine:                                                *
    *                                                               *
    *  This function parses the command line.                       *
    *                                                               *
    ****************************************************************/

ParseCmdLine()
  {
    WORD                i;              /* Index variable */
    SBTYPE              sbFile;         /* File name */
    SBTYPE              sbRunFile;      /* Run file name */
    APROPOVERLAYTYPE    *papropOverlay; /* Overlay property cell pointer */
    FTYPE               fObjects;       /* True if objects are specified */

    fLstFileOpen = FALSE;               /* Assume no list file */
    vfMap = FALSE;                      /* Don't show publics */
    vfLineNos = FALSE;                  /* Don't show line numbers */
    fObjects = FALSE;                   /* No objects yet */
    fIandD = FALSE;                     /* Assume no separate code and data */
    fLoWordFirst = FALSE;               /* Assume high word first in ranlib */
    fMedium = FALSE;                    /* Assume small model */
    fIgnoreCase = TRUE;                 /* Assume not case-sensitive */
    fSymbol = FALSE;                    /* Assume no symbol file */
    Move(6,"\005x.out",sbRunFile);      /* Default run file name */
    cbStack = 0;                        /* Assume default stack spec. */
#if OIAPX286
    saData = DFSADATA;                  /* Assume default value */
#endif
    i = 0;                              /* Initialize index */
    if(!carg)                           /* If no arguments specified */
      {
	fprintf(stderr,"usage: %s %s%s",lnknam,
#if OIAPX286
	  "[-C][-D <hex #>][-F <hex #>][-L][-M][-S][-i][-l <lib>]",
#else
	  "[-C][-F <hex #>][-L][-M][-S][-i][-l <lib>]",
#endif
	  "[-m <map>][-o <name>] <objects>");
					/* Usage message */
	NEWLINE(stderr);                /* Newline */
	fflush(stderr);                 /* Force message to screen */
	exit(1);                        /* Error exit */
      }
    while(i < carg)                     /* While arguments remain */
      {
	if(IsSwitch(varg[i]))           /* If argument is a switch */
	  {
	    switch(varg[i][1])          /* Switch on switch */
	      {
		case 'l':               /* Library specification */
		  if(++i == carg) Fatal("library name missing",1);
		  MakeSb(varg[i],sbFile);
					/* Make length-prefixed string */
		  AddLibrary(sbFile);   /* Add file name to search list */
		  break;

		case 'm':               /* Map file specification */
		  if(++i == carg) Fatal("list file name missing",1);
		  if(!fLstFileOpen)     /* If no list file yet */
		    {
		      if((bsLst = fopen(varg[i],WRTXT)) == NULL)
			Fatal("cannot create list file",1);
					/* Create list file */
		      fLstFileOpen = TRUE;
					/* We have a list file */
		      vfMap = TRUE;     /* Show public symbols */
		      vfLineNos = TRUE; /* Show line numbers */
		    }
		  break;

		case 'o':               /* Output file specification */
		  if(++i == carg) Fatal("run file name missing",1);
		  MakeSb(varg[i],sbRunFile);
					/* Set up run file name */
		  break;

		case 'i':               /* Separate code and data */
		  fIandD = TRUE;        /* Set flag */
		  break;

		case 't':               /* Map to std. output */
		  if(!fLstFileOpen)     /* If no list file */
		    {
		      fLstFileOpen = TRUE;
					/* Now there is a list file */
		      bsLst = stdout;   /* List to standard output */
		    }
		  break;

		case 'C':               /* Be case-sensitive */
		  fIgnoreCase = FALSE;  /* Do not ignore case */
		  break;

#if OIAPX286
		case 'D':               /* Specify virtual base of data seg */
		  if(++i == carg) Fatal("data selector missing",1);
					/* Check for argument */
		  if((saData = S2H(varg[i])) >= STBIAS)
		    Fatal("data selector too large",1);
					/* Convert arg to hex integer */
		  break;
#endif

		case 'F':               /* Specify fixed stack size */
		  if(++i == carg) Fatal("stack size missing",1);
					/* Check for argument */
		  cbStack = S2H(varg[i]);
					/* Convert arg to hex integer */
		  break;

		case 'L':               /* Low word first */
		  fLoWordFirst = TRUE;  /* Set flag */
		  break;

		case 'M':               /* Medium model */
		  fIandD = TRUE;        /* Separate code and data */
		  fMedium = TRUE;       /* Medium model */
		  break;

		case 'S':               /* Make symbol file */
		  fSymbol = TRUE;       /* Set flag */
		  break;

		default:                /* Unknown switch */
		  fprintf(stderr,"%s: unknown switch \"%s\"",lnknam,varg[i]);
					/* Message */
		  NEWLINE(stderr);      /* Newline */
		  Fatal("invalid command line",1);
					/* Fatal error */
	      }
	  }
	else                            /* Else if not switch */
	  {
	    MakeSb(varg[i],sbFile);     /* Make length-prefixed string */
	    ProcObject(sbFile);         /* Add to list of objects */
	    fObjects = TRUE;            /* We have some objects */
	  }
	++i;                            /* Increment index */
      }
    if(!fObjects) Fatal("no object files specified",1);
					/* There must be object files */
    papropOverlay =
      (APROPOVERLAYTYPE *) PropSymLookup(sbRunFile,ATTROVL,TRUE);
					/* Install run file in sym tab */
    papropOverlay->ao_iov = iovMac - 1; /* Set overlay number for run file */
  }
